/***********************************************************************************************************************
 * Schnittstelle fuer die Prozeduren zur Zeitmessung in der Regelmengenverwaltung 
 ***********************************************************************************************************************
 * Autor:     Juergen Sellentin. Implementiert am 13.02.94
 * Angepasst: Arnim am 04.03.94.  Makros und Funtionen eingesetzt.
 *            Arnim am 04.04.94.  Stack zur Verwaltung von geschachtelten Aufrufen eingebaut.
 *            Arnim bis 24.05.95. Komplett neues Konzept, gettimeofday und Zaehler eingebaut.  
 ***********************************************************************************************************************
 *
 * Bemerkungen zur Zeitmessung:
 * ----------------------------
 *
 * Die Zeitmessungen werden mit der Systemfunktion "getrusage" realisiert. Diese protokolliert die seit dem Beginn des
 * Prozesses verbrauchte Zeit (als user- und system- Anteil).
 *
 * Die Genauigkeit dieser Messung haengt sehr stark von der update-Frequenz von getrusage ab !!!
 * Bei sehr kurzen Prozeduraufrufen werden die getrusage-Daten evtl. NICHT veraendert, so dass die Laufzeit 0
 * als Ergebnis herauskommt. Andererseits kann eine zu hohe Laufzeit gemessen werden, wenn das Aktualisieren der Daten
 * (von getrusage) gerade in einer sehr kurzen Prozedur durchgefuehrt wird.
 * Im statistischen Mittel gleichen sich diese Fehler hoffentlich aus!
 *
 * ACHTUNG: Schachtelungstiefe der Aufrufe begrenzt (siehe Konstante ZM_StackSize im .c-File). Bei hoeherem Bedarf 
 *          die Konstante anpassen, da nicht auf Ueberlauf geprueft wird (Zeit!)
 *
 ***********************************************************************************************************************/


#ifndef ZEITMESSUNG_H
#define ZEITMESSUNG_H

#include "general.h"
#include "InfoStrings.h"
#include "Parameter.h"
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#endif


/* ***************************************************************** */
/* * (0) Granularitaet und anderes                                 * */
/* ***************************************************************** */

/* Granularitaet der Zeitmessung:

   Level 0: Nur die Gesamtzeit wird gemessen (dies geschieht immer mit getrusage)
   Level 1: Zusaetzlich werden die Gesamtzeiten fuer 
            * NFBildung,
            * KPBildung mit neuer Regel, 
            * KPBildung mit neuer Gleichung,
            * Interreduktion,
            * Subsumption,
            * Richten der vom Vervollstaendiger angefassten KPs
            gemessen.
   Level 2: Zusaetzlich wird die KP_Bildung in MGU-Suche, Groesser-Test und ggf. Kapur-Test aufgegliedert,
            jeweils unterschieden nach 'zu neuer Regel bzw. Gleichung'.
   Level 3: Zusaetzlich wird die NFBildung in Match-Suche, Termerzeugung, Groesser-Test etc. aufgespalten
   Level 4: Zusaetzlich wird der Zeitbedarf fuer die Verwaltung von KP-, Regel- und Gleichungsmenge sowie die Anlage und Initialisierung
            neuer Speicherbloecke gemessen.
   
   Ferner kann eingestellt werden, ob die Zeitmessung die Funktion gettimeofday (Realzeit mit Aufloesung 1us (nur user-time)) oder getrusage 
   (Prozesszeit mit Aufloesung 10ms (user- und system-time)) verwendet.
   Bei Verwendung von gettimeofday wird keine system-time gemessen. Stattdessen kann noch eingestellt werden, dass die Anzahl der
   Zeitmessungsaufrufe mitgemessen werden soll. Diese Anzahl erscheint dann in der Statistik als Sekunden-Anteil der system-time.
   Ferner wird die Gesamtzahl der Zeitmessungsaufrufe gezaehlt, wenn ZM_STATISTIK auf 1 gesetzt ist (auch bei getrusage).

   Module sind: ZeitMessung.h, ZeitMessung.c
*/

#define ZM_MESSLEVEL 0
/* Alternativen sind: 0 1 2 3 4 */

#define ZM_STATISTIK normal_power_comp(0,0,0)
/* Alternativen: 0 fuer aus, 1 fuer an */

/* ***************************************************************** */
/* * (II) Initialisierung und Allgemeine Zeitmessung               * */
/* ***************************************************************** */

void ZM_StartGesamtZeit(void);                       /* Die Gesamtzeitmessung wird immer ausgefuehrt! */
void ZM_StoppGesamtZeit(void);

/* ***************************************************************** */
/* * (III) Variablen                                               * */
/* ***************************************************************** */

extern struct timeval       ZM_GesamtZeit_ut;               /* insgesamt verbrauchte Zeit exklusive Parsen und Plattklopfen. */
extern struct timeval       ZM_GesamtZeit_st;


/* ***************************************************************** */
/* * (IV) Statistik                                                * */
/* ***************************************************************** */

void ZM_Statistik(InfoStringT IS);

#ifdef __cplusplus
}
#endif

#endif
