/**********************************************************************
* Filename : NewClassification.
*
* Kurzbeschreibung : Classifikation von CPs und CGs
*
* Autoren : Andreas Jaeger
*
* 10.03.98: Erstellung
*
*
*
**********************************************************************/
#define __EXTENSIONS__ 1
#include "antiwarnings.h"
#include <limits.h>
#include <string.h>

#include "compiler-extra.h"
#include "general.h"

#include "Ausgaben.h"
#include "CStatistics.h"
#include "ClasFunctions.h"
#include "ClasHeuristics.h"
#include "ClasPatch.h"
#include "Clas_CP_Goal.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "NewClassification.h"
#include "OccNest.h"
#include "Ordnungen.h"
#include "RUndEVerwaltung.h"
#include "SymbolGewichte.h"
#include "TermOperationen.h"

typedef enum  {Crit_KillerR,
               Crit_KillerRE,
               Crit_KillerE,
               Crit_EChild,
               Crit_LastMarker /* Ende Marker */
} CriteriaEnumType;


/* Act_normal ist nur der Vollstaendigkeit halber in der Liste - es wird nichts gemacht */
typedef enum {Act_ultimate, Act_normal, Act_double, Act_half, Act_never,
              Act_LastMarker
} ActionEnumType;


typedef enum {Heu_AddWeight, Heu_CP_In_Goal, Heu_GLAM, Heu_Goal_In_CP, Heu_GtWeight,
              Heu_MixWeight2, Heu_MixWeight, Heu_MaxWeight, Heu_OrdWeight, Heu_PolyWeight,
              Heu_OccNest, Heu_Unifikationsmass,
	      Heu_CPPatch, Heu_CGPatch,
              Heu_Fifo, Heu_Golem,
	      Heu_LastMarker
} HeuristicsEnumType;


typedef struct 
{
  CriteriaEnumType criterion;
  ActionEnumType action;
  CriteriaProcType critProc;
  CounterT fullfilled;
} CriterionType;


typedef struct 
{
  CPHistoryT context;
  ActionEnumType action;
} ContextType;

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe Classification.h) --------------------*/
/* ------------------------------------------------------------------------*/



void (* C_WeightStatisticsStartProc) (InfoStringT);
void (* C_WeightStatisticsEndProc) (void);
PQ_WatchingWalkthroughProcedureT C_WeightStatProc;

/* lokale, globale Vars */
/* fuer FIFO-Strategie: Einfacher Zaehler */
static long int CPNr;

static CriterionType *ClasCriteria;
static ClassifyProcT ClassifyProc;
static ClassifyProcT CGClassifyProc;
static CriterionType *ReClasCriteria;
static ClassifyProcT ReClassifyProc;

#if !COMP_POWER  
static struct {
  CPHistoryT crit;
  const char *name;
} NameContext[] = { /* Reihenfolge muss mit SUEBasis uebereinstimmen */
  {CPWithFather, "CPWithFather"},
  {CPWithMother, "CPWithMother"},
  {Initial, "Initial"},
  {KilledRE, "KilledRE"},
  {ChaseManhattanCP, "Database"},
  {TWreceived, "TWreceived"},
  {CPHistory_Last, "The End"}
};
#endif

static struct {
  CriteriaEnumType crit;
  const char *name;
  CriteriaProcType critProc;
} NameCriterion[] = {
  {Crit_EChild, "EChild", CF_EChildPraedikat},
  {Crit_KillerR, "KillerR", CF_KillerPraedikatR},
  {Crit_KillerRE, "KillerRE", CF_KillerPraedikatRE},
  {Crit_KillerE, "KillerE", CF_KillerPraedikatE},
  {Crit_LastMarker, "The End", NULL}
}
;

static struct 
{
  const char *name;
  ActionEnumType action;
} action_parse [] = {
    {"ultimate", Act_ultimate},
    {"normal", Act_normal},
    {"double", Act_double},
    {"half", Act_half},
    {"last", Act_never},
    {"", Act_LastMarker}
};


static struct 
{
  const char *name;
  HeuristicsEnumType heuristic;
  ClassifyProcT classProc;
} heuristics_parse [] = {
    {"addweight", Heu_AddWeight, &CH_AddWeight},
    {"cpingoal", Heu_CP_In_Goal, &CPinGoal},
    {"glam", Heu_GLAM, NULL},
    {"goalincp", Heu_Goal_In_CP, &GoalinCP},
    {"gtweight", Heu_GtWeight, &CH_GtWeight},
    {"Mixweight", Heu_MixWeight2, &CH_MixWeight2},
    {"mixweight", Heu_MixWeight, &CH_MixWeight},
    {"maxweight", Heu_MaxWeight, &CH_MaxWeight},
    {"ordweight", Heu_OrdWeight, &CH_OrdWeight},
    {"polyweight", Heu_PolyWeight, &CH_PolyWeight},
    {"occnest", Heu_OccNest, &CH_OccNest},
    {"unif", Heu_Unifikationsmass, &CH_Unifikationsmass},
    {"cppatch", Heu_CPPatch, &CPA_CPPatch},
    {"cgpatch", Heu_CGPatch, &CPA_CGPatch},
    {"fifo", Heu_Fifo, &CH_Fifo},
    {"Addweight", Heu_Golem, &CH_Golem},
    {"", Heu_LastMarker, NULL}
};

static ContextType HistorySpec[CPHistory_Last] =
{ /* Reihenfolge muss mit SUEBasis uebereinstimmen */
  {CPWithFather, Act_normal},
  {CPWithMother, Act_normal},
  {Initial, Act_normal},
  {KilledRE, Act_normal},
  {ChaseManhattanCP, Act_normal},
  {TWreceived, Act_normal}
};

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Prototypen ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static const char *
get_criterion (CriteriaEnumType singleCrit);


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Output ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void C_CleanUp(void)
{
   CS_CleanUp();
}


void C_Statistics(InfoStringT IS)
{
  int i;
  char str[100];
  
  if (!(IS_IsIn("C", IS) || IS_IsIn("all", IS)))
    return;

  if (IS_IsIn("standard", IS)) {
    IO_DruckeKurzUeberschrift("Classification", '+');
    if (ClasCriteria) {
      for (i = 0; ClasCriteria[i].criterion != Crit_LastMarker; i++)
        {
          strcpy (str, get_criterion (ClasCriteria[i].criterion));
          strcat (str, " during CLASSIFY");
          IO_DruckeStatistikZeile(str, ClasCriteria[i].fullfilled);
        }
    }
    if (ReClasCriteria) {
      for (i = 0; ReClasCriteria[i].criterion != Crit_LastMarker; i++)
        {
          strcpy (str, get_criterion (ReClasCriteria[i].criterion));
          strcat (str, " during RECLASSIFY");
          IO_DruckeStatistikZeile(str, ReClasCriteria[i].fullfilled);
        }
    }
  }
  else if (IS_IsIn("latex-head", IS)){
    if (ClasCriteria) {
      fprintf (IO_Strom, "CLAS: ");
      for (i = 0; ClasCriteria[i].criterion != Crit_LastMarker; i++)
        {
          fprintf (IO_Strom, " %s & ", get_criterion (ClasCriteria[i].criterion));
        }
    }
    if (ReClasCriteria) {
      fprintf (IO_Strom, "RECLAS: ");
      for (i = 0; ReClasCriteria[i].criterion != Crit_LastMarker; i++)
        {
          fprintf (IO_Strom, " %s & ", get_criterion (ReClasCriteria[i].criterion));
        }
    }
  }
}


/* ------------------------------------------------------------------------------------------------*/
/* ------------ Classifikationen ------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------*/

/* Normale Classifikation von CPs mit Criterium */

void
C_Classify (WeightEntryT *wtPtr, TermT lhs, TermT rhs,
            RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent,
            UTermT actualPosition,
            CPHistoryT History)
{
  int i;
  BOOLEAN res = FALSE;
  ActionEnumType action;

  action = HistorySpec [History].action;

  wtPtr->w2 = ++CPNr; /* FIFO */
  
  switch (action)
    {
    case Act_ultimate:
      wtPtr->w1 = LONG_MIN;
      return;
    case Act_never:
      wtPtr->w1 = LONG_MAX;
      return;
    default:
      break;
    }

  if (ClasCriteria) {
    for (i = 0; ClasCriteria[i].criterion != Crit_LastMarker; i++)
      {
        res = ClasCriteria[i].critProc (lhs, rhs, actualParent, otherParent);
        if (res) {
          ClasCriteria[i].fullfilled++;
          if (ClasCriteria[i].action == Act_ultimate) {
            wtPtr->w1 = LONG_MIN;
            return;
          }
          if (ClasCriteria[i].action == Act_never) {
            wtPtr->w1 = LONG_MAX;
            return;
          }
          action = ClasCriteria[i].action;
          /* OK, normale Classifikation durchfuehren */
          break;
        }
      }
  }

  /* normale Classifikation */
  wtPtr->w1 = (*ClassifyProc) (lhs, rhs, actualParent, otherParent, actualPosition, TRUE);

  switch (action)
    {
    case Act_double:
      wtPtr->w1 = wtPtr->w1 * 2;
      break;
    case Act_half:
      wtPtr->w1 = wtPtr->w1 / 2;
      break;
    default:
      break;
    }
}



void
C_CG_Classify (WeightEntryT *wtPtr, TermT lhs, TermT rhs,
               RegelOderGleichungsT actualParent,
               RegelOderGleichungsT otherParent)
{
  /* normale Classifikation */
  wtPtr->w2 = ++CPNr; /* FIFO */
  wtPtr->w1 = (*CGClassifyProc) (lhs, rhs, actualParent, otherParent, NULL, FALSE);

}


void
C_ReClassify (WeightEntryT *wtPtr, TermT lhs, TermT rhs,
               RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
{
  int i;
  BOOLEAN res = FALSE;
  ActionEnumType action = Act_normal;

  /* w2 wird nicht geaendert */

  if (ReClasCriteria)
    {
      for (i = 0; ReClasCriteria[i].criterion != Crit_LastMarker; i++)
        {
          res = ReClasCriteria[i].critProc (lhs, rhs, actualParent, otherParent);
          if (res) {
            ReClasCriteria[i].fullfilled++;
            if (ReClasCriteria[i].action == Act_ultimate) {
              wtPtr->w1 = LONG_MIN;
              return;
            }
            if (ReClasCriteria[i].action == Act_never) {
              wtPtr->w1 = LONG_MAX;
              return;
            }
            action = ReClasCriteria[i].action;
            /* OK, normale Classifikation durchfuehren */
            break;
          }
        }
    }
  
  /* normale Classifikation */
  wtPtr->w1 = (*ReClassifyProc) (lhs, rhs, actualParent, otherParent, NULL, TRUE);

  if (res)
    {
      switch (action)
        {
        case Act_double:
          wtPtr->w1 = wtPtr->w1 * 2;
          break;
        case Act_half:
          wtPtr->w1 = wtPtr->w1 / 2;
          break;
        default:
          break;
        }
    }
}




/* ------------------------------------------------------------------------------------------------*/
/* ------------ ParamInfo -------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------*/

static const char *
get_action (ActionEnumType action)
{
  int i;

  for (i = 0; action_parse[i].action != Act_LastMarker; i++)
    {
      if (action_parse[i].action == action)
        return action_parse[i].name;
    }

  return "";
}


static const char *
get_criterion (CriteriaEnumType singleCrit)
{
  int i;

  for (i = 0; NameCriterion[i].crit != Crit_LastMarker; i++)
    {
      if (NameCriterion[i].crit == singleCrit)
        return NameCriterion[i].name;
    }
  
  return "";
}

static CriteriaProcType
get_crit_proc (CriteriaEnumType singleCrit)
{
  int i;

  for (i = 0; NameCriterion[i].crit != Crit_LastMarker; i++)
    {
      if (NameCriterion[i].crit == singleCrit)
        return NameCriterion[i].critProc;
    }
  
  return NULL;
}


static const char *
get_heuristic (HeuristicsEnumType heuristic)
{
  int i;

  for (i = 0; heuristics_parse[i].heuristic != Heu_LastMarker; i++)
    {
      if (heuristics_parse[i].heuristic == heuristic)
        return heuristics_parse[i].name;
    }
  return "";
}



#if !COMP_POWER  
static void
print_options (CriterionType **Crit, HeuristicsEnumType heuristic,
               ContextType Cont[])
{
  int i;

  if ((Crit && *Crit) || Cont)
    fprintf (IO_Strom, "  Criteria:\n");
  if (Crit && *Crit)
    {
      for (i = 0; (*Crit)[i].criterion != Crit_LastMarker; i++)
        {
          fprintf (IO_Strom, "    %s: %s\n", get_criterion ((*Crit)[i].criterion),
                   get_action ((*Crit)[i].action));
        }
    }
  if (Cont)
    {
      for (i = 0; i < CPHistory_Last; i++)
        {
          if (Cont[i].action != Act_normal)
            fprintf (IO_Strom, "    %s: %s\n", NameContext[i].name, get_action (Cont[i].action));
        }
    }
  fprintf (IO_Strom, "  Heuristic: %s\n", get_heuristic (heuristic));
}
#endif

static
BOOLEAN parse_cpgoal_param (const char *str, WeightT *cigic_single_match,
                            WeightT *cigic_no_match, WeightT *cigic_min_struct)
{
   char *kopie, *kopieDummy, *token;

   kopie = (char *) our_alloc (sizeof (char) * strlen (str));
   strcpy (kopie, str);
   kopieDummy = kopie;
   
   /* wir betrachten den String bis zum naechsten : bzw. Stringende */
   token = strstr (kopie, ":");
   if (token)
      *token = '\0';
   
   token = strtok (kopie, ".");
   if (token) {
      *cigic_single_match = atol (token);
      token = strtok (NULL, ".");
      if (token) {
         *cigic_no_match = atol (token);
         token = strtok (NULL, ".");
         if (token) {
            *cigic_min_struct = atol (token);
         }
      }
   }
   our_dealloc (kopieDummy);

   return TRUE;
}

static
BOOLEAN parse_heuristic (const char *str, HeuristicsEnumType *heuristic)
{
  int i;
  
  for (i = 0; heuristics_parse[i].heuristic != Heu_LastMarker; i++)
    {
      if (strncmp (str, heuristics_parse[i].name, strlen (heuristics_parse[i].name)) == 0)
        {
          *heuristic = heuristics_parse[i].heuristic;
          return TRUE;
        }
    }

  return FALSE;
}


static BOOLEAN
add_action (CriteriaEnumType singleCrit, const char *str, CriterionType **Crit)
{
  int i, j;
  BOOLEAN found = FALSE;

  if (Crit == NULL)
    our_error ("criteria are only allowed for critical pair (re)classification.");
  
  
  for (i = 0; action_parse [i].action != Act_LastMarker && !found; i++)
    {
      if (strncasecmp (str, action_parse[i].name, strlen (action_parse[i].name)) == 0)
        {
          found = TRUE;
          break;
        }
    }
  
  if (!found)
    return FALSE;
  
  if (*Crit  == NULL)
    {
      *Crit = our_alloc (sizeof (CriterionType) * 2);
      (*Crit)[0].criterion = singleCrit;
      (*Crit)[0].action = action_parse[i].action;
      (*Crit)[0].critProc = get_crit_proc (singleCrit);
      (*Crit)[1].criterion = Crit_LastMarker;
    }
  else {
    /* Kriterium hinten anhaengen bzw. alten Wert ueberschreiben */
    for (j = 0; (*Crit)[j].criterion != Crit_LastMarker; j++)
      {
        if ((*Crit)[j].criterion == singleCrit)
          {
            (*Crit)[j].action = action_parse[i].action;
            return TRUE;
          }
      }
    *Crit = our_realloc (*Crit, sizeof (CriterionType) * (j + 2), "Parsing criteria");
    (*Crit)[j].criterion = singleCrit;
    (*Crit)[j].fullfilled = 0;
    (*Crit)[j].action = action_parse[i].action;
    (*Crit)[j].critProc = get_crit_proc (singleCrit);
    (*Crit)[j+1].criterion = Crit_LastMarker;
  }
  
  return TRUE;
}


static BOOLEAN
add_action_context (CPHistoryT context, const char *str, ContextType Cont[])
{
  int i;
  BOOLEAN found = FALSE;

  if (Cont == NULL)
    our_error ("History only allowed for critical pair classification.");
  
  for (i = 0; action_parse [i].action != Act_LastMarker && !found; i++)
    {
      if (strncasecmp (str, action_parse[i].name, strlen (action_parse[i].name)) == 0)
        {
          found = TRUE;
          break;
        }
    }
  
  if (!found)
    return FALSE;

  Cont[context].action = action_parse[i].action;
  
  return TRUE;
}


static BOOLEAN
parse_str (const char *options, CriterionType **Crit, HeuristicsEnumType *heuristic,
           ContextType Cont[],
           char **glam_file,
           WeightT *cigic_single_match, WeightT *cigic_no_match, WeightT *cigic_min_struct,
	   int *CPpatchVariante, int *CGpatchVariante)
{  char *kopie, *kopieDummy, *token;
  BOOLEAN ok = TRUE;

  kopie = (char *) our_alloc(sizeof(char) * strlen(options));

  strcpy (kopie, options);

  kopieDummy = kopie;
  token = strtok (kopie, ":");
  while (token != NULL && ok) {
    if (strncasecmp (token, "database=", 9) == 0)
      {
        ok = ok && add_action_context (ChaseManhattanCP, &token[9], Cont);
      }
    else if (strncasecmp (token, "echild=", 7) == 0)
      {
        ok = ok && add_action (Crit_EChild, &token[7], Crit);
      }
    else if (strncasecmp (token, "glamfile=", 9) == 0)
      {
        if (*glam_file)
          free (*glam_file);
        *glam_file = strdup (&token[9]);
      }
    else if (strncasecmp (token, "heuristic=", 10) == 0)
      {
        ok = ok && parse_heuristic (&token[10], heuristic);
      }
    else if (strncasecmp (token, "initial=", 8) == 0)
      {
        ok = ok && add_action_context (Initial, &token[8], Cont);
      }
    else if (strncasecmp (token, "killedRE=", 9) == 0)
      {
        ok = ok && add_action_context (KilledRE, &token[9], Cont);
      }
    else if (strncasecmp (token, "killerE=", 8) == 0)
      {
        ok = ok && add_action (Crit_KillerE, &token[8], Crit);
      }
    else if (strncasecmp (token, "killerRE=", 9) == 0)
      {
        ok = ok && add_action (Crit_KillerRE, &token[9], Crit);
      }
    else if (strncasecmp (token, "killerR=", 7) == 0)
      {
        ok = ok && add_action (Crit_KillerR, &token[7], Crit);
      }
    else if (strncasecmp (token, "twreceived=", 12) == 0)
      {
        ok = ok && add_action_context (TWreceived, &token[12], Cont);
      }
    else if (strncasecmp (token, "cpgoal=", 7) == 0)
       {
          ok = ok && parse_cpgoal_param (&token[7], cigic_single_match, cigic_no_match, cigic_min_struct);
       }
    else if (strncasecmp (token, "cppatchno=", 10) == 0)
       {
	 *CPpatchVariante = atoi (&token[10]);
       }
    else if (strncasecmp (token, "cgpatchno=", 10) == 0)
       {
	 *CGpatchVariante = atoi (&token[10]);
       }
    else
      {
        /* Funktionssymbol, wird ignoriert und an SymbolGewichte weitergereicht */
        /* statistics, statint wird an CStatistics weitergereicht */
      }
    
    token=strtok (NULL, ":");
  }

  if (!ok)
    {
      our_error ("Unknown Options for classification specified.");
    }
  
  our_dealloc (kopieDummy);

  return ok;
}



static
BOOLEAN parse_options (InfoStringT strDEF, InfoStringT strUSR,
                       CriterionType **Crit, ContextType Cont[],
                       HeuristicsEnumType defaultHeuristic,
                       HeuristicsEnumType *chosenHeuristic,
                       ClassifyProcT *ClassProc,
                       WeightT *cigic_single_match, WeightT *cigic_no_match, WeightT *cigic_min_struct,
		       int *CPpatchVariante, int *CGpatchVariante)
{
  BOOLEAN ok;
  HeuristicsEnumType heuristic;
  char *glam_file = NULL;
  
  heuristic = defaultHeuristic;
  ok = parse_str (strDEF, Crit, &heuristic, Cont, &glam_file,
		  cigic_single_match, cigic_no_match, cigic_min_struct,
		  CPpatchVariante, CGpatchVariante);

  ok = ok && parse_str (strUSR, Crit, &heuristic, Cont, &glam_file,
                        cigic_single_match, cigic_no_match, cigic_min_struct,
			CPpatchVariante, CGpatchVariante);

  *chosenHeuristic = heuristic;
  *ClassProc = heuristics_parse[heuristic].classProc;
  if (*ClassProc == NULL)
    our_error ("Heuristic not implemented yet");
  
#if !COMP_POWER  
  print_options (Crit, heuristic, Cont);
#endif
  
  return ok;
}


/* whichModule: 1 - CLAS
                2 - RECLAS
                3 - CGCLAS
*/
void C_ParamInfo(ParamInfoArtT what, int whichModule, 
                 InfoStringT clasDEF, InfoStringT clasUSR, 
                 InfoStringT reclasDEF, InfoStringT reclasUSR, 
                 InfoStringT cgDEF, InfoStringT cgUSR)
{
  CriterionType *Crit;
  ContextType Cont[CPHistory_Last];
  ClassifyProcT ClassProc;
  int i;
  WeightT cigic_single_match, cigic_no_match, cigic_min_struct;
  HeuristicsEnumType heuristic;
  int CPpatchVariante, CGpatchVariante;
  
  switch (what)
    {
    case ParamEvaluate:
      for (i = 0; i < CPHistory_Last; i++) {
        Cont[i].context = HistorySpec [i].context;
        Cont[i].action = Act_normal;
      }
  
      Crit = NULL;
      switch (whichModule)
        {
        case 1:
          fprintf(IO_Strom, "\n* Critical Pairs\n");
          parse_options (clasDEF, clasUSR, &Crit, Cont, Heu_MixWeight, &heuristic, &ClassProc,
                         &cigic_single_match, &cigic_no_match, &cigic_min_struct,
			 &CPpatchVariante, &CGpatchVariante);
          SG_ParamInfo(ParamEvaluate, clasDEF, clasUSR, TRUE);
          break;
        case 2:
          /*  CP-Re-Klassifikation */
          fprintf(IO_Strom, "\n* Critical Pairs Reclassification\n");
          parse_options (reclasDEF, reclasUSR, &Crit, NULL, Heu_MixWeight, &heuristic, &ClassProc,
                         &cigic_single_match, &cigic_no_match, &cigic_min_struct,
			 &CPpatchVariante, &CGpatchVariante);
          break;
        case 3:
          /* CG-Klassifikation */
          fprintf(IO_Strom, "\n* Critical Goals\n");
          parse_options (cgDEF, cgUSR, NULL, NULL, Heu_Unifikationsmass, &heuristic, &ClassProc,
                         &cigic_single_match, &cigic_no_match, &cigic_min_struct,
			 &CPpatchVariante, &CGpatchVariante);
          SG_ParamInfo(ParamEvaluate, cgDEF, cgUSR, TRUE);
          break;
        default:
          our_error ("only values 1-3 are allowed in C_ParamInfo");
          break; 
        }
      if (Crit != NULL)
        our_dealloc (Crit);
      break;
    case ParamExplain:
      switch (whichModule)
        {
        case 1:
          fprintf(IO_Strom, "Classification\n");
          fprintf(IO_Strom, "**************\n");
          fprintf(IO_Strom, "This parameter guides the assessment of critical pairs,\n");
          fprintf(IO_Strom, "i.e. the computation of one or more integers (w1/w2) that make up an ordering\n");
          fprintf(IO_Strom, "of the stored cps. This computation of basic weights can be influenced\n");
          fprintf(IO_Strom, "by assigning different weights to single symbols, or by some criteria.\n\n");
          break;
        case 2:
          fprintf(IO_Strom, "Re-Classification\n");
          fprintf(IO_Strom, "*****************\n");
          fprintf(IO_Strom, "This parameter guides the criteria to apply on critical pairs during the\n");
          fprintf(IO_Strom, "intermediate reprocessing of the CP-Set. The basic classifying weight is\n");
          fprintf(IO_Strom, "the same as given by the -clas option.\n\n");
          break;
        case 3:
          fprintf(IO_Strom, "CG-Classification\n");
          fprintf(IO_Strom, "*****************\n");
          fprintf(IO_Strom, "This parameter guides the classification of critical goals, and is similar to\n");
          fprintf(IO_Strom, "the -clas option. For critical goals, no criteria are available.\n\n");
          break;
        }
      if (whichModule < 3){
         fprintf(IO_Strom, "Criteria that can be applied to cps detect certain properties. If such a property\n");
         fprintf(IO_Strom, "has been detected, the cp will be prefered resp. defered as stated in <treatment>.\n");
         fprintf(IO_Strom, "\n");
         fprintf(IO_Strom, "<criterion>=<treatment>\n");
         fprintf(IO_Strom, "\n");
         fprintf(IO_Strom, "<criterion>=killerR|killerE|killerRE|echild|initial|killedRE|database\n");
         fprintf(IO_Strom, "<treatment>=ultimate|half|double|last\n");
         fprintf(IO_Strom, "\n");
         fprintf(IO_Strom, "killerR/E/RE  detects those cps that would remove a rule/equation via interreduction\n");
         fprintf(IO_Strom, "echild        detects cps a parent of which is an equation\n");
         fprintf(IO_Strom, "initial       detects cps from the initial specification\n");
         fprintf(IO_Strom, "killedRE      detects cps that arised from interreduction victims\n\n");
      }
      fprintf (IO_Strom,
               "Heuristics can be specified with:\n"
               "heuristic=addweight|cpingoal|glam|goalincp|gtweight|mixweight|mixweight2|maxweight|ordweight|occnest|\n"
               "          unif|fifo|golem|cppatch|cgpatch\n");
      fprintf(IO_Strom, "  Please note that heuristics cpingoal/goalincp and occnest only work on critical pairs.\n");
      fprintf(IO_Strom, "  The heuristics cpingoal/goalincp are parametrised by:\n");
      fprintf(IO_Strom, "    cpgoal=single-match.no-match.min-struct\n");
      fprintf(IO_Strom, "  The heuristics cppatch/cgpatch are experimental and are parametrised by:\n");
      fprintf(IO_Strom, "    cppatchno=X:cgpatchno=Y\n");
      
      switch (whichModule)
        {
        case 1:
          SG_ParamInfo(ParamExplain, clasDEF, clasUSR, TRUE);
        case 3:
          SG_ParamInfo(ParamExplain, cgDEF, cgUSR, TRUE);
          break;
        default:
          break;
          
        }
      break;
    }
}


BOOLEAN C_InitAposteriori(InfoStringT clasDEF, InfoStringT clasUSR, 
                          InfoStringT reclasDEF, InfoStringT reclasUSR, 
                          InfoStringT cgDEF, InfoStringT cgUSR)
{
  WeightT cigic_single_match, cigic_no_match, cigic_min_struct;
  HeuristicsEnumType ClasHeuristic, ReClasHeuristic, CGHeuristic;
  int CGpatchVariante = 0;
  int CPpatchVariante = 0;

  CPNr = 0;
  cigic_single_match = 0;
  cigic_no_match = 0;
  cigic_min_struct = 0;
  

#if !COMP_POWER  
   fprintf(IO_Strom, "\n* CLASSIFICATION\n");
#endif   
   
   SG_SymbGewichteEintragen (cgUSR, cgUSR, FALSE);
   SG_SymbGewichteEintragen (clasDEF, clasUSR, TRUE);
   CS_InitAposteriori(clasDEF, clasUSR);

   /* normale CP-Klassifikation */
#if !COMP_POWER  
   fprintf(IO_Strom, "\n* Critical Pairs\n");
#endif

   ClasCriteria = NULL;
   parse_options (clasDEF, clasUSR, &ClasCriteria, HistorySpec, Heu_MixWeight, &ClasHeuristic, &ClassifyProc,
                  &cigic_single_match, &cigic_no_match, &cigic_min_struct,
		  &CPpatchVariante, &CGpatchVariante);

   switch (ClasHeuristic)
     {
     case Heu_CP_In_Goal:
     case Heu_Goal_In_CP:
       CIGICInit (cigic_single_match, cigic_no_match, cigic_min_struct);
       break;
      case Heu_OccNest:
        ON_Init ();
       break;
      case Heu_CPPatch:
	CPA_CPInit (CPpatchVariante);
	break;
     case Heu_CGPatch:
	our_error ("This heuristic doesn't work on critical pairs!");
	break;
     case Heu_PolyWeight:
        CF_PolyWeightInitialisieren();
        break;
     default:
       break;
     }
   
   /*  CP-Re-Klassifikation */
#if !COMP_POWER  
   fprintf(IO_Strom, "\n* Critical Pairs Reclassification\n");
#endif
   ReClasCriteria = NULL;
   parse_options (reclasDEF, reclasUSR, &ReClasCriteria, NULL, ClasHeuristic, &ReClasHeuristic, &ReClassifyProc,
                  &cigic_single_match, &cigic_no_match, &cigic_min_struct,
		  &CPpatchVariante, &CGpatchVariante);

   if (ClasHeuristic != ReClasHeuristic)
      our_error ("Classification and Reclassification must use the same heuristics!");
   
   /* CG-Klassifikation */
#if !COMP_POWER  
   fprintf(IO_Strom, "\n* Critical Goals\n");
#endif
   parse_options (cgDEF, cgUSR, NULL, NULL, Heu_Unifikationsmass, &CGHeuristic, &CGClassifyProc,
                  &cigic_single_match, &cigic_no_match, &cigic_min_struct,
		  &CPpatchVariante, &CGpatchVariante);

   switch (CGHeuristic)
      {
      case Heu_OccNest:
      case Heu_CP_In_Goal:
      case Heu_Goal_In_CP:
      case Heu_CPPatch:
         our_error ("This heuristic doesn't work on critical goals!");
         break;
      case Heu_CGPatch:
	CPA_CGInit (CGpatchVariante);
	break;
      case Heu_PolyWeight:
        CF_PolyWeightInitialisieren();
        break;
      default:
         break;
      }
   
   /* Weight-Statistik */
   C_WeightStatProc = CS_WeightStatProc;
   C_WeightStatisticsStartProc = CS_WeightStatisticsStart;
   C_WeightStatisticsEndProc = CS_WeightStatisticsEnd;
   
   return TRUE;
}
