/**********************************************************************
* Filename : SelectClassify.c
*
* Kurzbeschreibung : Behandlung von Regel/Gleichungspaaren, die
*                    MNF bei Select erzeugt
*
* Autoren : Andreas Jaeger
*
* 05.09.96: Erstellung
* 22.02.97: Mehrfachaufrufclean. AJ.
* 23.02.97: komplette Ueberarbeitung des Moduls: Subsumption/TO_ACnGleich wird
*           richtig behandelt, es werden keine doppelten Gleichungen/Regeln
*           mehr aufgenommen, Subsumption der Gleichungen untereinander,
*           bessere Strukturierung, Parameter geaendert,...
*
*
* In der Datei SelectClassify.tex sind einige Ueberlegungen zu den
* Aufgaben dieses Moduls dargelegt.
*
* Benoetigt folgende Module:
*  - Ausgaben: Ausgeben von Termen (wird beim Testen benutzt)
*  - Hauptkomponenten: Ordnungsvergleich
*  - DSBaumOperationen: Normieren von Termpaaren
*
* Benutzte Globale Variablen aus anderen Modulen:
* ??
*
*
 Todo: Strategien einbauen (SC_ClassifyEquations), Statistik einbauen

   Fuer Optimierungen (Subsumption) sollte noch ein Test auf
   Monogleichungen rein. Aber wie sieht der Test aus? - und
   lohnt sich das ueberhaupt? bzw. Muessen wir die Gegenrichtung einer Gleichung
   normiert rein tun?
   
**********************************************************************/


#include <limits.h>

#include "general.h"
#include "Ausgaben.h"
#include "DSBaumOperationen.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "MNF_CPs.h"
#include "SelectClassify.h"
#include "SpeicherVerwaltung.h"
#include "Subsumption.h"
#include "SUEM.h"
#include "SymbolGewichte.h"
#include "TermOperationen.h"
#include "Ordnungen.h"


/*
  zum Testen des Moduls:
  SC_TEST auf 1 oder 2 setzen
  */
#define SC_TEST 0

#define SC_STATISTIK normal_power_comp(1,0,0)

#if SC_STATISTIK 
/* Statistikvariablen */
static CounterT SC_MaxNF,
   SC_NoCalls,
   SC_NoRulesGenerated,
   SC_NoEquGenerated,
   SC_NoNFSinglePair,
   SC_NoCallsGeneratePairs,
   SC_NoSubsumed,
   SC_NoSubsumedClassifyEquations,
   SC_NoDoppelte,
   SC_NoSubsummiertUntereinander,
   SC_ACGleich;
#endif /* SC_STATISTIK */

#if SC_STATISTIK 
# define SC_IncNoCalls             (SC_NoCalls++)
# define SC_IncNoRulesGenerated    (SC_NoRulesGenerated++)
# define SC_IncNoEquGenerated      (SC_NoEquGenerated++)
# define SC_IncNoNFSinglePair      (SC_NoNFSinglePair++)
# define SC_IncNoSubsumed          (SC_NoSubsumed++)
# define SC_IncCallsGeneratePairs  (SC_NoCallsGeneratePairs++)
# define SC_IncNoSubsumedClassifyEquations  (SC_NoSubsumedClassifyEquations++)
# define SC_IncNoDoppelte          (SC_NoDoppelte++)
# define SC_IncNoSubsummiertUntereinander (SC_NoSubsummiertUntereinander++)
# define SC_IncACGleich            (SC_ACGleich++)
#else
# define SC_IncNoCalls
# define SC_IncNoRulesGenerated
# define SC_IncNoEquGenerated
# define SC_IncNoNFSinglePair
# define SC_IncNoSubsumed
# define SC_IncCallsGeneratePairs
# define SC_IncNoSubsumedClassifyEquations
# define SC_IncNoDoppelte
# define SC_IncNoSubsummiertUntereinander
# define SC_IncACGleich
#endif /* SC_STATISTIK */


typedef struct SC_TermTS 
{
   int                      anzGroesserL;
   int                      anzGroesserR;
   int                      anzUnvglL;
   int                      anzUnvglR;
   BOOLEAN                  usedAsLeftSide;
   BOOLEAN                  usedAsRightSide;
   TermT                    Term;
   VergleichsT              *VergleichsVektor;
} SC_TermT;

typedef SC_TermT            *SC_CTermArrayT;

/* diverse Variablen */
static SV_SpeicherManagerT SC_PairsSpeicher;

static BOOLEAN SC_Initialised = FALSE;

/* Variablen fuer Heuristiken */
/* diese Variablen sind alle ueber Optionen setzbar */
static unsigned short SC_PermGrad_E = 0;
static unsigned short SC_PermGrad_R = 0;
static BOOLEAN SC_Subsumption = TRUE;
static BOOLEAN useK = TRUE;
static unsigned int SC_MaxAddRules;
static unsigned int SC_MaxAddEquations;

/* Funktionsprototypen */
static int SC_ClassifyPair (TermT left, TermT right);


/********************************************************************************/
/*                                                                              */
/*  Operationen auf der Paarstruktur                                            */
/*                                                                              */
/********************************************************************************/

static void SC_AddPair (SC_REPairsT **REPairs, TermT left, TermT right, VergleichsT vergleich,
                         BOOLEAN VerbindetLR, BOOLEAN keep)
{
   SC_REPairsPtrT         pairPtr;

#if SC_TEST
   VergleichsT            vgl = ORD_VergleicheTerme (left, right);

   if (vgl != vergleich) {
      our_error ("SC_AddPair: Vergleich ist falsch!");
   }
#endif
   
   SV_Alloc(pairPtr,SC_PairsSpeicher,SC_REPairsPtrT);

   pairPtr->next = *REPairs;
   pairPtr->left = TO_Termkopie (left);
   pairPtr->right= TO_Termkopie (right);
   BO_TermpaarNormierenAlsKP(pairPtr->left, pairPtr->right);

   pairPtr->keep=keep;
   pairPtr->weightP=SC_ClassifyPair (left, right);
   pairPtr->vergleich=vergleich;
   pairPtr->VerbindetLR=VerbindetLR;
   *REPairs=pairPtr;
}




/* Loescht alle Elemente, fuer die keep=FALSE ist */
static void SC_DeletePairsNoKeep (SC_REPairsT **REPairs) {
   SC_REPairsPtrT    tmpPtr;
   SC_REPairsPtrT    pairPtr = *REPairs;
   SC_REPairsPtrT    newPairPtr=NULL;
   
   while (pairPtr != NULL) {
      if (pairPtr->keep) {
         /* an neue Liste dranhaengen */
         tmpPtr  = pairPtr;
         pairPtr = pairPtr->next;
         tmpPtr->next = newPairPtr;
         newPairPtr = tmpPtr;
      }
      else {
         TO_TermLoeschen (pairPtr->left);
         TO_TermLoeschen (pairPtr->right);
         tmpPtr  = pairPtr;
         pairPtr = pairPtr->next;
         SV_Dealloc (SC_PairsSpeicher, tmpPtr);
      }
   }
   *REPairs = newPairPtr;
}


/*
  Loeschen der gesamten Datenstruktur,
  Termzellen werden nicht geloescht
 */
void SC_DeletePairs (SC_REPairsT **REPairs) {
   SC_REPairsPtrT    tmpPtr;
   SC_REPairsPtrT    pairPtr = *REPairs;
   
   while (pairPtr != NULL) {
      tmpPtr=pairPtr;
      pairPtr=pairPtr->next;
      SV_Dealloc (SC_PairsSpeicher, tmpPtr);
   }
   *REPairs=NULL;
}


/* Loeschen der Gesamten Datenstruktur einschliesslich
   der belegten Termzellen */
void SC_DeletePairsTermLoeschen (SC_REPairsT **REPairs) {
   SC_REPairsPtrT    tmpPtr;
   SC_REPairsPtrT    pairPtr = *REPairs;
   
   while (pairPtr != NULL) {
      TO_TermLoeschen(pairPtr->left);
      TO_TermLoeschen(pairPtr->right);
      tmpPtr=pairPtr;
      pairPtr=pairPtr->next;
      SV_Dealloc(SC_PairsSpeicher, tmpPtr);
   }
   *REPairs=NULL;
}


/********************************************************************************/
/*                                                                              */
/*  Classifizieren von Paaren                                                   */
/*                                                                              */
/********************************************************************************/
      
/* Berechnen eines Gewichtes fuer dieses Termpaar,
   Funktionalitaet muss noch erweitert werden auf GLAM */
static int SC_ClassifyPair (TermT left, TermT right)
{
   SymbolT     symb;
   int         length_left = 0;
   int         length_right= 0;
   
   TO_doSymbole(left, symb,{
      length_left  += SG_SymbolGewichtCP(symb);
   });
   TO_doSymbole(right, symb,{
      length_right  += SG_SymbolGewichtCP(symb);
   });

   return length_left + length_right;
}



/* Klassifikation der Paare danach, ob RE-Menge reduziert wird. Wenn ja,
   dann wird das Gewicht minimal gesetzt */
static void SC_ClassifyI (SC_REPairsT **REPairs) 
{
   SC_REPairsPtrT         pairPtr;
   BOOLEAN                killerKrit = FALSE;
   
   pairPtr=*REPairs;

   /* testen auf Killerkriterium */
   while (pairPtr != NULL) {
      if (pairPtr->vergleich == Unvergleichbar) {
         killerKrit = IR_TPEReduziertRE (pairPtr->left, pairPtr->right);
         killerKrit = killerKrit || IR_TPEReduziertRE (pairPtr->right, pairPtr->left);
      } else if (pairPtr->vergleich == Kleiner)
         killerKrit=IR_TPRReduziertRE (pairPtr->right, pairPtr->left);
      else
         killerKrit=IR_TPRReduziertRE (pairPtr->left, pairPtr->right);
      if (killerKrit) {
#if SC_TEST >= 2
         fprintf(IO_Strom, "\n\aSC_ClassifyI: Killerkriterium schlaegt zu:\n\a");
         IO_DruckeFlex("%t ? %t\n", pairPtr->right, pairPtr->left);
#endif
         pairPtr->weightP = 1; /* minimales Gewicht geben */
      }
      pairPtr=pairPtr->next;
   }
}


/*
  holen der maxPairs besten Paare aus REPairs
  Es muessen mindestens maxPairs Paare vorhanden sein.
  getRules == TRUE besagt, dass nur Regeln untersucht werden
  getRules == FALSE besagt, dass nur Gleichungen untersucht werden
  */
static SC_REPairsT ** GetBestPairs (SC_REPairsT **REPairs, unsigned int maxPairs, BOOLEAN getRules) 
{
   SC_REPairsPtrT         pairPtr;
   SC_REPairsT            **REPairsArray;
   unsigned int           i, minIndex;
   int                    worst;



   if (maxPairs == 0)
      return NULL;
   
   /* in REPairsArray wird eine sortierte Liste der noPairs
      besten Element gehalten */
   /* es wird von hinten nach vorne gefuellt, so dass
      REPairsArray[noPairs] das Minimum enthaelt */
   REPairsArray = (SC_REPairsT **)our_alloc (sizeof(SC_REPairsPtrT) * maxPairs);

   pairPtr = *REPairs;

   minIndex = maxPairs;

   worst = INT_MAX;

   /* jetzt suchen der naechsten Elemente. Durch den Wert von worst (INT_MAX)
      kommen mal erst die ersten Elemente komplett rein - und dann wird jeweils
      das schlechteste rausgeschmissen und das neue einsortiert */
   while (pairPtr != NULL) {
      /* Wenn getRules TRUE ist, werden nur Regeln genommen,
         andernfalls nur Gleichungen */
      /* Im folgenden ist also
         getRules XOR (pairPtr->vergleich == Unvergleichbar) programmiert:
         */
      pairPtr->keep = FALSE;  /* mal erst alle auf FALSE um einen Durchlauf zu sparen */
      if (getRules == (pairPtr->vergleich != Unvergleichbar)) {
         if (pairPtr->weightP < worst) {
            /* Einsortieren in Array */
            if (minIndex > 0) {
               minIndex--;
               REPairsArray[minIndex] = pairPtr;
            }
            else {
               REPairsArray[0] = pairPtr;
            }
            i = minIndex + 1;
            while ((i < maxPairs) && (pairPtr->weightP < REPairsArray[i]->weightP)) {
               REPairsArray[i-1] = REPairsArray[i];
               i++;
            }
            REPairsArray[i-1] = pairPtr;
            if (minIndex == 0)
               worst = REPairsArray[0]->weightP;
         }
         pairPtr=pairPtr->next;
      }
   }
   return REPairsArray;
}


/* Durchgehen durch REPairs und behalten der besten Termpaare */
static void SC_CheckSize (SC_REPairsT **REPairs, unsigned int maxRules,
                          unsigned int maxEquations)
{
   SC_REPairsPtrT         pairPtr;
   BOOLEAN                verbindetLR = FALSE; /* aufpassen, dass ein Element drin bleibt, welches verbindet */
   unsigned int           countEquations = 0;
   unsigned int           countRules = 0;
   SC_REPairsT            **RulePairsArray;
   SC_REPairsT            **EquationPairsArray;
   SC_REPairsPtrT         VerbindetLRPairPtr = NULL;
   unsigned int           i;
   int                    bestWeight;
   BOOLEAN                VerbindetMussNochRein = FALSE;
   
   /* mal erst Testen, ob wir ueberhaupt was zu tun haben */
   pairPtr = *REPairs;
   while (pairPtr != NULL) {
      if (pairPtr->vergleich == Unvergleichbar)
         countEquations++;
      else
         countRules++;
      pairPtr=pairPtr->next;
   }

   if ((countEquations <= maxEquations) &&
       (countRules <= maxRules))
      return;

   /* Trick 17:
      Wenn wir weniger Regeln/Gleichungen haben,
      als wir zulassen - dann setzten wir einfach
      die Grenze auf die richtige Anzahl
      */
   if (countEquations < maxEquations)
      maxEquations = countEquations;
   if (countRules < maxRules)
      maxRules = countRules;
   
   /* in REPairsArray wird eine sortierte Liste der maxRules/Equations
      besten Element gehalten */
   /* es wird von hinten nach vorne gefuellt, so dass
      REPairsArray[maxRules/Equations] das Minimum enthaelt */
   RulePairsArray = GetBestPairs (REPairs, maxRules, TRUE);
   EquationPairsArray = GetBestPairs (REPairs, maxEquations, FALSE);

   /* testen, ob das KP verbunden wird */
   for (i = 0; i < maxRules; i++)
      verbindetLR = verbindetLR || RulePairsArray[i]->VerbindetLR;

   if (!verbindetLR) {
      for (i = 0; i < maxEquations; i++)
         verbindetLR = verbindetLR || EquationPairsArray[i]->VerbindetLR;
   }
   /* Wenn unter den noPairs besten keine Regel/Gleichung dabei ist,
      die das kritische Paar verbindet, dann wird die beste Regel/Gleichung,
      die das tut hinzugetan und das schlechteste bisher ausgewaehlte R/E
      rausgeschmissen */

   if (!verbindetLR) {
      pairPtr = *REPairs;
      bestWeight = INT_MAX;
      while (pairPtr != NULL) {
         if (pairPtr->VerbindetLR && (pairPtr->weightP < bestWeight)) {
            VerbindetLRPairPtr = pairPtr;
            bestWeight = VerbindetLRPairPtr->weightP;
         }
         pairPtr=pairPtr->next;
      }
      /* VerbindetLRPairPtr ist nun die beste Regel/Gleichung, die verbindet.
         Sie muss jetzt noch in die richtige Struktur eingetragen werden.
         Es muss aufgepasst werden, dass maxRules/Equations != 0 ist
         */
      if (VerbindetLRPairPtr->vergleich == Unvergleichbar) {
         if (maxEquations == 0)
            VerbindetMussNochRein = TRUE;
         else
            EquationPairsArray[0] = VerbindetLRPairPtr;
      } else {
         if (maxRules == 0)
            VerbindetMussNochRein = TRUE;
         else
            RulePairsArray[0] = VerbindetLRPairPtr;
      }
   }

   /* Spezialfall: maxRules == 0 und maxEquations == 0,
      dann wurde keep nicht auf FALSE gesetzt,
      deswegen wird jetzt keep hier auf FALSE gesetzt */
   if ((maxEquations == 0) && (maxRules == 0)) {
      pairPtr = *REPairs;
      while (pairPtr != NULL) {
         pairPtr->keep = FALSE;
         pairPtr=pairPtr->next;
      }
   }

   /* die besten Elemente bleiben erhalten (alle Elemente wurden auf FALSE gesetzt) */
   for (i=0; i < maxRules; i++)
      RulePairsArray[i]->keep = TRUE;
   for (i=0; i < maxEquations; i++)
      EquationPairsArray[i]->keep = TRUE;
   if (VerbindetMussNochRein)
      VerbindetLRPairPtr->keep = TRUE;
   
   /* der Rest wird weggeschmissen */
   SC_DeletePairsNoKeep (REPairs);
   /* aufraeumen */
   if (RulePairsArray != NULL)
      our_dealloc (RulePairsArray);
   if (EquationPairsArray != NULL)
      our_dealloc (EquationPairsArray);
}

/********************************************************************************/
/*                                                                              */
/*  Testen von Subsumption/ACGleichheit der Paare                               */
/*                                                                              */
/********************************************************************************/
  
static BOOLEAN SC_ACGleichRegeln (SC_REPairsT   **REPairs) {
   SC_REPairsPtrT     pairPtr;
   BOOLEAN            finished = FALSE;
   BOOLEAN            acGleich;
   
   pairPtr=*REPairs;

   if (!SC_PermGrad_R)
      return FALSE;

   do {
      if ((pairPtr->keep) &&
          ((pairPtr->vergleich == Kleiner) || (pairPtr->vergleich == Groesser))) {
         acGleich= TO_ACnGleich (SC_PermGrad_R, pairPtr->left, pairPtr->right);
         if (acGleich) {
            SC_IncACGleich;
            if (pairPtr->VerbindetLR) {
               SUE_selectedCPMNFpermsubsumed ();
               finished = TRUE;
               HK_fair = FALSE;
            } else {
               pairPtr->keep = FALSE;
            }
         }
      }
      pairPtr=pairPtr->next;
   } while (!finished && (pairPtr != NULL));

   return finished;
}



static BOOLEAN SC_ACGleichGleichungen (SC_REPairsT   **REPairs) {
   SC_REPairsPtrT     pairPtr;
   BOOLEAN            finished = FALSE;
   BOOLEAN            acGleich;
   
   pairPtr=*REPairs;

   if (!SC_PermGrad_E)
      return FALSE;

   do {
      if (pairPtr->keep &&
          (pairPtr->vergleich == Unvergleichbar)) {
         acGleich= TO_ACnGleich (SC_PermGrad_E, pairPtr->left, pairPtr->right);
         if (acGleich) {
            SC_IncACGleich;
            if (pairPtr->VerbindetLR) {
               HK_fair = FALSE;
               SUE_selectedCPMNFpermsubsumed ();
               finished = TRUE;
            } else {
               pairPtr->keep = FALSE;
            }
         }
      }
      pairPtr=pairPtr->next;
   } while (!finished && (pairPtr != NULL));
   
   return finished;
}


static BOOLEAN SC_SubsumptionGleichungsmenge (SC_REPairsT **REPairs) {
   SC_REPairsPtrT     pairPtr;
   BOOLEAN            finished = FALSE;
   BOOLEAN            subsummiert;

   if (!SC_Subsumption)
      return FALSE;
   
   pairPtr=*REPairs;
   
   do {
      if ((pairPtr->vergleich == Unvergleichbar) &&
          pairPtr->keep) {
         subsummiert = SS_TermpaarSubsummiertVonGM (pairPtr->left, pairPtr->right);
         if (subsummiert) {
            SC_IncNoSubsumed;
            if (pairPtr->VerbindetLR) {
               SUE_selectedCPMNFsubsumed ();
               finished = TRUE;
            } else {
               pairPtr->keep = FALSE;
            }
         }
      }
      pairPtr=pairPtr->next;
   } while (!finished && (pairPtr != NULL));

   return finished;
}

static BOOLEAN SC_SubsumptionUntereinander (SC_REPairsT **REPairs) {
   SC_REPairsPtrT     pairPtr, pair2Ptr;
   BOOLEAN            subsummiert;
   
   /* Achtung: im folgenden kann es passieren, dass ein Gleichung, welche nicht links & rechts
      verbindet, Gleichungen loescht, welche links&rechts verbinden.
      Da diese Gleichung dann allgemeiner ist als die geloeschte und somit
      sicherlich auch l&r verbindet, bekommt sie einfach das verbindet-Flag
      */

   if (!SC_Subsumption)
      return FALSE;
   
   pairPtr = *REPairs;
   do {
      pair2Ptr = *REPairs;
      while (pair2Ptr != pairPtr) {
         if ((pairPtr->vergleich == Unvergleichbar)
             && (pairPtr->vergleich == Unvergleichbar)
             && pairPtr->keep
             && pair2Ptr->keep) {
            subsummiert = SS_TermpaarSubsummiertTermpaar (pairPtr->left, pairPtr->right,
                                                          pair2Ptr->left, pair2Ptr->right);
            if (subsummiert) {
               pair2Ptr->keep = FALSE;
               SC_IncNoSubsummiertUntereinander;
               if (pair2Ptr->VerbindetLR)
                  pairPtr->VerbindetLR = TRUE;
            }
         }
         pair2Ptr = pair2Ptr->next;
      }
      pairPtr = pairPtr->next;
   } while (pairPtr != NULL);
   return FALSE;
}


/* Da alle Paare jetzt normiert sind, kann es sein, dass Paare identisch sind.
   Dies wird im folgenden ueberprueft */
static void SC_SucheDoppelte (SC_REPairsT **REPairs) {
   SC_REPairsPtrT     pairPtr, pair2Ptr;


   /* Achtung: im folgenden kann es passieren, dass ein Paar, welches nicht links & rechts
      verbindet, einem Paar gleich, welche links&rechts verbinden.
      Da diese Paare gleich sind, muss das nicht geloeschte das Kennzeichen verbindetLR
      erhalten.
      */

   pairPtr = *REPairs;
   do {
      pair2Ptr = *REPairs;
      while (pair2Ptr != pairPtr) {
         if (pairPtr->keep && pair2Ptr->keep) {
            if (pairPtr->vergleich == pair2Ptr->vergleich) {
               if (TO_TermeGleich (pairPtr->left, pair2Ptr->left)
                   && TO_TermeGleich (pairPtr->right, pair2Ptr->right)) {
#if SC_TEST >= 2
                  fprintf(IO_Strom, "\n\aSC_SucheDoppelte: \n\a");
                  IO_DruckeFlex("%t ? %t\n", pairPtr->right, pairPtr->left);
                  IO_DruckeFlex("%t ? %t\n", pair2Ptr->right, pair2Ptr->left);
#endif
                  pair2Ptr->keep = FALSE;
                  SC_IncNoDoppelte;
                  if (pair2Ptr->VerbindetLR)
                     pairPtr->VerbindetLR = TRUE;
               }
            } else if (((pairPtr->vergleich == Groesser)
                        && (pair2Ptr->vergleich == Kleiner))
                       || ((pairPtr->vergleich == Kleiner)
                           && (pair2Ptr->vergleich == Groesser)))
               {
                  if (TO_TermeGleich (pairPtr->left, pair2Ptr->left)
                      && TO_TermeGleich (pairPtr->right, pair2Ptr->right)) {
#if SC_TEST >= 2
                     fprintf(IO_Strom, "\n\aSC_SucheDoppelte: \n\a");
                     IO_DruckeFlex("%t ? %t\n", pairPtr->right, pairPtr->left);
                     IO_DruckeFlex("%t ? %t\n", pair2Ptr->right, pair2Ptr->left);
#endif
                     pair2Ptr->keep = FALSE;
                     SC_IncNoDoppelte;
                     if (pair2Ptr->VerbindetLR)
                        pairPtr->VerbindetLR = TRUE;
                  }
               }
         }
         pair2Ptr = pair2Ptr->next;
      }
      pairPtr = pairPtr->next;
   } while (pairPtr != NULL);
}


/* Durchgehen durch alle Elemente und dabei Testen:
   - ob ein Element von der aktuellen Regel/Gleichungsmenge subsummiert wird
   - ob ein Element von einem anderen Element subsummiert wird.
   - ob Termpaare ACnGleich sind
   
   Die Funktion gibt TRUE zurueck, wenn ein Element subsummiert wird/ACnGleich ist,
   welches L und R verbindet.
   Es werden alle subsummierten Elemente entfernt
   Parameter:
   REPairs: Menge von Regeln/Gleichungen zu testen
   onlyRules: Flag. Wenn TRUE dann enthaelt REPairs nur Regeln und keine einzige Gleichung,
                    wenn false ist nichts bekannt.
*/

static BOOLEAN SC_SubsumptionTest (SC_REPairsT   **REPairs, BOOLEAN onlyRules) {
   BOOLEAN            finished = FALSE;

   /* TODO: Reihenfolge evtl. umstellen */
   SC_SucheDoppelte (REPairs);
   if (onlyRules)
      finished = SC_ACGleichRegeln (REPairs);
   else {
      finished = SC_SubsumptionGleichungsmenge (REPairs);
      
      finished = finished || SC_ACGleichRegeln (REPairs);
      finished = finished || SC_ACGleichGleichungen (REPairs);
      finished = finished || SC_SubsumptionUntereinander (REPairs);
   }
   
   if (finished) {
      SC_DeletePairsTermLoeschen (REPairs); /* NULL-Liste zurueck */
#if IO_AUSGABEN
      if ((IO_level3) || (IO_level4))
	 fprintf(IO_Strom, "... subsumed.\n");
#endif
      SC_IncNoSubsumedClassifyEquations;      
   } else
      SC_DeletePairsNoKeep (REPairs);
   
   return finished;
}


static BOOLEAN SC_SubsummiertEinPaar (TermT links, TermT rechts, VergleichsT vgl) 
{
   BOOLEAN ret = FALSE;
   
   if (vgl == Unvergleichbar) {
      if (SC_PermGrad_E) {
         ret = TO_ACnGleich (SC_PermGrad_E, links, rechts);
         if (ret) {
           HK_fair = FALSE;
            SC_IncACGleich;
            SUE_selectedCPMNFpermsubsumed ();
         }
      }
      if (SC_Subsumption) {
         if (!ret && SS_TermpaarSubsummiertVonGM (links, rechts)) {
            SC_IncNoSubsumed;
            SUE_selectedCPMNFsubsumed ();
            ret = TRUE;
         }
      }
   } else { /* Kleiner/Groesser */
      if (SC_PermGrad_R) {
         ret = TO_ACnGleich (SC_PermGrad_R, links, rechts);
         if (ret) {
            HK_fair = FALSE;
            SC_IncACGleich;
            SUE_selectedCPMNFpermsubsumed ();
         }
      }
   }
   
   return ret;
}

/********************************************************************************/
/*                                                                              */
/*  Hauptarbeit: Erzeugen von Paaren                                            */
/*                                                                              */
/********************************************************************************/


/* Erzeugen der Paare */
static void SC_GeneratePairs (MNF_NFSetT *NFLeft, MNF_NFSetT *NFRight,
                                 int countLeft, int countRight,
                                 SC_REPairsT **REPairs)
{
   int                  i, j;
   int                  countTotal = countLeft + countRight;
   MNF_NFSetPtrT        setPtr;
   SC_CTermArrayT       CTermArray;
   BOOLEAN              madeAllPairs = FALSE;
   BOOLEAN              verbindetLR;
   BOOLEAN              onlyRules = FALSE;
   BOOLEAN              subsummiert;
   int                  noTotalMin = -1;
   int                  Minimum = -1;
   int                  MinAnzGroesser = 0;
#if SC_TEST >= 2
   SC_REPairsPtrT       pairPtr;
#endif
   
   SC_IncCallsGeneratePairs;

   CTermArray = (SC_CTermArrayT) our_alloc(sizeof(SC_TermT) * countTotal);
   
   for (i=0; i < countTotal; i++) {
      CTermArray[i].VergleichsVektor = (VergleichsT *)our_alloc (sizeof(VergleichsT) * countTotal);
      CTermArray[i].anzGroesserR = 0;
      CTermArray[i].anzGroesserL = 0;
      CTermArray[i].anzUnvglL = 0;
      CTermArray[i].anzUnvglR = 0;
      CTermArray[i].usedAsLeftSide = FALSE;
      CTermArray[i].usedAsRightSide = FALSE;
      
   }

   setPtr=NFLeft;
#if SC_TEST >= 2
   fprintf(IO_Strom, "\nSC_GeneratePairs:\n");
#endif
   for (i=0; i < countLeft; i++, setPtr=setPtr->next) {
      CTermArray[i].Term = setPtr->Term;
#if SC_TEST >= 2
      fprintf(IO_Strom, "Left  %d: ", i );
      IO_DruckeFlex("%t\n", setPtr->Term);
#endif      
   }
   

   setPtr=NFRight;
   for (i = countLeft; i < countTotal; i++, setPtr=setPtr->next) {
      CTermArray[i].Term = setPtr->Term;
#if SC_TEST >= 2
      fprintf(IO_Strom, "Right %d: ",i );
      IO_DruckeFlex("%t\n", setPtr->Term);
#endif      
   }


   /* dies ist optimiert: es wird im Dreieck durch die Matrix gegangen und somit
      kein Vergleich doppelt ausgefuehrt */
   for (i=0; i < countTotal; i++) {
      CTermArray[i].VergleichsVektor[i] = Gleich;
      for (j=i+1; j < countTotal; j++) {
         VergleichsT vgl = ORD_VergleicheTerme (CTermArray[i].Term, CTermArray[j].Term);
         CTermArray[i].VergleichsVektor[j] = vgl;
         switch (vgl) {
         case Kleiner:
            CTermArray[j].VergleichsVektor[i] = Groesser;
            break;
         case Groesser:
            CTermArray[j].VergleichsVektor[i] = Kleiner;
            break;
         case Unvergleichbar:
            CTermArray[j].VergleichsVektor[i] = Unvergleichbar;
            break;
         default:
            our_error("SC_VergleichsMatrix: Termvergleich ergab gleich ?!");
         }  
      }
   }


   /* anzGroesserR/L bestimmen */
   for (i=0; i < countTotal; i++) {
      for (j=0; j < countTotal; j++) {
         VergleichsT vgl = CTermArray[i].VergleichsVektor[j];
         if (vgl == Kleiner) {
            if (j < countLeft) CTermArray[i].anzGroesserL++;
            else CTermArray[i].anzGroesserR++;
         }
         else if (vgl == Unvergleichbar) {
            if (j < countLeft) CTermArray[i].anzUnvglL++;
            else CTermArray[i].anzUnvglR++;
         }
      }
   }

   /* Suchen nach totalem Minimum */
   for (i=0; i < countTotal; i++) {
      if ((CTermArray[i].anzGroesserL + CTermArray[i].anzGroesserR) == (countTotal - 1)) { /* -1 wg. Element mit sich selber */
         noTotalMin = i;
      }
   }
   if (noTotalMin != -1) {
      BOOLEAN minIsLeft = noTotalMin < countLeft;
      /* Wenn wir totales  Minimum haben, dann ist dies unsere neue rechte Seite */
      /* alles reinschmeissen, was geht */
      CTermArray[noTotalMin].usedAsRightSide = TRUE;
      for (i=0; i < countTotal; i++) {
         if (i != noTotalMin) {
            CTermArray[i].usedAsLeftSide=TRUE;
            if (minIsLeft && (i < countLeft))
               SC_AddPair (REPairs, CTermArray[noTotalMin].Term, CTermArray[i].Term, Kleiner, FALSE, TRUE);
            else 
               SC_AddPair (REPairs, CTermArray[noTotalMin].Term, CTermArray[i].Term, Kleiner, TRUE, TRUE);
         }
      }
      madeAllPairs = TRUE;
      onlyRules = TRUE;
   }

   /* Fall 2 */
   if (!madeAllPairs) {
      /* Suchen nach Element, das Vergleichbar mit einem Element der anderen Seite ist und auch noch kleiner,
         dabei wird das Element genommen, welches bei den meisten Vergleichen verloren hat */
      for (i=0; i < countTotal; i++) {
         if (((i < countLeft) && (CTermArray[i].anzGroesserR > 0)) ||
             ((i >= countLeft) && (CTermArray[i].anzGroesserL > 0))) {
            /* Es gibt also ein Element auf der anderen Seite */
            if ((CTermArray[i].anzGroesserR + CTermArray[i].anzGroesserL) > MinAnzGroesser) {
               Minimum = i;
               MinAnzGroesser = CTermArray[i].anzGroesserR + CTermArray[i].anzGroesserL;
            }
         }
      }
      if (Minimum != -1) {
         /* Yeah - wir waren erfolgreich */
         BOOLEAN      minIsLeft = Minimum < countLeft;

         CTermArray[Minimum].usedAsRightSide=TRUE;
         /* Reinschmeissen aller Regeln, die mit Minimum als rechter Seite
            konstruiert werden koennen */
         for (i=0; i < countTotal; i++) {
            if ((i != Minimum) && (CTermArray[Minimum].VergleichsVektor[i] == Kleiner)) {
               CTermArray[i].usedAsLeftSide=TRUE;
               if (minIsLeft && (i < countLeft))
                  SC_AddPair (REPairs, CTermArray[Minimum].Term, CTermArray[i].Term, Kleiner, FALSE, TRUE);
               else 
                  SC_AddPair (REPairs, CTermArray[Minimum].Term, CTermArray[i].Term, Kleiner, TRUE, TRUE);
            }
         }

         /* Was koennten wir noch alles reinnehmen ? */
         /* es wird mal wieder im Dreieck durch die Matrix gegangen */
         for (i=0; i < countTotal; i++) {
            for (j = i+1; j < countTotal; j++) {
               verbindetLR = ((i < countLeft) && (countLeft <= j));
               switch (CTermArray[i].VergleichsVektor[j]) {
               case Kleiner:
                  if (!CTermArray[j].usedAsLeftSide && !CTermArray[j].usedAsRightSide
                      && !CTermArray[i].usedAsLeftSide) {
                     CTermArray[j].usedAsLeftSide = TRUE;
                     CTermArray[i].usedAsRightSide = TRUE;
                     SC_AddPair (REPairs, CTermArray[i].Term, CTermArray[j].Term, Kleiner, verbindetLR, TRUE);
                  }
                  break;
               case Groesser:
                  if (!CTermArray[i].usedAsLeftSide && !CTermArray[i].usedAsRightSide
                      && !CTermArray[j].usedAsLeftSide) {
                     CTermArray[i].usedAsLeftSide = TRUE;
                     CTermArray[j].usedAsRightSide = TRUE;
                     SC_AddPair (REPairs, CTermArray[i].Term, CTermArray[j].Term, Groesser, verbindetLR, TRUE);
                  }
                  break;
               case Unvergleichbar:
                  if (!CTermArray[i].usedAsLeftSide && !CTermArray[j].usedAsLeftSide) {
                     SC_AddPair (REPairs, CTermArray[i].Term, CTermArray[j].Term, Unvergleichbar,
                                 verbindetLR, TRUE);
                  }
                  break;
               default:
                  our_error("SC_VergleichsMatrix: Termvergleich ergab gleich ?!");
               }
            }
         }
         madeAllPairs = TRUE;
      }
   }
   
   /* Fall 3: Gleichungen generieren */
   if (!madeAllPairs) {
      for (i = 0; i < countLeft; i++) {
         for (j = countLeft; j < countTotal; j++) {
            CTermArray[i].usedAsLeftSide=TRUE;
            CTermArray[j].usedAsLeftSide=TRUE; /* Bei Gleichungen wird immer links genommen */
            SC_AddPair (REPairs, CTermArray[i].Term, CTermArray[j].Term, Unvergleichbar, TRUE, TRUE);
         }
      }
      madeAllPairs = TRUE;
   }

   subsummiert = SC_SubsumptionTest (REPairs, onlyRules);

   /* Classify & CheckSize zusammenhauen */
   /*   if (!onlyRules && !subsummiert) {
      SC_ClassifyEquations (REPairs);*/
   if (!subsummiert) {
      if (useK)
         SC_ClassifyI (REPairs);
      SC_CheckSize(REPairs, SC_MaxAddRules, SC_MaxAddEquations);
   }
   
   /* Aufraeumen: */
   /*   benutzten Speicher freigeben */
   for (i=0; i < countTotal; i++)
      our_dealloc(CTermArray[i].VergleichsVektor);

   our_dealloc (CTermArray);

#if SC_TEST >= 2
   pairPtr= *REPairs;
   fprintf (IO_Strom, "Inserted Pairs:\n");
   while (pairPtr != NULL) {
      switch (pairPtr->vergleich) {
      case Kleiner:
         IO_DruckeFlex("%t -> %t\n", pairPtr->right, pairPtr->left);
         break;
      case Groesser:
         IO_DruckeFlex("%t -> %t\n", pairPtr->left, pairPtr->right);
         break;
      case Unvergleichbar:
         IO_DruckeFlex("%t <> %t\n", pairPtr->right, pairPtr->left);
         break;
      default:
         our_error ("falscher Vergleich in SC_Classify!");
      }
      pairPtr=pairPtr->next;
   }
   fprintf (IO_Strom, "\n");
#endif
}


/********************************************************************************/
/*                                                                              */
/*  Zentrale exportierte Routine                                                */
/*                                                                              */
/********************************************************************************/

/* NFSetCount: Zaehlen der Elemente in set */
static unsigned int NFSetCount (MNF_NFSetT *set) 
{
   unsigned int count = 0;
   MNF_NFSetPtrT ptr;

   ptr=set;

   while (ptr != NULL) {
      count++;
      ptr=ptr->next;
   }

   return count;
}


/* Classifizieren der Regeln/Gleichungen */
void SC_Classify (MNF_NFSetT *NFLeft, MNF_NFSetT *NFRight, SC_REPairsT   **REPairs) {
   unsigned int           countLeft, countRight;
   

   SC_IncNoCalls;
   *REPairs = NULL;
   
   countLeft  = NFSetCount (NFLeft);
   countRight = NFSetCount (NFRight);

   if ((countLeft == 0) || (countRight == 0))
      our_error ("SC_Classify: count is 0");

   if ((countLeft == 1) && (countRight == 1)) {
      VergleichsT vgl = ORD_VergleicheTerme(NFLeft->Term, NFRight->Term);
      SC_IncNoNFSinglePair;

      if (!SC_SubsummiertEinPaar (NFLeft->Term, NFRight->Term, vgl)) {
         SC_AddPair (REPairs, NFLeft->Term, NFRight->Term, vgl, TRUE, TRUE);
      }
   }
   else {
      SC_GeneratePairs (NFLeft, NFRight, countLeft, countRight, REPairs);
   }
   
   /* wurde CP zusammengefuehrt -> nur dann darf NULL zurueck */
   if (*REPairs != NULL) {
      SUE_selectedCPMNFRE(); /* muss aufgerufen werden wg. Speicherfreigabe, es
                                muss immer genau eine der SUE_selectedCP* Funktionen
                                aufgerufen werden */

#if SC_STATISTIK
      do {
         SC_REPairsPtrT         pairPtr;
         pairPtr = *REPairs;
         do {
            switch (pairPtr->vergleich) {
            case Kleiner:
            case Groesser:
               SC_IncNoRulesGenerated;
               break;
            case Unvergleichbar:
               SC_IncNoEquGenerated;
            break;
            default:
               our_error ("falscher Vergleich in SC_Classify!");
            }
            pairPtr=pairPtr->next;
         } while (pairPtr != NULL);
      } while (0); /* Einmal ausfuehren */
#endif
   }
}


/********************************************************************************/
/*                                                                              */
/*  Initialisieren/Statistik/CleanUp                                            */
/*                                                                              */
/********************************************************************************/



static void SC_PrintOptions(unsigned int maxAddRules,
                            unsigned int maxAddEquations) 
{
   /* Ausgeben der Einstellung: */
   fprintf (IO_Strom, "  Select Classification:");
   if (maxAddRules)
      fprintf (IO_Strom, " MaxAddRules=%u", maxAddRules);
   if (maxAddEquations)
      fprintf (IO_Strom, " MaxAddEquations=%u", maxAddEquations);
   if (!maxAddRules && !maxAddEquations)
      fprintf (IO_Strom, " Add one rule/equation");
   fprintf (IO_Strom, "\n");
}


static void SC_ParseInfoStrings(InfoStringT DEF, InfoStringT USR,
                                unsigned int *maxAddRules,
                                unsigned int *maxAddEquations) 
{

   *maxAddRules = 0;
   *maxAddEquations = 0;
   
   if (IS_FindValuedOption(DEF,USR,1,"maxr") == 0) {
      char *str = IS_ValueOfOption(DEF, USR, "maxr");
      if ((str != NULL))
         *maxAddRules = (unsigned int)strtoul (str, NULL, 10);
   }
   if (IS_FindValuedOption(DEF,USR,1,"maxe") == 0) {
      char *str = IS_ValueOfOption(DEF, USR, "maxe");
      if ((str != NULL))
         *maxAddEquations = (unsigned int)strtoul (str, NULL, 10);
   }
}


void SC_ParamInfo (ParamInfoArtT what, InfoStringT DEF, InfoStringT USR) 
{
   unsigned int   maxAddE, maxAddR;
 
   switch (what) {
   case ParamEvaluate:
      SC_ParseInfoStrings (DEF, USR, &maxAddR, &maxAddE);
      SC_PrintOptions (maxAddR, maxAddE);
      break;
   case ParamExplain:
      fprintf (IO_Strom, "Classification of Rules/Equations when using MNF\n");
      fprintf (IO_Strom, "************************************************\n");
      fprintf (IO_Strom, "Parameter: [:maxr=<n>][:maxe=<n>]\n");
      fprintf (IO_Strom, "  maxr: maximal number of rules to add (default: 0)\n");
      fprintf (IO_Strom, "  maxr: maximal number of equations to add (default: 0)\n");
      fprintf (IO_Strom, "   (0 means to add only necessary rules/equations)\n");
      fprintf (IO_Strom, "\n");
      break;
   }
}


void SC_Statistik (InfoStringT IS) {

   if (!(IS_IsIn("all", IS) || IS_IsIn("SC", IS)))
      return;

   if (IS_IsIn("config", IS)) {
#if SC_STATISTIK      
      fprintf(IO_Strom, "SC: Statistics ON\n");
#else
      fprintf(IO_Strom, "SC: Statistics OFF\n");
#endif  /* SC_STATISTIK */     
   }

   if (!SC_Initialised)
      return;
   
#if SC_STATISTIK      
   if (IS_IsIn("standard", IS)) {
      int tiefe1;
      tiefe1 = IO_Einruecktiefe(0, Maximum(6, SC_NoCalls, SC_NoSubsumed, SC_NoSubsumedClassifyEquations,
                                           SC_NoCallsGeneratePairs,SC_NoNFSinglePair,SC_MaxNF));

      IO_DruckeKurzUeberschrift("Select Classification",'+');
      IO_DruckeStatistikZeileEingerueckt("selected CPs", SC_NoCalls, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("Subsumed", SC_NoSubsumed, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("Subsumed in Classify Equations",  SC_NoSubsumedClassifyEquations, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("Calls To Generate Pairs", SC_NoCallsGeneratePairs, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("NF contained only a single pair", SC_NoNFSinglePair, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("Max Elements in NF", SC_MaxNF, tiefe1);
   }
   else if (IS_IsIn("latex-head", IS)) { /* korrespondiert zu standard */
      fprintf(IO_Strom, "SC: sel CP & SC Subsumed & SC Subsumed CE & SC GPCalls & SC Single & SC MaxNF & ");
   }
   else if (IS_IsIn("short", IS)){
      IO_DruckeKurzUeberschrift("SelectClassify",'+');
      IO_DruckeStatistikZeile("Calls To Generate Pairs", SC_NoCallsGeneratePairs);
      IO_DruckeStatistikZeile("Subsumed in Generate Pairs", SC_NoSubsumedClassifyEquations);
   }
   else if (IS_IsIn("slatex-head", IS)){
      fprintf(IO_Strom, "GP Ca.&Sub.& ");
   }
   else if (IS_IsIn("intern", IS)) {
      IO_DruckeStatistikZeile("Rules Generated (SC)", SC_NoRulesGenerated);
      IO_DruckeStatistikZeile("Equations Generated (SC)", SC_NoEquGenerated);
      IO_DruckeStatistikZeile("AC Equal (SC)", SC_ACGleich);
      IO_DruckeStatistikZeile("No of Doubles (SC)", SC_NoDoppelte);
      IO_DruckeStatistikZeile("Pairs subsumed by each other (SC)", SC_NoSubsummiertUntereinander);
   }
#endif  /* SC_STATISTIK */     
}


BOOLEAN SC_InitAposteriori (InfoStringT DEF, InfoStringT USR,
                            unsigned short PermGrad_R, unsigned short PermGrad_E, 
                            BOOLEAN Subsumption) {

   SV_ManagerInitialisieren (&SC_PairsSpeicher, SC_REPairsT, SC_PairsBlock, "RE Pairs (SC)");

   SC_PermGrad_R = PermGrad_R;
   SC_PermGrad_E = PermGrad_E;
   SC_Subsumption = Subsumption;
   SC_Initialised = TRUE;
   SC_ParseInfoStrings (DEF, USR, &SC_MaxAddRules, &SC_MaxAddEquations);
   SC_PrintOptions (SC_MaxAddRules, SC_MaxAddEquations);

   return TRUE;
}
