/**********************************************************************
* Filename : Zielverwaltung.c
*
* Kurzbeschreibung : "Mittelschicht" Zielverwaltung,
*                    also gemeinsame Schnittstelle von NormaleZiele und
*                    MNF_Ziele nach oben
*
* Autoren : Andreas Jaeger
*
* 10.07.96: Erstellung aus der urspruenglichen Zielverwaltung heraus
* 22.02.97: Mehrfachaufrufclean. AJ.
**********************************************************************/

/* Konzept:
   Ziele werden in zwei einfachen linearen Listen verwaltet:
   ZielListe: Noch zu beweisende Ziele
   ZielFertigListe: Bewiesene Ziele

   Fuer jedes Ziel wird festgelegt von welcher Zielbehandlung es
   bearbeitet wird.
   */

/* TODO:
   - Paramodulation ordentlich machen (klappt fuer den Fall 1 Ziel)
   - Ausgaben ordentlich
   - Optimieren, doppelten Code eliminieren (teilweise gemacht)
   - Beobachtungslauf
   */

#include "compiler-extra.h"
#include "Ausgaben.h"
#include "DSBaumKnoten.h"
#include "DSBaumOperationen.h"
#include "DSBaumTest.h"
#include "InfoStrings.h"
#include "LispLists.h"
#include "MatchOperationen.h"
#include "MNF_Ziele.h"
#include "NFBildung.h"
#include "PhilMarlow.h"
#include "NormaleZiele.h"
#include "RUndEVerwaltung.h"
#include "Subsumption.h"
#include "Statistiken.h"
#include "SUEM.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "general.h"
#include "parser.h"
#include "pcl.h"
#include <stdlib.h>
#include <string.h>
#include "ClasPatch.h"

/*********************/
/* globale Variablen */
/*********************/

unsigned short int AnzahlZiele;
static Z_BeobachtungsProzT  ZielBeobachtungsProz;
static BOOLEAN          continueAfterFirstProved;        /* Fortfahren nach dem ersten bewiesenen Ziel? */
static BOOLEAN          statistikAfterProved;            /* Statistik nach jedem bewiesenem Ziel? */
static unsigned long    period; /* Zielbehandlung jeweils nach period Regeln und Gleichungen */
Z_ZielbehandlungsModulT Z_ZielModul = Z_NichtDefiniert;
Z_ZielGesamtStatusT Z_ZielGesamtStatus;

/* Achtung: Folgende Counter sind essentiell und duerfen nicht rausgenommen
   werden */
unsigned short int AnzErfSubsumptionenZiele,          /* Ziele, die von der GM subsummiert wurden.*/
                   AnzErfZusfuehrungenZiele;          /* Hack: Nicht statisch, denn Konsulat setzt diese zurueck. */

#define IncAnzErfSubsumptionenZiele        AnzErfSubsumptionenZiele++
#define IncAnzErfZusammenfuehrungenZiele   AnzErfZusfuehrungenZiele++

/* ------------------------------------------------------------------------- */
/*                              NEU-WIP                                      */
/* ------------------------------------------------------------------------- */

typedef struct GoalListTS {
   ZielePtrT   firstElem;
   ZielePtrT   lastElem;
} GoalListT;


/*********************/
/* globale Variablen */
/*********************/

/* Speicherverwaltung */
static SV_SpeicherManagerT Z_Speicher;

static GoalListT ZielListe;
static GoalListT ZielFertigListe;


/*
 * Forward-Declarationen
 */
static BOOLEAN TesteTrivialitaeten (void);


/* ------------------------------------------------------------------------- */
/*                               Narrowing                                   */
/* ------------------------------------------------------------------------- */

   
static void AddEqXXIsTrue (void)                                                   /* Aufnahme von eq(x,x) -> true */
{
   TermT links = TO_(SO_EQ,SO_ErstesVarSymb,SO_ErstesVarSymb),
         rechts = TO_(SO_TRUE);
   SUE_CPinitial(links,rechts);
   SUE_recentCPinsert();
}

static void ParamodulationsZiel (ZielePtrT element) 
{
   /*
     Hat man existenzquantifizierte Ziele s(i) = t(i), 
     1<=i<=n, so macht man in der Theorie daraus Regeln eq(s(i),t(i)) -> false(i)
     und nimmt zusaetzlich die Regel eq(x,x) -> true auf. Als regulaere Zielmenge
     nimmt man true = false(i). Das existenzquantifizierte Ziel i ist genau dann
     gezeigt, wenn true und false(i) zusammengefuehrt wurden. Natuerlich ist es
     kein Problem, zusaetzlich gewoehnliche Ziele zu behandeln.
   */
   TermT Flunder1 = TO_(SO_TRUE),
         Flunder2 = TO_(SO_FALSE),
         links = TO_Termkopie (element->orginalLinks),
         rechts = TO_Termkopie (element->orginalRechts);
#if COMP_PCL
   PCL_ZielAusDerSpezifikation(Flunder1, Flunder2, &(element->pclID));
#endif
   NZ_ZielEinfuegen (element, Flunder1, Flunder2);
   TO_TermeLoeschen (Flunder1,Flunder2);

   Z_UngleichungEinmanteln(&links,&rechts);
   SUE_CPinitial(links,rechts);
   SUE_recentCPinsert();
}


/* ------------------------------------------------------------------------- */
/*                                                                           */
/* ------------------------------------------------------------------------- */

static void InitZielBeobachtungsProz (ZielePtrT zielPtr) {

  switch (zielPtr->modul) {
  case Z_NormaleZiele:
  case Z_Paramodulation:
  case Z_Repeat:
    NZ_InitZielBeobachtungsProz (zielPtr);
    break;
  case Z_MNF_Ziele:
    MNFZ_InitZielBeobachtungsProz (zielPtr);
    break;
  case Z_NichtDefiniert:
    our_error ("Zielverwaltung nicht definiert!");
    break;     
  }
}


/* Aus SymbolOperationen geholt */
/* Testet ob Variablen im Term sind */
static BOOLEAN VariablenIn(term tt)
{
  if variablep(tt)
    return TRUE;
  else {
    func_symb fkt = term_func(tt);                                                    /* Topsymbol des Terms holen */
    int runny;
    if (sig_arity(fkt)!=0)                                             /* Funktions-Symbol. Argumente durchlaufen. */
      for (runny=0; runny<sig_arity(fkt); runny++)
        if (VariablenIn(term_arg(tt,runny)))
          return TRUE;
    return FALSE;
  }                                                                            /* 'else': Konstante => Ueberlesen. */
}


static void ListeEinfuegen (GoalListT *liste, ZielePtrT element) 
{
   if (liste->firstElem == NULL) {
      liste->firstElem = element;
      liste->lastElem = element;
      element->Vorg = NULL;
      element->Nachf = NULL;
   } else {
      element->Vorg = liste->lastElem;
      element->Nachf = NULL;
      liste->lastElem->Nachf = element;
      liste->lastElem = element;
   }
}


static void ListeEinfuegenSortiert (GoalListT *liste, ZielePtrT element) 
{
   ZielePtrT  listePtr;
   
   if (liste->firstElem == NULL) {
      /* erstes Element in Liste einfuegen */
      liste->firstElem = element;
      liste->lastElem = element;
      element->Vorg = NULL;
      element->Nachf = NULL;
   } else {
      /* Liste enthaelt bereits Elemente */
      listePtr = liste->firstElem;
      while (listePtr != NULL && (element->nummer > listePtr->nummer)) {
         listePtr = listePtr->Nachf;
      }
      /* Abbruch, also listePtr == NULL oder element->nummer <= listePtr->nummer */
      if (listePtr == NULL) {
         /* Ende der Liste erreicht */
         element->Vorg = liste->lastElem;
         element->Nachf = NULL;
         liste->lastElem->Nachf = element;
         liste->lastElem = element;
      } else {
         /* ab in die Mitte */
         /* einfuegen vor listePtr */
         element->Vorg = listePtr->Vorg;
         element->Nachf = listePtr;
         if (listePtr == liste->firstElem) {
            /* Vorne einfuegen */
            listePtr->Vorg = element;
            liste->firstElem = element;
         } else {
            /* in der Mitte einfuegen */
            listePtr->Vorg->Nachf = element;
            listePtr->Vorg = element;
         }
      }  
   }
}


/* Element aus ZielListe nach ZielFertigListe umhaengen */
static void UmhaengenInListen (ZielePtrT element) 
{
   /* Letztes Element raus ? */
   if (ZielListe.firstElem == ZielListe.lastElem) {
      ZielListe.firstElem = NULL;
      ZielListe.lastElem = NULL;
   } else if (ZielListe.firstElem == element) {
      ZielListe.firstElem = element->Nachf;
      ZielListe.firstElem->Vorg = NULL;
   } else if (ZielListe.lastElem == element) {
      ZielListe.lastElem = element->Vorg;
      ZielListe.lastElem->Nachf = NULL;
   } else {
      element->Vorg->Nachf = element->Nachf;
      element->Nachf->Vorg = element->Vorg;
   }
   ListeEinfuegenSortiert (&ZielFertigListe, element);
   if (element->beobachtProz) {
     element->beobachtProz = NULL;
     if (ZielListe.firstElem) {
       ZielListe.firstElem->beobachtProz = ZielBeobachtungsProz;
       InitZielBeobachtungsProz (ZielListe.firstElem);
     }
   }
}


/* Fuegt Ziel in Liste ein, dabei muss left != right sein */
void Z_EinZielEinfuegen (TermT left, TermT right, unsigned short int Index, Z_ZielbehandlungsModulT modul)
{
   ZielePtrT zielPtr;

   SV_Alloc (zielPtr, Z_Speicher, ZielePtrT);

#if COMP_PCL
   if (modul != Z_Paramodulation) {
     /* bei Paramodulation geschieht initialisierung in ParamodulationsZiel */
     PCL_ZielAusDerSpezifikation(left, right, &(zielPtr->pclID));
   }
#endif
   zielPtr->nummer = Index;
   zielPtr->orginalLinks = left;
   zielPtr->orginalRechts = right;
   zielPtr->status = Z_Open;
   zielPtr->privat.NZ = NULL;
   zielPtr->subSumNr = 0;
   if (Index == 1)
     zielPtr->beobachtProz = ZielBeobachtungsProz;
   else
     zielPtr->beobachtProz = NULL;
   zielPtr->modul = modul;
   ListeEinfuegen (&ZielListe, zielPtr);
   /* Modulspezifische Initialisierung */
   switch (modul) {
   case Z_NormaleZiele:
   case Z_Repeat:
     NZ_ZielEinfuegen (zielPtr, left, right);
     break;
   case Z_Paramodulation:
     ParamodulationsZiel (zielPtr);
     break;
   case Z_MNF_Ziele:
     MNFZ_ZielEinfuegen (zielPtr);
     break;
   case Z_NichtDefiniert:
     our_error ("Zielverwaltung nicht definiert!");
     break;
   }
}



void Z_ZieleAusSpezifikationHolen(void)
{
   ListenT ZielIndex;
   BOOLEAN SchonEinExistenzziel = FALSE;
   ZielePtrT zielePtr;

   AnzahlZiele = 0;
   for (ZielIndex=PM_Ziele; ZielIndex; Cdr(&ZielIndex)) {
      ListenT Eintrag = car(ZielIndex);
      TermT links = TO_Termkopie(car(Eintrag)),
            rechts = TO_Termkopie(cdr(Eintrag));
      /* Trivialitaeten werden spaeter in TesteTrivialitaeten behandelt */
      if ((TO_TermIstNichtGrund(links) || TO_TermIstNichtGrund(rechts))
          && !TO_TermeGleich(links,rechts)) {              /* Existenzquantifizierung */
         if (PA_ExtModus==completion)
            our_error("No existential quantification of goals possible in COMPLETION mode.");
         if (SchonEinExistenzziel) 
            our_error("Existential quantification is possible for one goal only.");
         AddEqXXIsTrue();
         SchonEinExistenzziel = TRUE;
         Z_EinZielEinfuegen (links, rechts, ++AnzahlZiele, Z_Paramodulation);
      }
      else {
         Z_EinZielEinfuegen(links, rechts, ++AnzahlZiele, Z_ZielModul);
         if (ZielBeobachtungsProz)
            (*ZielBeobachtungsProz)(links, rechts);
      }
   }
#if !COMP_ACIDWM
   if ((zielePtr = ZielListe.firstElem)) {
      fprintf(IO_Strom,  "\nGoals:\n------\n\n" );
      do {
        fprintf(IO_Strom, "(%4u)  ", zielePtr->nummer);
        if (zielePtr->modul == Z_Paramodulation)
          NZ_AusgabeZiel (zielePtr, Z_AusgabeInitial);
        else 
          IO_DruckeFlex("%t ?=? %t\n", zielePtr->orginalLinks, zielePtr->orginalRechts);
      } while ((zielePtr = zielePtr->Nachf));
   }
#endif
   if (AnzahlZiele)
     Z_ZielGesamtStatus = Z_Offen;
   else
     Z_ZielGesamtStatus = Z_KeineZiele;

   /* Trivialitaeten testen, Resultat ignorieren */
   (void)TesteTrivialitaeten ();
}



BOOLEAN Z_ZielmengeLeer(void)
/* Liefert TRUE, wenn es kein zu beweisendes Ziel mehr gibt
   (entweder wurden alle Ziele gezeigt, oder es gab keine) */
{
   return (ZielListe.firstElem == NULL);
}


void Z_ZielBeobachtungsProzAnmelden (Z_BeobachtungsProzT proz)
/* Anmeldung einer Prozedur proz, die aufgerufen wird, sobald sich das 1. Ziel aendert.
   Wurde das erste Ziel bewiesen, so wird proz mit dem naechsten Ziel aufgerufen.
   proz wird insbesondere auch dann aufgerufen, sobald das erste Ziel eingefuegt wird.
*/
{
   ZielBeobachtungsProz = proz;
}

/* Liefert TRUE, wenn alle gewuenschten Ziele (gemaess Einstellungen alle oder nur */
/* das erste) bewiesen werden konnten.*/
BOOLEAN Z_ZieleBewiesen(void) 
{
   return (Z_ZielmengeLeer () ||
           (((AnzErfSubsumptionenZiele > 0) || (AnzErfZusfuehrungenZiele > 0))
            && !continueAfterFirstProved));
}

/* Anzahl bewiesener Ziele */
int Z_AnzahlBewiesenerZiele (void)
{
  return (AnzErfSubsumptionenZiele + AnzErfZusfuehrungenZiele);
}

/* Anzahl spezifizierter Ziele */
int Z_AnzahlZiele (void) {
  return (int)AnzahlZiele;
}

/* Mindestens ein Ziel wurde bewiesen */
BOOLEAN Z_MindEinZielBewiesen (void)
{
  return (AnzErfSubsumptionenZiele > 0) || (AnzErfZusfuehrungenZiele > 0);
}


void Z_ZielmengenBeobachtungsDurchlauf(Z_BeobachtungsDurchlaufProcT AP)   
/* Wendet AP auf jedes Ziel an, wobei AP die Ziele nicht veraendert.*/
{
   ZielePtrT zielePtr;

   zielePtr = ZielListe.firstElem;

   while (zielePtr) {
      switch (zielePtr->modul) {
      case Z_MNF_Ziele:
         MNFZ_ZielBeobachtung (zielePtr, AP);
         break;
      case Z_Paramodulation:
      case Z_NormaleZiele:
      case Z_Repeat:
         NZ_ZielBeobachtung (zielePtr, AP);
	 break;
      default:
         /* TODO */
         break;
      }
      zielePtr = zielePtr->Nachf;
   }
}

INLINE static void AusgabeStatus (ZielePtrT ziel) {

  switch (ziel->status) {
  case Z_Open:
    fprintf (IO_Strom, " still open");
    break;
  case Z_Joined: 
    fprintf (IO_Strom, " joined");
    break;
  case Z_Subsumed: 
    fprintf (IO_Strom, " subsumed by equation %ld", ziel->subSumNr);
    break;
  case Z_Unified:
    fprintf (IO_Strom, " unified");
    break;
  }
}


INLINE static void AusgabeListe (GoalListT *liste, const char *title, Z_AusgabeArtT Z_AusgabeArt) 
{
   ZielePtrT zielePtr;

   zielePtr = liste->firstElem;
   if (zielePtr != NULL) {
      if (Z_AusgabeArt != Z_AusgabeSpez)
         IO_DruckeKurzUeberschrift (title, '*');
      while (zielePtr) {
         if (Z_AusgabeArt != Z_AusgabeSpez) {
            fprintf (IO_Strom, "No. %4u\n", zielePtr->nummer);
            fprintf (IO_Strom, "--------\n");
            if (zielePtr->status != Z_Open) {
               AusgabeStatus (zielePtr);
               fprintf (IO_Strom, "\n");
            }
           IO_DruckeFlex(" original %t ?= %t\n", zielePtr->orginalLinks, zielePtr->orginalRechts);
         }
         switch (zielePtr->modul) {
         case Z_MNF_Ziele:
            MNFZ_AusgabeZiel (zielePtr, Z_AusgabeArt);
            break;
         case Z_NormaleZiele:
         case Z_Paramodulation:
	 case Z_Repeat:
            NZ_AusgabeZiel (zielePtr, Z_AusgabeArt);
            break;
         default:
            /* TODO */
            break;
         }
         zielePtr = zielePtr->Nachf;
      }
   }
}


void Z_AusgabeZielmenge (Z_AusgabeArtT Z_AusgabeArt)
{
   AusgabeListe (&ZielListe, "Open Goals", Z_AusgabeArt);
   AusgabeListe (&ZielFertigListe, "Proven Goals", Z_AusgabeArt);
}

INLINE static void AusgabeZieleEndeListe (GoalListT *liste, const char *title) 
{
   ZielePtrT zielePtr;


   zielePtr = liste->firstElem;
   if (zielePtr != NULL) {
     fprintf (IO_Strom, "\n%s:\n", title);
     while (zielePtr) {
       fprintf (IO_Strom, "No. %2u: ", zielePtr->nummer);
       if (zielePtr->modul == Z_Paramodulation) {
         NZ_AusgabeZiel (zielePtr, Z_AusgabeKurz);
       } else {
         IO_DruckeFlex(" %t ?= %t", zielePtr->orginalLinks, zielePtr->orginalRechts);
         if (zielePtr->status != Z_Open) {
           AusgabeStatus (zielePtr);
           fprintf (IO_Strom, ", current: ");
           switch (zielePtr->modul) {
           case Z_MNF_Ziele:
             MNFZ_AusgabeZiel (zielePtr, Z_AusgabeKurz);
             break;
           case Z_NormaleZiele:
	   case Z_Repeat:
             NZ_AusgabeZiel (zielePtr, Z_AusgabeKurz);
             break;
           default:
             break;
           }
         }
       }
       fprintf (IO_Strom, "\n");
       zielePtr = zielePtr->Nachf;
     }
   }
}

static void printAnzahl (const char *status, const char *was, const char *were, int Anzahl, int AnzahlGesamt)
{
  if (Anzahl) {
    if (Anzahl == 1 && AnzahlGesamt == 1)
      fprintf (IO_Strom, "which %s %s", was, status);
    else if (Anzahl == AnzahlGesamt)
      fprintf (IO_Strom, "all %s %s", were, status);
    else if (Anzahl == 1)
      fprintf (IO_Strom, "1 %s %s", was, status);
    else
      fprintf (IO_Strom, "%d %s %s", Anzahl, were, status);
  }
}

/* Zusammenfassung der Ziele mit Statistik, completed ist TRUE wenn das System vervollstaendig ist,
   so ist unterscheidbar, ob die nicht bewiesenen Ziele Wiederlegt wurden oder nicht */
void Z_AusgabeZieleEnde (BOOLEAN completed)
{
  unsigned short int AnzahlBewiesen, AnzahlOpen;

  if (completed)
    AusgabeZieleEndeListe (&ZielListe, "Refuted Goals");
  else
    AusgabeZieleEndeListe (&ZielListe, "Open Goals");

  AusgabeZieleEndeListe (&ZielFertigListe, "Proved Goals");

  AnzahlBewiesen = AnzErfSubsumptionenZiele + AnzErfZusfuehrungenZiele;
  AnzahlOpen = AnzahlZiele - AnzahlBewiesen;

  switch (AnzahlZiele) {
  case 0:
    fprintf (IO_Strom, "No goals were specified.\n");
    return;
  case 1:
    fprintf (IO_Strom, "1 goal was specified, ");
    break;
  default:
    fprintf (IO_Strom, "%hu goals were specified, of which ", AnzahlZiele);
    break;
  }
  
  printAnzahl ("proved", "was", "were", AnzahlBewiesen, AnzahlZiele);
  if (AnzahlBewiesen != AnzahlZiele && AnzahlBewiesen > 0)
    fprintf (IO_Strom, " and ");
  if (completed) {
    printAnzahl ("refuted", "was", "were", AnzahlOpen, AnzahlZiele);
  } else {
    printAnzahl ("still open", "is", "are", AnzahlOpen, AnzahlZiele);
  }
  fprintf (IO_Strom, ".\n");
  if (AnzahlZiele) {
    if (AnzahlBewiesen == AnzahlZiele)
      Z_ZielGesamtStatus = Z_AlleBewiesen;
    else if (AnzahlBewiesen == 0 && completed)
      Z_ZielGesamtStatus = Z_AlleWiederlegt;
    else if (AnzahlBewiesen && completed)
      Z_ZielGesamtStatus = Z_EinigeWiederlegt;
    else if (AnzahlBewiesen && !completed)
      Z_ZielGesamtStatus = Z_EinigeBewiesen;
    else
      Z_ZielGesamtStatus = Z_Offen;
  }
}

/* Parameter sagt, ob return TRUE ausgefuehrt wird oder nicht */
#define ErledigtErfolgsTest(/*BOOLEAN*/abort_if_result_true)                   \
      if (result) {                                                             \
         ZielePtrT nachf;                                                       \
         if (zielePtr->status == Z_Joined)                                      \
            IncAnzErfZusammenfuehrungenZiele;                                   \
         else if (zielePtr->status == Z_Subsumed)                               \
            IncAnzErfSubsumptionenZiele;                                        \
         nachf = zielePtr->Nachf;                                               \
         UmhaengenInListen (zielePtr);                                          \
         zielePtr = nachf;                                                      \
         if (!continueAfterFirstProved && abort_if_result_true)                 \
            return TRUE;                                                        \
         if (statistikAfterProved)                                              \
           ST_GoalProvedStatistik ("Proved a Goal");                            \
      } else {                                                                  \
         zielePtr = zielePtr->Nachf;                                            \
      }

   
/* durch alle Ziele durchlaufen und auf Trivialitaeten testen. */
static BOOLEAN TesteTrivialitaeten (void)
{
   ZielePtrT zielePtr;
   BOOLEAN result;
   
   zielePtr = ZielListe.firstElem;
   while (zielePtr) {
      switch (zielePtr->modul) {
      case Z_MNF_Ziele:
         result = MNFZ_TesteZielErledigt (zielePtr);
         break;
      case Z_NormaleZiele:
      case Z_Repeat:
         result = NZ_TesteZielErledigt (zielePtr);
         break;
      case Z_Paramodulation:
      default:
         result = FALSE;
         break;
      }
      ErledigtErfolgsTest (FALSE);
   }
   return TRUE;
}

/* Alle Ziele auf Reduzierbarkeit mit der neuen Regel pruefen. Falls
   (je nach Modus) ein bzw. alle Ziele zusammengefuehrt werden
   konnten, wird TRUE zurueckgegeben, sonst FALSE. Zusammengefuehrte
   Ziele werden aus der Zielmenge entfernt.*/
BOOLEAN Z_ZieleErledigtMitNeuerRegel (RegelT Regel)
{
   ZielePtrT zielePtr;
   BOOLEAN   result = FALSE;
   BOOLEAN   durchlaufen;
   
   durchlaufen = ((period == 1)
                  || (RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt())%period == 0);

   zielePtr = ZielListe.firstElem;
   while (zielePtr) {
      switch (zielePtr->modul) {
      case Z_MNF_Ziele:
         result = MNFZ_ZielErledigtMitNeuerRegel (zielePtr, Regel, durchlaufen);
         break;
      case Z_Paramodulation:
      case Z_NormaleZiele:
      case Z_Repeat:
         if (period == 1)
            result = NZ_ZielErledigtMitNeuerRegel (zielePtr, Regel);
         else if (durchlaufen)
            result = NZ_ZielFinalerDurchlauf (zielePtr);
         break;
      default:
         break;
      }
      ErledigtErfolgsTest (TRUE);
   }
   return Z_ZielmengeLeer ();
}


/* Alle Ziele auf Reduzierbarkeit und Subsummierbarkeit mit der neuen
   Gleichung pruefen. Falls (je nach Modus) ein bzw. alle Ziele
   zusammengefuehrt werden konnten, wird TRUE zurueckgegeben, sonst
   FALSE. Zusammengefuehrte Ziele werden aus der Zielmenge entfernt.*/
BOOLEAN Z_ZieleErledigtMitNeuerGleichung(GleichungsT Gleichung)
{
   ZielePtrT zielePtr;
   BOOLEAN   result = FALSE;
   BOOLEAN   durchlaufen;
   
   durchlaufen = ((period == 1)
                  || (RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt())%period == 0);

   zielePtr = ZielListe.firstElem;
   while (zielePtr) {
      switch (zielePtr->modul) {
      case Z_MNF_Ziele:
         result = MNFZ_ZielErledigtMitNeuerGleichung (zielePtr,
                                                      Gleichung,
                                                      durchlaufen);
         break;
      case Z_Paramodulation:
      case Z_NormaleZiele:
      case Z_Repeat:
         if (period == 1)
            result = NZ_ZielErledigtMitNeuerGleichung (zielePtr, Gleichung);
         else if (durchlaufen)
            result = NZ_ZielFinalerDurchlauf (zielePtr);
         break;
      default:
         /* TODO */
         break;
      }
      ErledigtErfolgsTest (TRUE);
   }
   return Z_ZielmengeLeer ();
}


/* Falls nicht nach jeder Einfuegung die Ziele behandelt wurden,
   muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden.
 */
BOOLEAN Z_FinalerZielDurchlauf(void)
{
   ZielePtrT zielePtr;
   BOOLEAN   result = FALSE;

   if ((PA_ExtModus == proof) && (period == 1))
      return FALSE;             /* Kein Durchlauf => keine Aenderung gegenueber vorher. */

   zielePtr = ZielListe.firstElem;

   while (zielePtr) {
      switch (zielePtr->modul) {
      case Z_MNF_Ziele:
         result = MNFZ_ZielFinalerDurchlauf (zielePtr);
         break;
      case Z_Paramodulation:
      case Z_NormaleZiele:
      case Z_Repeat:
         result = NZ_ZielFinalerDurchlauf (zielePtr);
         break;
      default:
         /* TODO */
         break;
      }
      ErledigtErfolgsTest (FALSE);
   }
   return Z_ZielmengeLeer ();
}


static void Z_CleanupListe (GoalListT *liste) 
{
   ZielePtrT zielePtr, nachf;

   zielePtr = liste->firstElem;

   while (zielePtr) {
      switch (zielePtr->modul) {
      case Z_MNF_Ziele:
         MNFZ_LoeschePrivateData (zielePtr);
         break;
      case Z_Paramodulation:
      case Z_NormaleZiele:
      case Z_Repeat:
         NZ_LoeschePrivateData (zielePtr);
         break;
      default:
         /* TODO */
         break;
      }
      TO_TermLoeschen (zielePtr->orginalLinks);
      TO_TermLoeschen (zielePtr->orginalRechts);
      
      nachf = zielePtr->Nachf;
      SV_Dealloc (Z_Speicher, zielePtr);
      zielePtr = nachf;
   }
}

/* Alle Ziele incl. Termen wegwerfen. */
void Z_Cleanup (void) 
{
   Z_CleanupListe (&ZielListe);
   Z_CleanupListe (&ZielFertigListe);
}

/*=================================================================================================================
  =  Zielbehandlung                                                                                               =
  =================================================================================================================*/

void Z_Statistik(InfoStringT str)
{
   if (!(IS_IsIn("all", str) || IS_IsIn("Z", str)))
      return;
   if (IS_IsIn("standard", str)){
      IO_DruckeKurzUeberschrift("Treatment of Goals", '+');
      IO_DruckeStatistikZeile("Goals proven by subsumption", AnzErfSubsumptionenZiele);
      IO_DruckeStatistikZeile("Goals proven by rewriting", AnzErfZusfuehrungenZiele);
   }
   else if (IS_IsIn("latex-head", str))
      fprintf(IO_Strom, "Goals: Subs & Rewr & ");

   if (Z_ZielModul == Z_MNF_Ziele)
      MNFZ_Statistik (str);
   NZ_Statistik (str);
}


static void Z_ParseInfo (InfoStringT DEF, InfoStringT USR,
                         BOOLEAN *cont, unsigned long int *infoperiod, Z_ZielbehandlungsModulT *ZielModul)
{
   char *value;

   /* Wenn Paramodulation benutzt wird, dann wird NormaleZiele genommen,
      ansonsten wird - wenn nichts anderes eingegeben - NormaleZiele
      genommen */
   switch (IS_FindOption (DEF, USR, 3, "snf", "rnf", "mnf")) {
   default:
   case 0:
      *ZielModul = Z_NormaleZiele;
      break;
   case 1:
      *ZielModul = Z_Repeat;
      break;
   case 2:
      *ZielModul = Z_MNF_Ziele;
      break;
   }
   
   switch (IS_FindOption(DEF, USR, 2, "stop", "continue")){
   case 0: *cont = FALSE;break;
   case 1: default: *cont = TRUE; break;
   }
   switch (IS_FindOption(DEF, USR, 1, "xstat")){
   case 0:
     statistikAfterProved = TRUE;
     break;
   default:
     statistikAfterProved = FALSE;
   }
   
   if (IS_FindValuedOption(DEF, USR, 1, "period") == 0) {
      value = IS_ValueOfOption(DEF, USR, "period");
      if (value)
         *infoperiod = strtoul(value, NULL, 10);
   }
   else
      *infoperiod = 1;

#if !COMP_POWER
   fprintf(IO_Strom, "\n* GOALS:   Period = %lu, %s after success. Module: ", 
           *infoperiod, *cont?"continue":"stop");
   switch (*ZielModul) {
   case Z_NormaleZiele:
      fprintf (IO_Strom, "Normal");
      break;
   case Z_MNF_Ziele:
      fprintf (IO_Strom, "MNF");
      break;
   case Z_NichtDefiniert:
      fprintf (IO_Strom, "Not specified ");
      break;
   case Z_Paramodulation:
      break;
   case Z_Repeat:
      fprintf (IO_Strom, "RNF");
      break;
   }
   if (PM_Existenzziele())
      fprintf (IO_Strom, " with Paramodulation");
   fprintf (IO_Strom, ".\n");
#endif
}


void Z_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   BOOLEAN cont; 
   unsigned long int infoperiod;
   Z_ZielbehandlungsModulT ZielModul;
   

   switch (what) {
   case ParamEvaluate:
      Z_ParseInfo (DEF, USR, &cont, &infoperiod, &ZielModul);
      switch (ZielModul) {
      case Z_NormaleZiele:
      case Z_Repeat:
         NZ_ParamInfo (what, DEF, USR);
         break;
      case Z_MNF_Ziele:
         MNFZ_ParamInfo (what, DEF, USR);
         break;
      case Z_NichtDefiniert:
      case Z_Paramodulation:
         our_error ("ZielModul is not defined. :-(");
         break;
      }
      fprintf (IO_Strom, "\n");
      break;
   case ParamExplain:
      fprintf(IO_Strom, "Treatment of Goals\n");
      fprintf(IO_Strom, "******************\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. [continue|stop]\n");
      fprintf(IO_Strom, "2. [period=<n>]\n");
      fprintf(IO_Strom, "3. [xstat]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. After the first goal has been proved, Waldmeister will\n");
      fprintf(IO_Strom, "continue resp. stop.\n");
      fprintf(IO_Strom, "2. Goals will be touched after <n> new rules/equations have been\n");
      fprintf(IO_Strom, "generated.\n");
      fprintf(IO_Strom, "3. Output statistics after every proven goal.\n");
      NZ_ParamInfo (what, DEF, USR);
      MNFZ_ParamInfo (what, DEF, USR);   
      break;
   default:
      our_error ("Zielverwaltung: Z_ParamInfo: case Label fehlt");
      break;
   }
}


BOOLEAN Z_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   SV_ManagerInitialisieren (&Z_Speicher, ZieleT, Z_SpeicherBlock, "Zielmenge");
   AnzahlZiele = 0;

   ZielListe.firstElem = NULL;
   ZielListe.lastElem = NULL;
   ZielFertigListe.firstElem = NULL;
   ZielFertigListe.lastElem = NULL;

   AnzErfSubsumptionenZiele = 0;
   AnzErfZusfuehrungenZiele = 0;

   ZielBeobachtungsProz = NULL;
 
   Z_ParseInfo (DEF, USR, &continueAfterFirstProved, &period, &Z_ZielModul);
   
   if (Z_ZielModul == Z_MNF_Ziele)
      MNFZ_InitAposteriori (DEF, USR);
   
   NZ_InitAposteriori (DEF, USR);
   
   return TRUE;
}



void Z_UngleichungEinmanteln(UTermT *links, UTermT *rechts)
{
   TermT Neu;
   TO_NachfBesetzen(TO_TermEnde(*links),*rechts);
   TO_TermzelleHolen(Neu);
   TO_SymbolBesetzen(Neu,SO_EQ);
   TO_NachfBesetzen(Neu,*links);
   TO_EndeBesetzen(Neu,TO_TermEnde(*rechts));
   *links = Neu;
   *rechts = TO_(SO_FALSE);
}


BOOLEAN Z_IstGemantelteUngleichung(TermT links, TermT rechts)
{
  if (PM_Existenzziele()) {
    SymbolT f = TO_TopSymbol(links),
            g = TO_TopSymbol(rechts);
    return (f==SO_EQ && g==SO_FALSE) || (f==SO_FALSE && g==SO_EQ);
  }
  else
    return FALSE;
}


void Z_UngleichungEntmanteln(UTermT *OhneMantelLinks, UTermT *OhneMantelRechts, TermT MitMantelLinks, TermT MitMantelRechts)
{
  TermT MitMantel = SO_SymbGleich(TO_TopSymbol(MitMantelLinks),SO_EQ) ? MitMantelLinks : MitMantelRechts;
  *OhneMantelLinks = TO_ErsterTeilterm(MitMantel);
  *OhneMantelRechts = TO_NaechsterTeilterm(*OhneMantelLinks);
}


BOOLEAN Z_UngleichungEntmantelt(UTermT *OhneMantelLinks, UTermT *OhneMantelRechts, TermT MitMantelLinks, TermT MitMantelRechts)
{
  if (PM_Existenzziele()) {
    SymbolT f = TO_TopSymbol(MitMantelLinks),
            g = TO_TopSymbol(MitMantelRechts);
    TermT MitMantel = f==SO_EQ && g==SO_FALSE ? MitMantelLinks : 
                      f==SO_FALSE && g==SO_EQ ? MitMantelRechts : NULL;
    if (MitMantel) {
      *OhneMantelLinks = TO_ErsterTeilterm(MitMantel);
      *OhneMantelRechts = TO_NaechsterTeilterm(*OhneMantelLinks);
      return TRUE;
    }
  }
  return FALSE;
}


BOOLEAN Z_IstSkolemvarianteZuUngleichung(TermT links, TermT rechts)
{
  TermT s, t;
  if (Z_UngleichungEntmantelt(&s,&t, links,rechts)) {
    RegelT Ungleichung;
    RE_forRegelnRobust(Ungleichung, {
      TermT u; TermT v;
      if (RE_UngleichungEntmantelt(&u,&v, Ungleichung) && MO_TupelSkolemAllgemeiner(v,u, s,t))
        return TRUE;
    })
  }
  return FALSE;
}


BOOLEAN Z_IstGrundungleichung(TermT links, TermT rechts)
{
  return Z_IstGemantelteUngleichung(links,rechts) &&
    TO_TermIstGrund(links) && TO_TermIstGrund(rechts);
}
