/*=================================================================================================================
  =                                                                                                               =
  =  Subsumption.c                                                                                                =
  =                                                                                                               =
  =  Subsumption mit der Gleichungsmenge bzw. einzelnen Gleichungen                                               =
  =                                                                                                               =
  =  27.03.96  Arnim. Aus der Interreduktion ausgegliedert.                                                       =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "Hauptkomponenten.h"
#include "MatchOperationen.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "Subsumption.h"
#include "TermOperationen.h"
#include "general.h"

/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Subsumption --------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* - Definition: Eine Gleichung s=t subsummiert eine weitere Gleichung l=r, falls eine Stelle p aus O(l,r),     */
/* -             eine Substitution sigma und ein Term u existieren, so dass gilt:                               */
/* -              l = u[p <- sigma(s)] und r = u[p <- sigma(t)].                                                */
/* -             "s=t ist Verallgemeinerung von l=r"                                                            */
/* -                                                                                                            */
/* - Definition: Eine Gleichung l=r heisst Subgleichung von s=t, falls eine Stelle p aus O(s,t) existiert, so   */
/* -             dass l = s|p und r = t|p, wobei fuer beliebige Terme u gilt:                                   */
/* -                      s [p <- u] == t [p <- u].                                                             */
/* -             "l=r entsteht aus s=t durch Entfernen von Kontext"                                             */
/* -                                                                                                            */
/* - Vorgehensweise: Es gibt zwei Situationen, in denen ein Subsumptionstest noetig ist:                        */
/* -                                                                                                            */
/* - 1. Vor dem Einfuegen einer neuen Gleichung wird geprueft, ob diese nicht durch eine bereits vorhandene     */
/* -    Gleichung aus GM subsummiert wird. Ist dies der Fall, so wird die neue Gleichung gar nicht erst aufge-  */
/* -    nommen; andernfalls wird die neue Gleichung in die GM aufgenommen.                                      */
/* - 2. Wird die neue Gleichung nicht von einer aelteren subsummiert, so wird umgekehrt geprueft, ob vielleicht */
/* -    eine der aelteren Gleichungen von der neuen subsummiert wird und somit aus der GM entfernt werden kann. */
/* -                                                                                                            */
/* - (1. heisst auch forward Subsumption, 2. bachward Subsumption (bzgl. GM))                                   */
/* -                                                                                                            */
/* - Bei 2. muessen die Subgleichungen aller in der GM befindlichen Gleichungen bestimmt und geprueft werden.   */
/* - Diese Subgleichungen werden aber bereits unter 1. beim Einfuegen jeder neuen Gleichung berechnet. Deshalb  */
/* - werden die unter 1. erhaltenen Resultate in einer Struktur gespeichert, die das Vorgehen bei 2. erleich-   */
/* - tert, der Subgleichungen-Menge. Diese ist im Modul Subgleichungen realisiert.                              */
/* -                                                                                                            */
/* -                                                                                                            */
/* - Tatsaechlich gewaehlte Struktur/Vorgehensweise:                                                            */
/* -                                                                                                            */
/* - + forward Subsumption: neue Gleichung l=r. Mit Gleichungsbaum Matches suchen und auf Blattebene in der/den */
/* -   rechten Seiten fortsetzen. Dies geschieht fuer alle Subgleichungen von l=r, und zwar nur einmal je       */
/* -   Subgleichung, da die Gleichungen bereits in beiden Richtungen im E-Baum gespeichert sind.                */
/* - + backward Subsumption: die Gleichungen werden zunaechst auf Top-Ebene getestet, und zwar geschieht dies   */
/* -   wiederum mit Hilfe des E-Baums: verlaengertes Interreduktionsproblem=Fortsetzung auf Blattebene in der/  */
/* -   den rechten Seiten der betroffenen Gleichung(en). Anschliessend werden in einer gesonderten Struktur     */
/* -   alle echten Subgleichungen der Gleichungen aus GM ueberprueft. Hierfuer wurde folgende Struktur gewaehlt:*/
/* -   - eine doppelt verkettete Liste von Kopf-Zellen, die jeweils fuer solche Gleichungen eingerichtet werden,*/
/* -     die echte Subgleichungen enthalten. Die Kopf-Zellen enthalten dann einen Verweis auf die Gleichungen   */
/* -     selbst sowie einen Zeiger auf die erste Subgleichung.                                                  */
/* -   - die Gleichungen erhalten einen Verweis auf ihre Kopf-Zelle in der Subgleichungs-Liste, um beim         */
/* -     Loeschen der Gleichung die Subgleichungen mitloeschen zu koennen, ohne lange suchen zu muessen.        */
/* -   - die an den Kopf-Zellen angehaengten Subgleichungen sind als einseitig verkettete Liste realisiert,     */
/* -     da sie nur als ganzes geloescht werden muessen.                                                        */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* - Ausblick : Da bei der backward Subsumption   (wird eine alte Gleichung von einer neuen subs.) nur Matches  */
/* -            in einen Baum berechnet werden muessen, bietet sich hier vielleicht eine ganz andere Struktur   */
/* -            an.                                                                                             */
/* -                                                                                                            */
/* -                                                                                                            */
/* - Subsumption der Ziele :                                                                                    */
/* -           Da die Ziele nach dem Einfuegen einer neuen Gleichung auf Subsummierbarkeit geprueft werden,     */
/* -           bietet sich auch hier eine Speicherung der Subgleichungen zu jedem Ziel an. Dies haette den      */
/* -           Vorteil, dass nicht bei jeder Ueberpruefung die Subgleichungen neu erzeugt werden muessen.       */
/* -           Ein Nachteil ist, dass dann nach jeder Reduktion die Subgleichungsliste neu berechnet werden     */
/* -           muesste. Abhilfe kann hier ein Flag schaffen, welches die Aktualitaet der Subgleichungsliste     */
/* -           meldet. Damit wird die SG-Liste nur dann berechnet, wenn dies zur Subsumption tatsaechlich noetig*/
/* -           wird, und dann gespeichert und das Flag auf 'aktuell' gesetzt. Wird nun das Ziel reduziert, so   */
/* -           wird nur das Flag geaendert, die SG-Liste jedoch nicht aktualisiert. Dies geschieht erst dann,   */
/* -           wenn das Ziel mit einer neuen Gleichung subsummiert werden soll. Damit werden insbesondere dann, */
/* -           wenn es keine Gleichungen im System gibt, keine Subgleichungen ueberfluessigerweise berechnet.   */
/* -------------------------------------------------------------------------------------------------------------*/

#if IO_AUSGABEN
#define DruckeMeldungGleichungWirdSubsummiertVon(/* long */ Nummer, /* TermT */ links, /* TermT */ rechts)   \
{                                                                                                               \
    if (IO_level3)                                                                                              \
       fprintf(IO_Strom, "... subsumed by equation no. %ld.\n", Nummer);                                        \
    else if (IO_level4){                                                                                        \
       fprintf(IO_Strom, "... subsumed by equation no. %ld:\n       ", Nummer);                                 \
       IO_DruckeFaktum(links, FALSE, rechts);                                                                   \
    }                                                                                                           \
 } 

#define DruckeMeldungGleichungWirdPermSubsummiert(/*unsigned short*/ Grad)   \
{                                                                               \
    if (IO_level3)                                                              \
       fprintf(IO_Strom, "... perm-subsumed (deg. %u).\n", Grad);               \
    else if (IO_level4)                                                         \
       fprintf(IO_Strom, "... perm-subsumed (deg. %u).\n", Grad);               \
 } 
#else
#define DruckeMeldungGleichungWirdSubsummiertVon(/*long*/ Nummer, /*TermT*/ links, /*TermT*/ rechts) /* nix tun */
#define DruckeMeldungGleichungWirdPermSubsummiert(/*unsigned short*/ Grad) /* nix tun */
#endif /* IO_AUSGABEN */


/* Das folgende Makro uebernimmt die folgenden Operationen auf der zu subsummierenden Gleichung:
   - Suche nach dem naechst-tieferen unterschiedlichen Teilterm,
   - Test der uebrigen Teilterme auf Gleichheit,
   - Falls die Unterschiede auf einen einzigen Teilterm beschraenkt bleiben, in diesen Teilterm absteigen.
   Voraussetzung ist, dass sich die beiden Terme links, rechts unterscheiden.
*/

#define Weiterschalten(links, rechts, teilterm1, teilterm2)                     \
{                                                                               \
   if (TO_TopSymboleGleich(links, rechts)){                                     \
      teilterm1 = TO_ErsterTeilterm(links);                                     \
      teilterm2 = TO_ErsterTeilterm(rechts);                                    \
      while (TO_TermeGleich(teilterm1, teilterm2)){                             \
         /* Ersten unterschiedlichen Teilterm suchen,                           \
            bricht auf jeden Fall ab, da die Terme ja unterschiedlich sind */   \
         teilterm1 = TO_NaechsterTeilterm(teilterm1);                           \
         teilterm2 = TO_NaechsterTeilterm(teilterm2);                           \
      }                                                                         \
      if (TO_NachfolgendeTeiltermeGleich(links, teilterm1, teilterm2)){         \
         links = teilterm1;                                                     \
         rechts = teilterm2;                                                    \
      }                                                                         \
      else{                /* mehr als ein unterschiedlicher Teilterm */        \
         return FALSE;                                                          \
      }                                                                         \
   }                                                                            \
   else{                  /* unterschiedliche Topsymbole */                     \
      return FALSE;                                                             \
   }                                                                            \
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Termpaar mit SG-Pufferung bzgl. GM/Permsub subsummieren ------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* TRUE, wenn die angegebene Gleichung links = rechts bzw. rechts = links von einer bereits vorhandenen 
   Gleichung oder einer Permgleichung vom Grad >= angegeb. Grad subsummiert wird. 
   MIT SUBGLEICHUNGEN PUFFERN.
   Vorbedingung: - links = rechts ist noch nicht in GM aufgenommen => kein Ausschluss zu beachten.
                 - links ist nicht identisch zu rechts
   Seiteneffekt: - Meldung, wenn Gleichung subummiert wird, und ggf. von welcher Gleichung.
                 - Bei Antwort FALSE koennen alle Subgleichungen von l = r in die Subgleichungsmenge aufgenommen werden
		   (da sie bereits berechnet und in den Puffer geschrieben wurden. Dies kann nicht in dieser Prozedur
                   erledigt werden, da die entsprechenden Subgleichngen an das Objekt Gleichung angehaengt werden muessen.).
*/

/* Nur besetzen der Puffer, wird beim Arbeiten mit MNF/SelectClassify benutzt */
/* gibt immer FALSE zurueck */
BOOLEAN SS_OnlyInitSGPuffer(TermT links, TermT rechts)
{
   TermT teilterm1, teilterm2;
   
   PU_SGPufferNeu();
   
   for (;;) {
      /* implicites return FALSE im Macro Weiterschalten */
      Weiterschalten(links, rechts, teilterm1, teilterm2);
      PU_SGEinfuegen(links, rechts);
   }

}

BOOLEAN SS_GleichungWirdSubsummiertMitSGPuffernGM(TermT links, TermT rechts)
{
   TermT teilterm1, teilterm2;

   if (RE_GleichungsmengeLeer())
      SS_OnlyInitSGPuffer(links,rechts);
   else {
     PU_SGPufferNeu();

     if (MO_SubsummierendeGleichungGefunden(links, rechts)){
       DruckeMeldungGleichungWirdSubsummiertVon(TP_ElternNr(MO_SubsummierendeGleichung),
                                                   TP_LinkeSeite(MO_SubsummierendeGleichung),
                                                   TP_RechteSeite(MO_SubsummierendeGleichung));
       return TRUE;
     }
     Weiterschalten(links, rechts, teilterm1, teilterm2);

     while (TRUE){              /* Abbruch gesichert, wg. l != r */
       if (MO_SubsummierendeGleichungGefunden(links, rechts)){
         DruckeMeldungGleichungWirdSubsummiertVon(TP_ElternNr(MO_SubsummierendeGleichung),
                                                     TP_LinkeSeite(MO_SubsummierendeGleichung),
                                                     TP_RechteSeite(MO_SubsummierendeGleichung));
         return TRUE;
       }
       else{
         PU_SGEinfuegen(links, rechts);
         Weiterschalten(links, rechts, teilterm1, teilterm2);

       }
     }
   }
   return FALSE;
}

BOOLEAN SS_GleichungWirdSubsummiertMitSGPuffernPermsub(TermT links, TermT rechts, unsigned short Grad)
{
  if (TO_ACnGleich(Grad,links,rechts)) {
    DruckeMeldungGleichungWirdPermSubsummiert(Grad);
    HK_fair = FALSE;
    return TRUE;
  }
  else { /* Schleife nur zur Berechnung und Pufferung der Subgleichungen */
    TermT teilterm1, teilterm2;
    PU_SGPufferNeu();
    Weiterschalten(links, rechts, teilterm1, teilterm2);
    do {              /* Abbruch gesichert, wg. l != r */
      PU_SGEinfuegen(links, rechts);
      Weiterschalten(links, rechts, teilterm1, teilterm2);
    } while (TRUE);
  }
}


BOOLEAN SS_GleichungWirdSubsummiertMitSGPuffernGMUndPermsub(TermT links, TermT rechts, unsigned short Grad)
{
  return SS_GleichungWirdSubsummiertMitSGPuffernGM(links,rechts) || SS_TermpaarSubsummiertVonPermsub (links,rechts,Grad);
}



/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Termpaar bzgl. GM subsummieren -------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* Liefert TRUE, wenn die Gleichungsmenge bzw. eine Permgleichung (Grad) das Termpaar links = rechts 
   subsummiert. Dabei werden alle Subgleichungen des Termpaares berechnet und getestet, und 
   zwar in beiden Richtungen, aber nicht gepuffert.*/

BOOLEAN SS_TermpaarSubsummiertVonPermsub (TermT links, TermT rechts, unsigned short Grad)
{
   if (TO_ACnGleich(Grad,links,rechts)) {
     DruckeMeldungGleichungWirdPermSubsummiert(Grad);
     HK_fair = FALSE;
     return TRUE;
   }
   return FALSE;                /* Meldung in uebergeordneter Prozedur */
}

BOOLEAN SS_TermpaarSubsummiertVonGM (TermT links, TermT rechts)
{
   TermT teilterm1, teilterm2;

   if (RE_GleichungsmengeNichtLeer()){
     while (TRUE){              /* Abbruch gesichert, wg. l != r */
       if (MO_SubsummierendeGleichungGefunden(links, rechts)){ /* Ueberpruefung in beiden Richtungen nicht noetig, da
                                                                  GM Stereogleichungen 2x enthaelt.*/
         return TRUE;
       }
       else
          Weiterschalten(links, rechts, teilterm1, teilterm2);
     }
   }
   return FALSE;                /* Meldung in uebergeordneter Prozedur */
}

BOOLEAN SS_TermpaarSubsummiertVonGMUndPermsub (TermT links, TermT rechts, unsigned short Grad)
{
  return SS_TermpaarSubsummiertVonGM(links,rechts) || SS_TermpaarSubsummiertVonPermsub(links,rechts,Grad);
}


/* -------------------------------------------------------------------------------------------------------------*/
/* ---------------------------- Termpaar bzgl. einzelner G subsummieren ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


BOOLEAN SS_GleichungSubsummiertTermpaar (GleichungsT gleichung, TermT links, TermT rechts)
/* Liefert TRUE, wenn die gleichung das Termpaar links, rechts subsummiert.
   Dabei werden alle Subgleichungen des Termpaar berechnet und getestet, und zwar in beiden Richtungen.*/
{
   TermT teilterm1, teilterm2;
   
   while (TRUE){                /* Abbruch gesichert, wg. l != r */
      if (MO_TermpaarSubsummiertZweites(TP_LinkeSeite(gleichung), TP_RechteSeite(gleichung), links, rechts) ||
          (TP_IstKeineMonogleichung(gleichung) && 
           MO_TermpaarSubsummiertZweites(TP_LinkeSeite(TP_Gegenrichtung(gleichung)), 
                                         TP_RechteSeite(TP_Gegenrichtung(gleichung)), 
                                         links, rechts))){
        return TRUE;
      }
      else
         Weiterschalten(links, rechts, teilterm1, teilterm2);
   }
   return FALSE;                /* Meldung in uebergeordneter Prozedur */
}



/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------ Termpaar bzgl. einzelner G subsummieren (G liegt als Termpaar vor) -----------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN SS_TermpaarSubsummiertTermpaar (TermT Glinks, TermT Grechts, TermT links, TermT rechts)
/* Liefert TRUE, wenn die gleichung das Termpaar links, rechts subsummiert.
   Dabei werden alle Subgleichungen des Termpaar berechnet und getestet, und zwar in beiden Richtungen.*/
{
   TermT teilterm1, teilterm2;
   
   while (TRUE){                /* Abbruch gesichert, wg. l != r */
      if (MO_TermpaarSubsummiertZweites(Glinks, Grechts, links, rechts) ||
          MO_TermpaarSubsummiertZweites(Grechts, Glinks, links, rechts)) {
        return TRUE;
      }
      else
         Weiterschalten(links, rechts, teilterm1, teilterm2);
   }
   return FALSE;                /* Meldung in uebergeordneter Prozedur */
}


