/****************************************************************************************************************/
/*                                                                                                              */
/*   pcl.c                                                                                                      */
/*   *****                                                                                                      */
/* -                                                                                                            */
/* - Enthaelt alle fuer den PCL-Modus notwendigen Routinen, Zaehler etc..                                       */
/* -                                                                                                            */
/* - ACHTUNG: Das Abklemmen des PCL-Modus bewirkt, dass - aehnlich wie im Modul                                 */
/* -          Ausgaben.h die Makros - die Funktionen auf 'Nix tun'-Makros  umgestellt werden. Es ist also       */
/* -          keine weitere Aenderung erforderlich. Ferner finden auch alle Abfragen innerhalb dieser Funktionen*/
/* -          statt, so dass an den betreffenden Stellen im Code nur der jeweilige Makro-Aufruf erscheint.      */
/* -                                                                                                            */
/* - 30.11.94 Arnim. Angelegt.                                                                                  */
/* - Mai 1996 Arnim. Chaining, Mikro-PCL eingebaut                                                              */
/* - Mai 1997 Bernd. iPCL                                                                                       */
/* - August 1998 Bernd: Slim down: nur noch normales PCL und iPCL.                                              */
/*                                                                                                              */
/****************************************************************************************************************/

#if COMP_PCL 

/* -------------------------------------------------------------------------------------------------------------
   Aufteilung in iPCL, sPCL und fPCL:
   - Es gibt folgende Modi:
     * den integrierten PCL-Modus (iPCL) einschliesslich Extraktion,
     * den standard PCL-Modus (sPCL) bei dem alles (bis auf triviale Ketten) rausgeschrieben wird
     * den full PCL-Modus (fPCL) bei dem wirklich alles rausgeschrieben wird
     * den abgeschalteten PCL-Modus.
   - im normalen PCL Modus werden die PCL-Schritte direkt rausgeschrieben
   - im iPCL-Modus werden die Rekonstruktionsdaten in ein dynamisches Array geschrieben. Am Ende wird
     ein extract auf dem Array gemacht und die benoetigten Inferenzen werden nachvollzogen.
   - Weiterhin koennen auf Wunsch ``Select''-Schritte erzeugt werden. Diese koennen z.Zt. nicht weiterver-
     arbeitet werden, sind aber zur manuellen Inspektion ganz nuetzlich
-------------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"
#include "ChaseManhattan.h"
#include "DSBaumOperationen.h"
#include "FehlerBehandlung.h"
#include "MatchOperationen.h"
#include "Parameter.h"
#include "SUEBasis.h"
#include "StrahlMemorium.h"
#include "TermOperationen.h"
#include "Unifikation1.h"
#include "compiler-extra.h"
#include "general.h"
#include "pcl.h"

#include <stdio.h>

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe pcl.h) -------------------------------*/
/* ------------------------------------------------------------------------*/

struct MeldesatzS PCL_Proc;
struct ChainsatzS PCL_Chain;

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Modul-lokale Variablen und Typen ------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

typedef enum {DummyMode, nPCLMode, iPCLMode} PCL_ModeT;
typedef enum {DummyChainMode, FULLChainMode} Chain_ModeT;

static PCL_ModeT PCL_Mode;
static Chain_ModeT Chain_Mode;
static BOOLEAN PCL_DoSelect;

typedef enum {noChain, inChain, inFeyChain} Chain_StatusT;
static Chain_StatusT ChainStatus;
static long ChainStartFilePos;
static PCL_ObjektNummerT ChainStartPclID;

static PCL_ObjektNummerT PCL_Objektzaehler;
static PCL_ObjektNummerT *AktuelleID;
static PCL_ObjektT AktuelleObjektArt;
static PCL_SeitenT AktuelleSeite;
static TermT AktuelleAndereSeite;

static char *ZieldateiName;
static FILE *Zieldatei;

typedef enum {
  iPCL_init, iPCL_hypo, iPCL_red, iPCL_cp, iPCL_orient, iPCL_select, iPCL_inst, iPCL_proved, iPCL_final, iPCL_conclusio, 
  iPCL_reconstructed
} iPCL_EntryType;

typedef enum {
  iPCL_eqn, iPCL_rule, iPCL_goal, iPCL_finalr
} iPCL_ResultType;

static const char *restpstr[] = {"eqn", "rule", "goal", "final"};

typedef union {
  struct {
    unsigned etp    : 4;          /* EntryType of Record */
    unsigned isUsed : 1;          /* used bit */
  } a;                            /* identical start segment of all alternatives */
  struct {
    unsigned etp    : 4;
    unsigned isUsed : 1;

    TermT lhs, rhs;
  } b;                            /* for init and hypo */
  struct {
    unsigned etp    : 4;
    unsigned isUsed : 1;

    unsigned pos    : 24;
    unsigned pside  : 1;
    unsigned side   : 1;
    PCL_ObjektNummerT id1, id2;
  } c;                           /* for red, cp and inst */
  struct {
    unsigned etp    : 4;
    unsigned isUsed : 1;

    unsigned heu    : 4;
    unsigned ux     : 1;
    PCL_ObjektNummerT id;
    long int nr;
  } d;                           /* for orient and select */
  struct {
    unsigned etp    : 4;
    unsigned isUsed : 1;

    PCL_ObjektNummerT id;
  } e;                           /* for proved and final */
  struct {
    unsigned etp    : 4;
    unsigned isUsed : 1;

    unsigned no     : 16;
    char *ts;
    char *text;
  } f;                           /* for conclusio */
  struct {
    unsigned etp    : 4;
    unsigned isUsed : 1;

    unsigned restp  : 2;
    TermT lhs, rhs;
  } r;                            /* for reconstructed terms */
} iPCL_Rec;

#define SetB(t,tp,ls,rs)        {t.a.etp = tp; t.a.isUsed = 0; t.b.lhs = ls; t.b.rhs = rs;}
#define SetC(t,tp,i1,i2,p,ps,s) {t.a.etp = tp; t.a.isUsed = 0; t.c.pos = p; t.c.pside = ps; t.c.side = s; t.c.id1 = i1; t.c.id2 = i2;}
#define SetD(t,tp,i,tw,h,n)     {t.a.etp = tp; t.a.isUsed = 0; t.d.heu = h; t.d.ux = tw; t.d.id = i; t.d.nr = n;}
#define SetE(t,tp,i)            {t.a.etp = tp; t.a.isUsed = 1; t.e.id = i;}
#define SetF(t,tp,terms,i,txt)  {t.a.etp = tp; t.a.isUsed = 0; t.f.ts = terms; t.f.no = i; t.f.text = txt;}
#define SetR(t,rt,ls,rs)        {t.a.etp = iPCL_reconstructed; t.a.isUsed = 1; t.r.restp = rt; t.r.lhs = ls; t.r.rhs = rs;}

static iPCL_Rec *iPCL_Memo;
static DomaenenT ipcl_Domaene;
static DomaenenT iPCL_StringBuffer;
static unsigned long *iPCL_UsedBuffer;

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CHAINING ------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* Chaining kommt in zwei Auspraegungen: Entweder es ist eingeschaltet oder komplett abgeklemmt (== Dummy_*Chain).
 * Ist es eingeschaltet, wird mittels StartChain ein Ruecksetzpunkt festgelegt. Wird danach DeleteChain
 * aufgerufen, wird das PCL-Protokoll auf diesen Punkt zurueckgesetzt. Mit SaveChain wird der Punkt geloescht.
 * Dies reduziert die Protokollgroesse z.T. erheblich, da die direkt zusammenfuehrbaren Kritischen Paare nicht mehr
 * das Protokoll aufblaehen. Werden die Terme implicit dargestellt, weiss man schon im Vorhinein, dass die
 * protokollierten Inferenzen nicht zum Beweis beitragen koennen. Deshalb gibt es "feychains". Bei StartFeyChain
 * werden dann die Protokollierungsfunktionen auf Nichtstun umgestellt. Dies spart nochmals erheblich an Zeit.
 * DeleteChain setzt dann die Zeiger wieder zurueck.
 *
 * (Eingeschaltetes) Chaining macht sich einen einfachen Endlichen Automaten mit drei Zustaenden zunutze.
 * Die vier Zugriffsfunktionen testen den Zustand des Automaten und fuehren dann neben ihren eigentlichen Aufgaben
 * noch einen Zustandsuebergang durch. Dieses Design ist erheblich einfacher, als das vorherige, das u.a. 
 * Funktionpointer verbogen hat.
 *
 *
 *     +----------+                         +----------+                        +------------+
 *     |          |	StartChain          |          |   StartFeyChain	|            |
 *     | inChain  |<------------------------| noChain  |----------------------->| inFeyChain |
 *     |          |		  	    |          |			|            |
 *     |          | SaveChain/DeleteChain   |          |   DeleteChain  	|            |
 *     |          |------------------------>|          |<-----------------------|            |
 *     +----------+			    +----------+			+------------+
 *
 *  Nicht eingetragene Uebergaenge fuehren zu einem Fehler!
 *
 * Bemerkung: Die Fehlerbehandlung (auch bei fseek/ftell) koennte man etwas defensiver gestalten, da nicht-essentielle
 * Programmteile betroffen sind.
 */

static void FunktionspointerBesetzen(PCL_ModeT Mode); /* forward declaration... */

static void FULL_StartFeyChain(void){
   if (ChainStatus != noChain) our_error("called PCL_StartFeyChain being in chain\n");
   FunktionspointerBesetzen(DummyMode);  /* PCL-Protokoll-Pointer abklemmen */
   ChainStatus = inFeyChain;
}

static void FULL_StartChain(void){
   if (ChainStatus != noChain) our_error(">>>>> ERROR: PCL_StartChain while being in chain\n");
   ChainStartPclID = PCL_Objektzaehler;                          /* Aktuelle Pcl-ID merken */
   if(PCL_Mode == nPCLMode) {
     ChainStartFilePos = ftell(Zieldatei); /* Position in Pcl-Datei merken */
     if (ChainStartFilePos == -1L) our_error(">>>>> Error in ftell!\n");
   }
   ChainStatus = inChain;
}

static void FULL_SaveChain(void){
   if (ChainStatus != inChain) our_error(">>>>> ERROR: PCL_SaveChain without being in chain\n");
   ChainStatus = noChain;
}

static void FULL_DeleteChain(void){
  switch (ChainStatus){
  case noChain:
    our_error(">>>>> ERROR: PCL_DeleteChain without being in chain\n");
    break;
  case inChain:
   PCL_Objektzaehler = ChainStartPclID; /* In StartChain gemerkte ID als aktuelle rueck-uebernehmen */
   if(PCL_Mode == nPCLMode){            /* Pcl-Datei zuruecksetzen */
     if (fseek(Zieldatei, ChainStartFilePos, SEEK_SET) != 0) our_error(">>>>> ERROR in fseek!\n");
   }
   break;
  case inFeyChain:
    FunktionspointerBesetzen(PCL_Mode); /* PCL-Protokoll-Pointer zuruecksetzen */
    break;
  default:
    our_error(">>>>> ERROR: PCL_DeleteChain: inconsistent chain status\n");
  }
  ChainStatus = noChain;
}

static void Dummy_StartChain(void)   {}
static void Dummy_StartFeyChain(void){}
static void Dummy_SaveChain(void)    {}
static void Dummy_DeleteChain(void)  {}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Gleichung/Ziel aus der Spezifikation --------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr){
  *nr = ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : initial\n", PCL_Objektzaehler, links, rechts);
}

static void iPCL_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr){
  *nr = ++PCL_Objektzaehler;
  SetB(iPCL_Memo[PCL_Objektzaehler], iPCL_init, TO_Termkopie(links), TO_Termkopie(rechts));
}

static void nPCL_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr){
  *nr = ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-goal : %t = %t : hypothesis\n", PCL_Objektzaehler, links, rechts);
}

static void iPCL_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr){
  *nr = ++PCL_Objektzaehler;
  SetB(iPCL_Memo[PCL_Objektzaehler], iPCL_hypo, TO_Termkopie(links), TO_Termkopie(rechts));
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------------- REDUKTION -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_Reduktionsbeginn(PCL_ObjektT objektArt, PCL_ObjektNummerT *nr, PCL_SeitenT seite, TermT andereSeite){
  AktuelleID          = nr;
  AktuelleObjektArt   = objektArt;
  AktuelleSeite       = seite;
  AktuelleAndereSeite = andereSeite;
}

static void iPCL_Reduktionsbeginn(PCL_ObjektT objektArt MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *nr,
				  PCL_SeitenT seite, TermT andereSeite MAYBE_UNUSEDPARAM){
  AktuelleID          = nr;
  AktuelleSeite       = seite;
}

static void nPCL_Reduktion(TermT ergebnis, UTermT betroffenerTeilterm, PCL_ObjektNummerT *AngewandtesFakt, PCL_SeitenT angewandteSeite){
  ++PCL_Objektzaehler;
  if ((AktuelleObjektArt == Regel) && (AktuelleSeite == linkeSeite)) AktuelleObjektArt = Gleichung;   /* Regel links reduziert!! */
  IO_DruckeFlexStrom(Zieldatei,"%u : tes-%s : %t %s %t : tes-red(%u,%p,%u,%q)\n",
		     PCL_Objektzaehler,
		     (AktuelleObjektArt == Regel) ? "rule" : (AktuelleObjektArt == Ziel) ? "goal" : "eqn",
		     (AktuelleSeite == linkeSeite) ? ergebnis : AktuelleAndereSeite,
		     (AktuelleObjektArt == Regel)  ? "->" : "=",
		     (AktuelleSeite == linkeSeite) ? AktuelleAndereSeite : ergebnis,
		     *AktuelleID, AktuelleSeite, ergebnis, betroffenerTeilterm,
		     *AngewandtesFakt, angewandteSeite);
  *AktuelleID = PCL_Objektzaehler;
}

static void iPCL_Reduktion(TermT ergebnis, UTermT betroffenerTeilterm, PCL_ObjektNummerT *AngewandtesFakt, PCL_SeitenT angewandteSeite){
  ++PCL_Objektzaehler;
  SetC(iPCL_Memo[PCL_Objektzaehler], iPCL_red, *AktuelleID, *AngewandtesFakt, 
       TO_StelleInTerm(betroffenerTeilterm,ergebnis), AktuelleSeite, angewandteSeite);
  *AktuelleID = PCL_Objektzaehler;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Unifikation ---------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_Unifikation(PCL_ObjektNummerT *nr, TermT links, TermT rechts, RegelOderGleichungsT Vater,
			     TermT betroffenerVaterTerm, UTermT betroffenerVaterTeilterm, RegelOderGleichungsT Mutter){
  *nr = ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : cp(%u,%p,%u,%q)\n",
		     *nr, links, rechts,
		     *(TP_PclID(Vater)),  (RE_IstRegel(Vater) || TP_RichtungAusgezeichnet(Vater))  ? linkeSeite : rechteSeite, 
		     betroffenerVaterTerm, betroffenerVaterTeilterm,
		     *(TP_PclID(Mutter)), (RE_IstRegel(Mutter) || TP_RichtungAusgezeichnet(Mutter))? linkeSeite : rechteSeite);
}

static void iPCL_Unifikation(PCL_ObjektNummerT *nr, TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM,
			     RegelOderGleichungsT Vater, TermT betroffenerVaterTerm, UTermT betroffenerVaterTeilterm, 
			     RegelOderGleichungsT Mutter){
  *nr = ++PCL_Objektzaehler;
  SetC(iPCL_Memo[PCL_Objektzaehler], iPCL_cp, *(TP_PclID(Vater)), *(TP_PclID(Mutter)),
       TO_StelleInTerm(betroffenerVaterTeilterm,betroffenerVaterTerm),
       (RE_IstRegel(Vater)  || TP_RichtungAusgezeichnet(Vater)) ? linkeSeite : rechteSeite,
       (RE_IstRegel(Mutter) || TP_RichtungAusgezeichnet(Mutter))? linkeSeite : rechteSeite);
}

static void nPCL_ReUnifikation(PCL_ObjektNummerT *nr, TermT links, TermT rechts,
			       RegelOderGleichungsT Vater, int NummerBetroffeneVaterStelle, RegelOderGleichungsT Mutter){
  TermT VaterLinkeSeite = TP_LinkeSeite(Vater);

  /*  NACHPRUEFEN: MUSS DA NICHT IN ETWA FOLGENDES HIN ????
   *  VaterLinkeSeite = (RE_IstRegel(Vater) || TP_RichtungAusgezeichnet(Vater)) ? TP_LinkeSeite(Vater) : TP_RechteSeite(Vater);
   */

  UTermT BetroffenerTeilterm = TO_TermAnStelle(VaterLinkeSeite, NummerBetroffeneVaterStelle);
  *nr = ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : cp(%u,%p,%u,%q)\n",
		     *nr, links, rechts,
		     *(TP_PclID(Vater)),  (RE_IstRegel(Vater) || TP_RichtungAusgezeichnet(Vater))  ? linkeSeite : rechteSeite, 
		     VaterLinkeSeite, BetroffenerTeilterm,
		     *(TP_PclID(Mutter)), (RE_IstRegel(Mutter) || TP_RichtungAusgezeichnet(Mutter))? linkeSeite : rechteSeite);
}

static void iPCL_ReUnifikation(PCL_ObjektNummerT *nr, TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM,
			       RegelOderGleichungsT Vater, int NummerBetroffeneVaterStelle, RegelOderGleichungsT Mutter){
  *nr = ++PCL_Objektzaehler;
  SetC(iPCL_Memo[PCL_Objektzaehler], iPCL_cp, *(TP_PclID(Vater)), *(TP_PclID(Mutter)),
       NummerBetroffeneVaterStelle,
       RE_IstRegel(Vater) ? linkeSeite : TP_RichtungAusgezeichnet(Vater) ? linkeSeite : rechteSeite,
       RE_IstRegel(Mutter)? linkeSeite : TP_RichtungAusgezeichnet(Mutter)? linkeSeite : rechteSeite);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Aufnahme von KPs als Regeln oder Gleichungen ------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_KPZuRegel(TermT links, TermT rechts, PCL_RichtungsRichtungT RichtungsRichtung, 
			   PCL_ObjektNummerT *RegelID, long int nr){
  ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-rule : %t -> %t : orient(%u,%s) ### %l%s\n",
		     PCL_Objektzaehler, links, rechts, *RegelID, (RichtungsRichtung == VonLinksNachRechts)?"u":"x",
		     nr, (SUE_LastSelectedWasFifo) ? " ## FIFO ##":"");
  *RegelID = PCL_Objektzaehler;
}

static void iPCL_KPZuRegel(TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM,
			   PCL_RichtungsRichtungT RichtungsRichtung, PCL_ObjektNummerT *RegelID, long int nr){
  ++PCL_Objektzaehler;
  SetD(iPCL_Memo[PCL_Objektzaehler], iPCL_orient, *RegelID, RichtungsRichtung, SUE_LastSelectedWasFifo,nr);
  *RegelID = PCL_Objektzaehler;
}

static void nPCL_KPZuGleichung(TermT links, TermT rechts, PCL_ObjektNummerT *GleichungsID, long int nr){
  if (PCL_DoSelect){
    ++PCL_Objektzaehler;
    IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : select(%u) ### %l%s\n",
		       PCL_Objektzaehler, links, rechts, *GleichungsID,
		       nr, (SUE_LastSelectedWasFifo) ? " ## FIFO ##":"");
    *GleichungsID = PCL_Objektzaehler;
  }
}

static void iPCL_KPZuGleichung(TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM, 
			       PCL_ObjektNummerT *GleichungsID, long int nr){
  if (PCL_DoSelect){
    ++PCL_Objektzaehler;
    SetD(iPCL_Memo[PCL_Objektzaehler], iPCL_select, *GleichungsID, 0, SUE_LastSelectedWasFifo, nr);
    *GleichungsID = PCL_Objektzaehler;
  }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Ziele subsummieren/ zusammenfuehren ---------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel, TermT links, TermT rechts,
				 PCL_ObjektNummerT *subsummierendeGleichung){
  ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-final : %t = %t : instance(%u,%u)\n",
		     PCL_Objektzaehler, links, rechts, *subsummiertesZiel, *subsummierendeGleichung);
  *subsummiertesZiel = PCL_Objektzaehler;
}

static void iPCL_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel, TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM,
				 PCL_ObjektNummerT *subsummierendeGleichung){
  ++PCL_Objektzaehler;
  SetC(iPCL_Memo[PCL_Objektzaehler], iPCL_inst, *subsummiertesZiel, *subsummierendeGleichung, 
       0, 0, 0); /* pos, pside, and side not yet supported by proof tools of Stephan! */
  iPCL_Memo[PCL_Objektzaehler].a.isUsed = 1;
  *subsummiertesZiel = PCL_Objektzaehler;
}

static void nPCL_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel, TermT links, TermT rechts){
  ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-final : %t = %t : %u\n",
		     PCL_Objektzaehler, links, rechts, *bewiesenesZiel);
  *bewiesenesZiel = PCL_Objektzaehler;
}

static void iPCL_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel, TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM){
  ++PCL_Objektzaehler;
  SetE(iPCL_Memo[PCL_Objektzaehler], iPCL_proved, *bewiesenesZiel);
  *bewiesenesZiel = PCL_Objektzaehler;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Abschluss der Vervollstaendigung: alle Fakten ausgeben --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art){
  ++PCL_Objektzaehler;
  IO_DruckeFlexStrom(Zieldatei, "%u : tes-final : %t %s %t : %u\n",
		     PCL_Objektzaehler, links, (art==Regel) ? "->" : "=", rechts, *nr);
  *nr = PCL_Objektzaehler;
}

static void iPCL_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM, 
					PCL_ObjektT art MAYBE_UNUSEDPARAM){
  ++PCL_Objektzaehler;
  SetE(iPCL_Memo[PCL_Objektzaehler], iPCL_final, *nr);
  *nr = PCL_Objektzaehler;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------------- DATENBANK -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static void nPCL_DatenbankKonklusion(RegelOderGleichungsT Konklusion){
  RegelOderGleichungsT Praemisse;

  *TP_PclID(Konklusion) = ++PCL_Objektzaehler;

  IO_DruckeFlexStrom(Zieldatei,"%u : tes-eqn : %t = %t : ",
		     *TP_PclID(Konklusion), TP_LinkeSeite(Konklusion), TP_RechteSeite(Konklusion));

  Praemisse = CM_ErstePraemisse(Konklusion);
  if (Praemisse != NULL) {
    fprintf(Zieldatei, "db_ceq(");
    do {
      IO_DruckeFlexStrom(Zieldatei, "db_ceq(%t = %t", TP_LinkeSeite(Praemisse), TP_RechteSeite(Praemisse));
      Praemisse = CM_NaechstePraemisse();
      if (Praemisse != NULL)
	fprintf(Zieldatei, ",");
    } while (Praemisse != NULL);
    IO_DruckeFlexStrom(Zieldatei, " : %t = %t ; ",
		       TP_LinkeSeite(CM_Konklusion(Konklusion)), TP_RechteSeite(CM_Konklusion(Konklusion)));
    for (Praemisse = CM_ErstePraemisseninstanz(Konklusion); Praemisse != NULL; ){
      fprintf(Zieldatei,"%lu", *TP_PclID(Praemisse));
      Praemisse = CM_NaechstePraemisseninstanz();

      if (Praemisse != NULL) fprintf(Zieldatei, ";");
    }
    fprintf(Zieldatei, ")\n");
  }
  else {
    fprintf(Zieldatei, "db_eq\n");
  }
}

static void iPCL_DatenbankKonklusion(RegelOderGleichungsT Konklusion){
  RegelOderGleichungsT Praemisse = CM_ErstePraemisse(Konklusion);
  char *buf;
  char *buf_start = iPCL_StringBuffer->Start;
  char *str;
  int i = 0;
  int size;
  TermT links = TO_Termkopie(TP_LinkeSeite(Konklusion)),
        rechts = TO_Termkopie(TP_RechteSeite(Konklusion));

  *TP_PclID(Konklusion) = ++PCL_Objektzaehler;


  if (iPCL_UsedBuffer == NULL){
    iPCL_UsedBuffer = our_alloc(2*sizeof(TermT)+CM_MaximaleAnzahlPraemissen()*sizeof(PCL_ObjektNummerT));
  }
  iPCL_UsedBuffer[0] = ptr2ul(links);
  iPCL_UsedBuffer[1] = ptr2ul(rechts);

  buf = buf_start;
  buf += our_sprintf((buf,"%lu : tes-eqn : ", PCL_Objektzaehler));
  buf += IO_DruckeTermString(buf, TP_LinkeSeite(Konklusion));
  buf += our_sprintf((buf, " = "));
  buf += IO_DruckeTermString(buf, TP_RechteSeite(Konklusion));
  buf += our_sprintf((buf, " : "));
  if (Praemisse) {
    buf += our_sprintf((buf, "db_ceq("));
    do {
      buf += IO_DruckeTermString(buf, TP_LinkeSeite(Praemisse));
      buf += our_sprintf((buf, " = "));
      buf += IO_DruckeTermString(buf, TP_RechteSeite(Praemisse));
      if ((Praemisse = CM_NaechstePraemisse()))
        buf += our_sprintf((buf, ","));
    } while (Praemisse);
    buf += our_sprintf((buf, " : "));
    buf += IO_DruckeTermString(buf, TP_LinkeSeite(CM_Konklusion(Konklusion)));
    buf += our_sprintf((buf, " = "));
    buf += IO_DruckeTermString(buf, TP_RechteSeite(CM_Konklusion(Konklusion)));
    buf += our_sprintf((buf, " ; "));
    for (Praemisse = CM_ErstePraemisseninstanz(Konklusion); Praemisse; ){
      iPCL_UsedBuffer[2+i] = *TP_PclID(Praemisse);
      i++;
      buf += our_sprintf((buf,"%lu", *TP_PclID(Praemisse)));
      if ((Praemisse = CM_NaechstePraemisseninstanz()))
        buf += our_sprintf((buf, ";"));
    }
    buf += our_sprintf((buf, ")\n"));
  }
  else
    buf += our_sprintf((buf, "db_eq\n"));


  str = our_alloc((buf - buf_start) + 1);
  strcpy(str, buf_start);
  size = 2*sizeof(TermT)+i*sizeof(PCL_ObjektNummerT);
  buf = our_alloc(size);
  memcpy(buf, iPCL_UsedBuffer, size);

  SetF(iPCL_Memo[PCL_Objektzaehler],iPCL_conclusio, buf, i, str);
}

static void Dummy_GleichungAusDerSpezifikation(TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM,
					       PCL_ObjektNummerT *nr MAYBE_UNUSEDPARAM)
{}

static void Dummy_ZielAusDerSpezifikation(TermT links MAYBE_UNUSEDPARAM, TermT rechts MAYBE_UNUSEDPARAM,
					  PCL_ObjektNummerT *nr MAYBE_UNUSEDPARAM)
{}

static void Dummy_Reduktionsbeginn(PCL_ObjektT a MAYBE_UNUSEDPARAM,PCL_ObjektNummerT *b MAYBE_UNUSEDPARAM,
				   PCL_SeitenT c MAYBE_UNUSEDPARAM,TermT d MAYBE_UNUSEDPARAM)
{}

static void Dummy_Reduktion(TermT a MAYBE_UNUSEDPARAM,UTermT b MAYBE_UNUSEDPARAM,PCL_ObjektNummerT *c MAYBE_UNUSEDPARAM,
			    PCL_SeitenT d MAYBE_UNUSEDPARAM)
{}

static void Dummy_Unifikation(PCL_ObjektNummerT *a MAYBE_UNUSEDPARAM,TermT b MAYBE_UNUSEDPARAM, TermT c MAYBE_UNUSEDPARAM,
			      RegelOderGleichungsT d MAYBE_UNUSEDPARAM,TermT e MAYBE_UNUSEDPARAM, UTermT f MAYBE_UNUSEDPARAM,
			      RegelOderGleichungsT g MAYBE_UNUSEDPARAM)
{}

static void Dummy_ReUnifikation(PCL_ObjektNummerT *a MAYBE_UNUSEDPARAM,TermT b MAYBE_UNUSEDPARAM, TermT c MAYBE_UNUSEDPARAM,
				RegelOderGleichungsT d MAYBE_UNUSEDPARAM, int e MAYBE_UNUSEDPARAM, 
				RegelOderGleichungsT f MAYBE_UNUSEDPARAM)
{}

static void Dummy_KPZuRegel(TermT a MAYBE_UNUSEDPARAM, TermT b MAYBE_UNUSEDPARAM, PCL_RichtungsRichtungT c MAYBE_UNUSEDPARAM,
			    PCL_ObjektNummerT *d MAYBE_UNUSEDPARAM, long int nr MAYBE_UNUSEDPARAM)
{}

static void Dummy_KPZuGleichung(TermT a MAYBE_UNUSEDPARAM, TermT b MAYBE_UNUSEDPARAM, 
                                PCL_ObjektNummerT *d MAYBE_UNUSEDPARAM, long int nr MAYBE_UNUSEDPARAM)
{}

static void Dummy_ZielSubsummiert(PCL_ObjektNummerT *a MAYBE_UNUSEDPARAM, TermT b MAYBE_UNUSEDPARAM, TermT c MAYBE_UNUSEDPARAM,
				  PCL_ObjektNummerT *d MAYBE_UNUSEDPARAM)
{}

static void Dummy_ZielBewiesen(PCL_ObjektNummerT *a MAYBE_UNUSEDPARAM, TermT b MAYBE_UNUSEDPARAM, TermT c MAYBE_UNUSEDPARAM)
{}

static void Dummy_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr MAYBE_UNUSEDPARAM, TermT links MAYBE_UNUSEDPARAM,
					 TermT rechts MAYBE_UNUSEDPARAM, PCL_ObjektT art MAYBE_UNUSEDPARAM)
{}

static void Dummy_DatenbankKonklusion(RegelOderGleichungsT Konklusion MAYBE_UNUSEDPARAM)
{}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Funktionspointer besetzen -------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define UnserKribskrabs(_) {                                                                                     \
  _##GleichungAusDerSpezifikation, _##ZielAusDerSpezifikation, _##Reduktionsbeginn,       _##Reduktion,          \
  _##Unifikation,                  _##ReUnifikation,           _##KPZuRegel,              _##KPZuGleichung,      \
  _##ZielSubsummiert,              _##ZielBewiesen,            _##VVAbschlussFaktDrucken, _##DatenbankKonklusion \
}

static void FunktionspointerBesetzen(PCL_ModeT Mode)
{
  static struct MeldesatzS PCL_Prozeduren[] = {
    UnserKribskrabs(Dummy_),
    UnserKribskrabs(nPCL_),
    UnserKribskrabs(iPCL_)
  };
  PCL_Proc = PCL_Prozeduren[Mode];
}


#define AlleKetten(_) {                                                                                          \
  _##StartChain, _##StartFeyChain, _##SaveChain, _##DeleteChain                                                  \
}

static void ChainingBesetzen(Chain_ModeT Mode)
{
  static struct ChainsatzS PCL_Chains[] = {
    AlleKetten(Dummy_),
    AlleKetten(FULL_)
  };
  PCL_Chain = PCL_Chains[Mode];
}


/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Initialisierung/Abschluss des Protokolls ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  char *selectArg = NULL;

  PCL_ParamInfo(ParamEvaluate, DEF, USR);

  if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "iPCL")) || IS_FindOption(DEF, USR, 1, "iPCL") == 0){
    PCL_Mode = iPCLMode; Chain_Mode = FULLChainMode;
  }
  else if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "sPCL"))){
    PCL_Mode = nPCLMode; Chain_Mode = FULLChainMode;
  }
  else if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "fPCL"))){
    PCL_Mode = nPCLMode; Chain_Mode = DummyChainMode;
  }
  else {
    PCL_Mode = DummyMode; Chain_Mode = DummyChainMode;
  }

  if (PCL_Mode != DummyMode){
    selectArg = IS_ValueOfOption(DEF, USR, "select");
    PCL_DoSelect = (strcmp(selectArg, "on") == 0);
	
    if (ZieldateiName != NULL){
      Zieldatei = fopen(ZieldateiName, "w");
      if (Zieldatei == NULL){
	if (PCL_Mode == iPCLMode){
	  fprintf(stderr, "\n >>> unable to open file %s, iPCL redirected to <stdout> <<<\n", ZieldateiName);
	  Zieldatei = stdout;
	}
	else {
	  fprintf(stderr, "\n >>> unable to open file %s, PCL turned off <<<\n", ZieldateiName);
	  PCL_Mode = DummyMode; Chain_Mode = DummyChainMode;
	}
      }
    }
    else { /* muss iPCL ohne Dateiname sein! */
      Zieldatei = stdout;
    }
    if (PCL_Mode == iPCLMode){
      ipcl_Domaene    = SM_NeueStrahldomaene(DM_iPCLMemo,10,"iPCL_Memo");
      iPCL_Memo       = ipcl_Domaene->Start;
      iPCL_StringBuffer = SM_NeueStrahldomaene(DM_iPCLBuffer,10,"iPCL_Buffer");
    }

    PCL_Objektzaehler = 0;
  }
  FunktionspointerBesetzen(PCL_Mode);
  ChainingBesetzen(Chain_Mode);
  ChainStatus = noChain;
}

BOOLEAN PCL_EsWirdProtokolliert(void)
{
  return PCL_Mode!=DummyMode;
}

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  char *arg = NULL;

  if (COMP_POWER) return;

  if (what == ParamEvaluate){
    if ((arg = IS_ValueOfOption(DEF, USR, "iPCL"))){
      fprintf(IO_Strom, "\n* PCL: write integrated PCL to %s", arg);
    }
    else if (IS_FindOption(DEF, USR, 1, "iPCL") == 0){
      fprintf(IO_Strom, "\n* PCL: write integrated PCL to <stdout>");
    }
    else if ((arg = IS_ValueOfOption(DEF, USR, "sPCL"))){
      fprintf(IO_Strom, "\n* PCL: write standard PCL (with chaining) to %s", arg);
    }
    else if ((arg = IS_ValueOfOption(DEF, USR, "fPCL"))){
      fprintf(IO_Strom, "\n* PCL: write full PCL (without chaining) to %s", arg);
    }
    else {
      fprintf(IO_Strom, "\n* PCL: off\n"); 
      
      return;
    }
    arg = IS_ValueOfOption(DEF, USR, "select");
    if (strcmp(arg,"on") == 0){
      fprintf (IO_Strom, "\n       (with protocol of select-steps!)\n");
    }
    else{
      fprintf (IO_Strom, "\n       (without protocol of select-steps)\n");
    }
  }
  else 
    fprintf (IO_Strom, 
	     "possible settings:\n"
	     " iPCL[=<file>]   write integrated PCL (i.e. extracted) to <file> or <stdout>\n"
	     " sPCL=<file>     write standard PCL (i.e. with chaining) to <file>\n"
             " fPCL=<file>     write full PCL (i.e. every step!) to <file> \n"
	     "\n"
	     " select=[on|off] write select steps (default: off)\n\n");
}

static void iPCL_Beenden(void)
{
  unsigned long i;
  int j, pos, side, pside ;
  TermT lhs, rhs;
  TermT l1,r1,l2,r2;
  UTermT actTermBefore;
  unsigned long *buf;
  char *str;
  iPCL_ResultType restp;

  /* Phase 1: downwards scanning & marking */

  for (i = PCL_Objektzaehler; i > 0; i--){
    if (iPCL_Memo[i].a.isUsed) {
      switch (iPCL_Memo[i].a.etp){
      case iPCL_init: 
      case iPCL_hypo:
        break;
      case iPCL_red: 
      case iPCL_cp:
      case iPCL_inst:
	iPCL_Memo[iPCL_Memo[i].c.id1].a.isUsed = 1;
	iPCL_Memo[iPCL_Memo[i].c.id2].a.isUsed = 1;
        break;
      case iPCL_orient: 
      case iPCL_select:
	iPCL_Memo[iPCL_Memo[i].d.id].a.isUsed = 1;
        break;
      case iPCL_proved:
      case iPCL_final:
	iPCL_Memo[iPCL_Memo[i].e.id].a.isUsed = 1;
        break;
      case iPCL_conclusio:
	buf = (unsigned long*) iPCL_Memo[i].f.ts;
	pos = iPCL_Memo[i].f.no;
	for (j=0; j < pos; j++){
	  iPCL_Memo[buf[2+j]].a.isUsed = 1;
	}
	break;
      default:
        our_error("reconstruction in iPCL failed because of unknown iPCL entry type");
      }
    }
  }

  /* Phase 2: upwards reconstructing with output */

  if (Zieldatei == stdout  && !PA_WWWAn){
    fprintf (IO_Strom, "\n********************************************************************************\n");
    fprintf (IO_Strom, "*                         P R O O F   P R O T O C O L                          *\n");
    fprintf (IO_Strom, "********************************************************************************\n\n");
  }

  for (i = 1; i <= PCL_Objektzaehler; i++){
    if (iPCL_Memo[i].a.isUsed){
      switch (iPCL_Memo[i].a.etp){
      case iPCL_init:
	lhs = iPCL_Memo[i].b.lhs;
	rhs = iPCL_Memo[i].b.rhs;
        IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : initial\n", i, lhs, rhs);
	SetR(iPCL_Memo[i],iPCL_eqn,lhs,rhs);
        break;
      case iPCL_hypo:
        lhs = iPCL_Memo[i].b.lhs;
	rhs = iPCL_Memo[i].b.rhs;
        IO_DruckeFlexStrom(Zieldatei, "%u : tes-goal : %t = %t : hypothesis\n", i, lhs, rhs);
	SetR(iPCL_Memo[i],iPCL_goal,lhs,rhs);
        break;
      case iPCL_red:
        lhs   = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].c.id1].r.lhs);
        rhs   = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].c.id1].r.rhs);
        pos   = iPCL_Memo[i].c.pos;
        pside = iPCL_Memo[i].c.pside;
        side  = iPCL_Memo[i].c.side;
	l1    = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].c.id2].r.lhs : iPCL_Memo[iPCL_Memo[i].c.id2].r.rhs;
	r1    = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].c.id2].r.rhs : iPCL_Memo[iPCL_Memo[i].c.id2].r.lhs;
        BO_TermpaarNormierenAlsKP(l1,r1);

	if (pos > 0){
	  actTermBefore = TO_TermAnStelle((pside == linkeSeite) ? lhs : rhs, pos - 1);
	  if (MO_TermpaarPasst(l1, r1, TO_Schwanz(actTermBefore)))
            TO_NachfBesetzen(actTermBefore, MO_SigmaRInLZ());
	  else
	    our_error("reconstruction of reduction in iPCL fails (case #1)");
	}
	else
	  if (MO_TermpaarPasst(l1, r1, (pside == linkeSeite) ? lhs : rhs))
	    MO_SigmaRInEZ();
	  else
	    our_error("reconstruction of reduction in iPCL fails (case #1)");

	restp = iPCL_Memo[iPCL_Memo[i].c.id1].r.restp;
        if ((restp == iPCL_rule) && (pside == linkeSeite)) restp = iPCL_eqn; /* Regel links reduziert ! */

	BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(Zieldatei, "%u : tes-%s : %t %s %t : tes-red(%u,%p,%u,%q)\n",
			   i, restpstr[restp], 
			   lhs, (restp == iPCL_rule) ? "->" : "=", rhs, 
			   iPCL_Memo[i].c.id1, 
			   pside, (pside == linkeSeite) ? lhs : rhs, TO_TermAnStelle ((pside == linkeSeite) ? lhs : rhs, pos),
			   iPCL_Memo[i].c.id2, side);
	SetR(iPCL_Memo[i],restp,lhs,rhs);
        break;
      case iPCL_cp:
	pos   = iPCL_Memo[i].c.pos;
        pside = iPCL_Memo[i].c.pside;
        side  = iPCL_Memo[i].c.side;

	l1 = (pside == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].c.id1].r.lhs : iPCL_Memo[iPCL_Memo[i].c.id1].r.rhs;
	r1 = (pside == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].c.id1].r.rhs : iPCL_Memo[iPCL_Memo[i].c.id1].r.lhs;
	BO_TermpaarNormierenAlsKP(l1, r1);

	l2 = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].c.id2].r.lhs : iPCL_Memo[iPCL_Memo[i].c.id2].r.rhs;
	r2 = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].c.id2].r.rhs : iPCL_Memo[iPCL_Memo[i].c.id2].r.lhs;

	U1_KPAusTermenRekonstruieren(l1, r1, l2, r2, pos, &lhs, &rhs);
	BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : cp(%u,%p,%u,%q)\n",
			   i, lhs, rhs, iPCL_Memo[i].c.id1,  pside, l1, TO_TermAnStelle(l1, pos),  iPCL_Memo[i].c.id2, side);
        SetR(iPCL_Memo[i],iPCL_eqn,lhs,rhs);
        break;
      case iPCL_orient:
        side = iPCL_Memo[i].d.ux;
	lhs = TO_Termkopie((side == UeberKreuz) ? iPCL_Memo[iPCL_Memo[i].d.id].r.rhs : iPCL_Memo[iPCL_Memo[i].d.id].r.lhs);
	rhs = TO_Termkopie((side == UeberKreuz) ? iPCL_Memo[iPCL_Memo[i].d.id].r.lhs : iPCL_Memo[iPCL_Memo[i].d.id].r.rhs);

        BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(Zieldatei, "%u : tes-rule : %t -> %t : orient(%u,%s) ### %l%s\n",
                           i, lhs, rhs, iPCL_Memo[i].d.id, (side == UeberKreuz) ? "x" : "u", 
			   iPCL_Memo[i].d.nr, iPCL_Memo[i].d.heu ? " ## FIFO ##" : "");
        SetR(iPCL_Memo[i],iPCL_rule,lhs,rhs);
        break;
      case iPCL_select:
	lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].d.id].r.lhs);
	rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].d.id].r.rhs);

        BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(Zieldatei, "%u : tes-eqn : %t = %t : select(%u) ### %l%s\n",
			   i, lhs, rhs,iPCL_Memo[i].d.id, 
			   iPCL_Memo[i].d.nr, iPCL_Memo[i].d.heu ? " ## FIFO ##" : "");
        SetR(iPCL_Memo[i],iPCL_eqn,lhs,rhs);
        break;
      case iPCL_inst:
        lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].c.id1].r.lhs);
        rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].c.id1].r.rhs);
        BO_TermpaarNormierenAlsKP(lhs, rhs);
        IO_DruckeFlexStrom(Zieldatei, "%u : tes-final : %t = %t : instance(%u,%u)\n",
			   i, lhs, rhs, iPCL_Memo[i].c.id1, iPCL_Memo[i].c.id2);
        SetR(iPCL_Memo[i],iPCL_finalr,lhs,rhs);
        break;
      case iPCL_proved:
        lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].e.id].r.lhs);
        rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].e.id].r.rhs);
        BO_TermpaarNormierenAlsKP(lhs, rhs);
        IO_DruckeFlexStrom(Zieldatei, "%u : tes-final : %t = %t : %u\n", i, lhs, rhs, iPCL_Memo[i].e.id);
        SetR(iPCL_Memo[i],iPCL_finalr,lhs,rhs);
        break;
      case iPCL_final:
        lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].e.id].r.lhs);
        rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].e.id].r.rhs);
        BO_TermpaarNormierenAlsKP(lhs, rhs);
        IO_DruckeFlexStrom(Zieldatei, "%u : tes-final : %t %s %t : %u\n", i, lhs, 
			   (iPCL_Memo[iPCL_Memo[i].d.id].r.restp == iPCL_rule) ? "->" : "=", rhs, iPCL_Memo[i].e.id);
        SetR(iPCL_Memo[i],iPCL_finalr,lhs,rhs);
        break;
      case iPCL_conclusio:
	str = iPCL_Memo[i].f.text;
	buf = (unsigned long*) iPCL_Memo[i].f.ts;

	lhs = (TermT) buf[0];
	rhs = (TermT) buf[1];
	fprintf(Zieldatei, str);

	our_dealloc(buf);
	our_dealloc(str);

	SetR(iPCL_Memo[i],iPCL_eqn,lhs,rhs);
        break;
      default:
        our_error("reconstruction in iPCL failed because of unexpected iPCL entry type");
      }
    }
  }

  /* Phase 3: dispose everything */

  for (i = PCL_Objektzaehler; i > 0; i--){
    if (iPCL_Memo[i].a.isUsed){
      TO_TermeLoeschen(iPCL_Memo[i].r.lhs,iPCL_Memo[i].r.rhs);
    }
    else if (iPCL_Memo[i].a.etp == iPCL_init || iPCL_Memo[i].a.etp == iPCL_hypo){
      TO_TermeLoeschen(iPCL_Memo[i].b.lhs,iPCL_Memo[i].b.rhs);
    }
    else if (iPCL_Memo[i].a.etp == iPCL_conclusio){
      buf = (unsigned long*) iPCL_Memo[i].f.ts;
      TO_TermeLoeschen((TermT) buf[0], (TermT) buf[1]);
      our_dealloc(iPCL_Memo[i].f.ts);
      our_dealloc(iPCL_Memo[i].f.text);
    }
  }
  DM_DomaeneLoeschen(ipcl_Domaene);
  DM_DomaeneLoeschen(iPCL_StringBuffer);

  if (Zieldatei != stdout) fclose (Zieldatei);
}

void PCL_Beenden(void)
/* PCL-Protokoll-Datei wieder schliessen.*/
{
  if (PCL_Mode == iPCLMode){
    iPCL_Beenden();
  }
  else {
   if (Zieldatei) {
      long Pos;
      Pos = ftell(Zieldatei);
      fclose(Zieldatei);
      truncate(ZieldateiName, Pos);
   }
  }
}

#endif
