/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFInnermost.c                                                                                              */
/* - *************                                                                                              */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (Leftmost-Innermost)                                       */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* - 20.04.94 Arnim. Fertiggemacht incl. Reduktion mit Gleichungen.                                             */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/
#include "general.h"
#include "TermOperationen.h"
#include "Parameter.h"
#include "NFBasis.h"
#include "pcl.h"
#include "NFInnermost.h"

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Initialisieren -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
void NFI_Init(void)
/* Die Initialisierung geschieht in Abhaengigkeit von der eingestellten NF-Strategie. Deshalb wird sie auch erst beim
   Initialisieren der NFBildung aufgerufen, und nicht direkt aus IN_AllesAprioriInit.*/
{
   NFB_Init();
}


static void InitialisierePfadAufTerm(TermT Term, ObjektT objekt)
/* Pfad auf Term initialisieren, dann Term 'an linkem Zweig' bis nach ganz unten absteigen, dabei den Pfad anlegen.*/
{
   NFB_PfadInit(Term,objekt);
   while (EsGibtNochZuBehNachf()){
      NFB_PfadEinsWeiterSetzen();
   }
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Hilfsprozeduren ----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static void SetzePfadAufNaechstenInNachOrdnung(void)
/* Setzt den aktuellen Pfad auf die naechste Stelle in Nach-Ordnung. Dabei wird vorausgesetzt,
   dass die akt. Stelle bereits erfolglos versucht wurde und nicht die Wurzel ist (dh. der Term nicht schon fertig
   durchlaufen ist.*/
{
   PfadEinsAuf();		/* Auf jeden Fall einen hochsteigen.(Da akt. Stelle erfolglos versucht wurde, muss
                                   ein Geschwister-Teilterm gesucht bzw. weiter aufgestiegen werden.*/
   if (EsGibtNochZuBehNachf()){ /* Geschwister-Teilterm gefunden => absteigen. */
      do{
	 NFB_PfadEinsWeiterSetzen();
      } while (EsGibtNochZuBehNachf());
   }
/* else{
      - Den Vorgaenger des zuletzt behandelten Termes behandeln -
      Dieser ist aber schon der, auf dem der Pfad jetzt steht, da bereits aufgestiegen wurde.
   }*/
}


      

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*    >>>Wird benoetigt, wenn das Regelsystem leer ist<<<                                                                  */
/* Bringt den uebergebenen Term bzgl. des aktuellen Gleichungssystems auf Normalform.                                      */
/* Diese Prozedur kommt zur Anwendung, wenn es keine Regeln im System gibt. Deshalb muss nicht jeder Teilterm des Anfrage- */
/* terms mit dem leeren Regelsystem 'versucht' werden; es kann vielmehr sofort ausschliesslich auf die Reduktion mit Glei- */
/* chungen umgeschaltet werden.                                                                                            */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFI_NormalformZuGleichungenDirekt (TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Gleichungssystems.
   Nach Reduktion wird sofort der betroffene Teilterm von vorne behandelt.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   if (RE_GleichungsmengeLeer()){
      return FALSE;
   }
   InitialisierePfadAufTerm(Term, objekt);                                        /* Pfad auf Term initialisieren */
   do{
      if (NFB_GleichungAngewendet()){	                                 /* Regel wurde angewendet => Teilterm von vorne */
	 while (EsGibtNochZuBehNachf()){
	    NFB_PfadEinsWeiterSetzen();
	 }
      }
      else if (Wurzel()){						  /* Erfolglos an Wurzel => Ende des Durchlaufes */
	 break;
      }
      else{
	 SetzePfadAufNaechstenInNachOrdnung();				  /* Sonst : Weiter in Nachordnung */
      }
   }while(TRUE);
   return PfadWurdeReduziert();                                           /* klar */
}

BOOLEAN NFI_NormalformZuGleichungenNochmal (TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Gleichungssystems.
   Nach Reduktion wird der betroffene Teilterm zunaechst fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   if (RE_GleichungsmengeLeer()){
      return FALSE;
   }
   InitialisierePfadAufTerm(Term, objekt);                                        /* Pfad auf Term initialisieren */
   do{
      if (NFB_GleichungAngewendet()){	                                 /* Gleichung wurde angewendet => */
         while(NFB_GleichungAngewendet());                               /*  => zunaechst fertigreduzieren, */
	 while (EsGibtNochZuBehNachf()){                                 /*     dann Teilterm von vorne */
	    NFB_PfadEinsWeiterSetzen();
	 }
      }
      else if (Wurzel()){						  /* Erfolglos an Wurzel => Ende des Durchlaufes */
	 break;
      }
      else{
	 SetzePfadAufNaechstenInNachOrdnung();				  /* Sonst : Weiter in Nachordnung */
      }
   }while(TRUE);
   return PfadWurdeReduziert();                                           /* klar */
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Bringt den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform.                                           */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFI_NormalformZuRegelnDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird sofort der betroffene Teilterm von vorne behandelt.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   InitialisierePfadAufTerm(Term, objekt);                               /* Pfad auf Term initialisieren und schon am
                                                                            linken Zweig entlang absteigen*/
   do{
/*    IO_DruckeFlex("Akt.Teilterm : %t\n",Pfad.EigentlicherPfad[Pfad.aktuelleStelle].Stelle);*/
      if (NFB_RegelAngewendet()){                                   /* Regel wurde angewendet => Teilterm von vorne */
	 while (EsGibtNochZuBehNachf()){
	    NFB_PfadEinsWeiterSetzen();
	 }
      }
      else if (Wurzel()){						  /* Erfolglos an Wurzel => Ende des Durchlaufes */
	 break;
      }
      else{
	 SetzePfadAufNaechstenInNachOrdnung();				  /* Sonst(dh. erfolglos an akt. Stelle und nicht Wurzel):
                                                                             Weiter in Nachordnung */
      }
   }while(TRUE);
   return PfadWurdeReduziert();
}

BOOLEAN NFI_NormalformZuRegelnNochmal(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird der betroffene Teilterm zunaechst fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   InitialisierePfadAufTerm(Term, objekt);                               /* Pfad auf Term initialisieren und schon am
                                                                            linken Zweig entlang absteigen*/
   do{
/*    IO_DruckeFlex("Akt.Teilterm : %t\n",Pfad.EigentlicherPfad[Pfad.aktuelleStelle].Stelle);*/
      if (NFB_RegelAngewendet()){                                   /* Regel wurde angewendet */
         while(NFB_RegelAngewendet());                                   /*  => zunaechst fertigreduzieren, */
	 while (EsGibtNochZuBehNachf()){                                 /*     dann Teilterm von vorne */
	    NFB_PfadEinsWeiterSetzen();
	 }
      }
      else if (Wurzel()){						  /* Erfolglos an Wurzel => Ende des Durchlaufes */
	 break;
      }
      else{
	 SetzePfadAufNaechstenInNachOrdnung();				  /* Sonst(dh. erfolglos an akt. Stelle und nicht Wurzel):
                                                                             Weiter in Nachordnung */
      }
   }while(TRUE);
   return PfadWurdeReduziert();
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Bringt den uebergebenen Term bzgl. des aktuellen Regel- und Gleichungssystems auf Normalform. Dazu wird zunaechst eine  */
/* 'abstrakte' Variable Pfad initialisiert, die dann gemaess der Leftmost-Innermost-Strategie das Durchlaufen des Termes   */
/* organisiert, wobei das weitere Vorgehen in Abhaengigkeit vom Erfolg der letzten versuchten Reduktion sein kann.         */
/* Dies wird solange fortgefuehrt, bis es keine reduzierbaren Stellen mehr im Anfrageterm gibt, wobei Gleichungen erst dann*/
/* auf einen Teilterm angewendet werden, wenn es keine anwendbaren Regeln mehr gibt.                                       */
/* Nach einem Reduktionsschritt muss der gerade behandelte Teilterm wieder von vorne durchlaufen werden, da es keine Moeg- */
/* lichkeit gibt, die bereits als irreduzibel erkannten Variablen-Substitute im reduzierten Term als solche zu erkennen    */
/* und ggf. zu markieren.                                                                                                  */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFI_NormalformBildenDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird sofort der betroffene Teilterm von vorne behandelt.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   NFB_PfadInit(Term, objekt);                                                         /* Pfad auf Term initialisieren */
   
   do{
/* IO_DruckeFlex("Akt.Teilterm : %t\n",Pfad.EigentlicherPfad[Pfad.aktuelleStelle].Stelle);*/
      if (NFB_RegelAngewendet() || NFB_GleichungAngewendet()){	          /* Regel o. Gleichung wurde angewendet => Teilterm von vorne */
	 while (EsGibtNochZuBehNachf()){
	    NFB_PfadEinsWeiterSetzen();
	 }
      }
      else if (Wurzel()){						  /* Erfolglos an Wurzel => Ende des Durchlaufes */
	 break;
      }
      else{
	 SetzePfadAufNaechstenInNachOrdnung();				  /* Sonst : Weiter in Nachordnung */
      }
   }while(TRUE);
   return PfadWurdeReduziert();                                           /* klar */
}

BOOLEAN NFI_NormalformBildenNochmal(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach Reduktion wird der betroffene Teilterm zunaechst fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   NFB_PfadInit(Term, objekt);                                                         /* Pfad auf Term initialisieren */
   
   do{
/* IO_DruckeFlex("Akt.Teilterm : %t\n",Pfad.EigentlicherPfad[Pfad.aktuelleStelle].Stelle);*/
      if (NFB_RegelAngewendet() || NFB_GleichungAngewendet()){	          /* Regel o. Gleichung wurde angewendet */
         while(NFB_RegelAngewendet() || NFB_GleichungAngewendet());       /*  => zunaechst fertigreduzieren, */
	 while (EsGibtNochZuBehNachf()){                                  /*     dann Teilterm von vorne */
	    NFB_PfadEinsWeiterSetzen();
	 }
      }
      else if (Wurzel()){						  /* Erfolglos an Wurzel => Ende des Durchlaufes */
	 break;
      }
      else{
	 SetzePfadAufNaechstenInNachOrdnung();				  /* Sonst : Weiter in Nachordnung */
      }
   }while(TRUE);
   return PfadWurdeReduziert();                                           /* klar */
}


