/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFTopDown.c                                                                                                */
/* - ***********                                                                                                */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (TopDown-Strategie)                                        */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "general.h"
#include "TermOperationen.h"
#include "Parameter.h"
#include "NFBasis.h"
#include "Ausgaben.h"
#include "limits.h"
#include "pcl.h"
#include "NFTopDown.h"

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Makros aus NFBasis etc. --------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static StellenT        letzteRed,         /* Verweis auf die letzte reduzierte Stelle   */
                       hoechsteRed;       /* Verweis auf die hoechste reduzierte Stelle */


#define PfadNeu(/*TermT*/ Term, /*ObjektT*/ objekt)\
{\
    NFB_PfadInit(Term, objekt);\
    letzteRed =0;\
    hoechsteRed = 0;\
    aktFrame=0;\
    FrameStack[0].oben=0;\
    FrameStack[0].unten= MaxStelle; /* "Au\3erster Frame ist 'bodenlos' */\
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------------------------------ Der Frame-Stack --------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

typedef struct{
   StellenT oben, unten;
} NFB_FrameZellenT;

NFB_FrameZellenT *FrameStack;

StellenT aktFrame,
         aktuelleGroesseFrameStack;

#define GipfelAktFrame() (FrameStack[aktFrame].oben)
#define BodenAktFrame() (FrameStack[aktFrame].unten)
#define aktSeiGipfelAktFrame() (aktuelleStelle = FrameStack[aktFrame].oben)
#define aktFrameWegwerfen() (aktFrame--)
#define /*BOOLEAN*/ LiegtOberhalbAktFrame(/*StellenT*/ tiefe) (tiefe<FrameStack[aktFrame].oben)
#define /*BOOLEAN*/ EsGibtUebergeordneteFrames() (aktFrame>0)

static void NFB_NeuenFrameAnlegen(StellenT oben, StellenT unten)
/* Legt einen neuen Frame auf dem Frame-Stack an. Vorausgesetzt wird bei dieser Prozedur, da\3 der neue Frame gem"a\3 Konven-
   tion im gerade aktuellen Frame enthalten ist.*/
{
   NFB_FrameZellenT *neuesArray;

   if (aktFrame == aktuelleGroesseFrameStack){                 /* Im Pfad kein Platz mehr zum Tiefersteigen => realloc versuchen*/
      aktuelleGroesseFrameStack += NF_PfadArrayGroesse;
      neuesArray = our_realloc(NFPfad, /* Dieses Stueck Code ist noch auf die neue our-realloc-Fkt. einzustellen. */
                               sizeof(NFB_FrameZellenT)*(aktuelleGroesseFrameStack+1), "Neuen Frame anlegen");
   }
   aktFrame++;
   FrameStack[aktFrame].oben = oben;
   FrameStack[aktFrame].unten = unten;
}
   

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Initialisieren -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN NFT_SchonInitialisiert = FALSE;
void NFT_Init(void)
/* Die Initialisierung geschieht in Abhaengigkeit von der eingestellten NF-Strategie. Deshalb wird sie auch erst beim
   Initialisieren der NFBildung aufgerufen, und nicht direkt aus IN_AllesAprioriInit.*/
{
   NFB_Init();
   if (!NFT_SchonInitialisiert){
     FrameStack = our_alloc (NF_PfadArrayGroesse*sizeof(NFB_FrameZellenT));
     aktuelleGroesseFrameStack = NF_PfadArrayGroesse -1;
   }
   NFT_SchonInitialisiert = TRUE;
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Termdurchlauf ------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/



static BOOLEAN WeiterenKnotenDerTiefeGefunden(StellenT Gipfel, StellenT Tiefe)
/* Sucht den naechsten Knoten der aktuellen Tiefe, steigt dabei aber nicht ueber Gipfel hinaus.
   Zurueckgegeben wird TRUE, wenn ein weiterer Knoten der angegebenen Tiefe gefunden wurde, aktStelle ist dann dieser Knoten.
   Wurde kein weiterer Knoten der aktuellen Tiefe gefunden, so steht aktStelle am Ende im Knoten Gipfel, der dann komplett
   abgearbeitet ist.
*/
{
   if (aktuelleStelle == Gipfel)      /* Es gibt nur eine Stelle in Wurzel-Tiefe: die Wurzel */
      return FALSE;
   do{
      do                                                     /* Mindestens einen aufsteigen */
         PfadEinsAuf();
      while( (aktuelleStelle != Gipfel) && (EsGibtKeineZuBehNachf()));
      if (EsGibtKeineZuBehNachf())                           /* Gipfel erreicht, aber keine Nachfolger mehr gefunden */
         return FALSE;
      do                                                     /* Abwaerts, bis Tiefe oder Term-Boden erreicht */
         NFB_PfadEinsWeiterSetzen();
      while((EsGibtNochZuBehNachf()) && (aktuelleStelle<Tiefe));
   }while (aktuelleStelle< Tiefe);
   return TRUE;
}


static BOOLEAN ErstenKnotenDerTiefeGefunden(StellenT Gipfel, StellenT tiefe)
/* Sucht den ersten Knoten der Tiefe tiefe im in Gipfel beginnenden Teilterm. Falls ein solcher Knoten gefunden
   wurde, ist aktStelle entsprechend gesetzt, andernfalls steht aktStelle auf Gipfel, der dann komplett abgearbeitet
   ist.
*/
{
   aktStelleSei(Gipfel);                                      /* Beginne in Gipfel, diesen Teilterm von vorne behandeln */
   aktStelleVonVorne();
   while (aktuelleStelle < tiefe){
      if (EsGibtNochZuBehNachf()){                            /* Einen absteigen, wenn moeglich */
         NFB_PfadEinsWeiterSetzen();
      }
      else if (aktuelleStelle == Gipfel){                             /* Am Gipfel schon kein Abstieg mehr moeglich => tiefe nicht erreichbar */
         return FALSE;
      }
      else{
         do{                                                  /* Aufsteigen soweit wie noetig */
            PfadEinsAuf();
         }while((aktuelleStelle != Gipfel) && (EsGibtKeineZuBehNachf()));
         if (EsGibtKeineZuBehNachf()){
            return FALSE;
         }
      }
   }
   return TRUE;
}


static BOOLEAN EsGibtNaechstenImAktFrameAbTiefe(StellenT tiefe)
/* Liefert TRUE, wenn es im aktuellen Frame noch nicht abgearbeitete Stellen gibt, wobei die Suche an der aktuellen Stelle
   begonnen wird und 'tiefe' die Tiefe der zuletzt behandelten Stelle ist. Wenn es im akt. Frame eine weitere Stelle dieser
   Tiefe gibt, wird aktStelle dorthin gesetzt, andernfalls wird im akt. Frame eine Stelle auf einer tieferen Ebene gesucht.
   War auch dies erfolglos, so wird FALSE zurueckgeliefert; aktStelle steht dann in der Wurzel des aktuellen Frames.
*/
{
   if (WeiterenKnotenDerTiefeGefunden(GipfelAktFrame(), tiefe)){
      return TRUE;
   }
   else if (tiefe == BodenAktFrame()){
      return FALSE;
   }
   else{
      return ErstenKnotenDerTiefeGefunden(GipfelAktFrame(), tiefe+1);
   }
}

static BOOLEAN NaechsterTopDown(void)
{
   StellenT aktTiefe = aktuelleStelle;
   BOOLEAN weiter;
   
   while( !(weiter = EsGibtNaechstenImAktFrameAbTiefe(aktTiefe))
         && (EsGibtUebergeordneteFrames())){
      aktTiefe = BodenAktFrame();
      aktFrameWegwerfen();
   }
   return weiter;
}



/*static void PipiPfad()
{
   int i;
   fprintf(IO_Strom, "akt Stelle : %d\n", aktuelleStelle);
   fprintf(IO_Strom, "Reduziert  : %s\n", (Reduziert)? "Ja":"Nein");
   fprintf(IO_Strom, "Gipfel     : %d\n", Gipfel);
   fprintf(IO_Strom, "LetzteRed  : %d\n", letzteRed);
   fprintf(IO_Strom, "HoechsteRed: %d\n", hoechsteRed);
   i=0;
   do{
      IO_DruckeFlex("Tiefe %d : %t. AnzTterme: %d",i, EigentlicherPfad[i].Stelle, EigentlicherPfad[i].AnzNochZuBehandelnderTeilterme);
      if (EigentlicherPfad[i].EinfuegePunkt){
         IO_DruckeFlex("EinfuegePunkt: %t", EigentlicherPfad[i].EinfuegePunkt);
      }
      i++;
      IO_LF();
   }while(i<aktuelleStelle);
}*/

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Regeln ---------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zur Wurzel auf Reduzierbarkeit geprueft und ggf.      */
/* reduziert.                                                                                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN GAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (NFB_GleichungAngewendet())
         return TRUE;
   }
   return FALSE;
}
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN GAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop'
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (NFB_GleichungAngewendet())
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Glgs.systems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Top-Down-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste             */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFT_NormalformZuGleichungenAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   StellenT u;

   PfadNeu(Term, objekt);
   while (NFB_GleichungAngewendet());                           /* Es gibt keine leeren Terme, also mind. ein Red.versuch */

   while(NaechsterTopDown()){
      if (NFB_GleichungAngewendet()){
         while(NFB_GleichungAngewendet());
         u = letzteRed = aktuelleStelle;
         while (GAlleAufsteigenBisReduziertOderStop(letzteRed, 0)){
            while(NFB_GleichungAngewendet());
            letzteRed = aktuelleStelle;
         }
         if (letzteRed == u)
            aktuelleStelle = u;
         else{
            while(LiegtOberhalbAktFrame(letzteRed)){ /* => alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()){ /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed,u); /* sonst: neuen Frame anlegen */
            }
            aktSeiGipfelAktFrame(); /* Dies ist die Stelle der hoechsten Reduktion und damit bereits zuende reduziert, */
         }                          /* braucht also nicht weiter behandelt zu werden */
      }
   }
   return PfadWurdeReduziert(); 
}

BOOLEAN NFT_NormalformZuGleichungenAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg */
{
   BOOLEAN weiter;
   StellenT u;

   PfadNeu(Term, objekt);
   while (NFB_GleichungAngewendet());
   weiter =  NaechsterTopDown();
   while(weiter){
      if (NFB_GleichungAngewendet()){
         letzteRed = u = aktuelleStelle;        /* Stelle der Reduktion merken, hoechste red. Stelle dito */
         while(GAlleAufsteigenBisReduziertOderStop(letzteRed, 0))
            letzteRed = aktuelleStelle;
         if (letzteRed == u)
            aktuelleStelle = u;
         else{
            while(LiegtOberhalbAktFrame(letzteRed)){ /* Sonst alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()){ /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed,u);
            }                          /* sonst: neuen Frame anlegen */
            aktSeiGipfelAktFrame();
         }
      }
      else
         weiter = NaechsterTopDown(); /* Gibt es noch eine unbehandelte Stelle im Term? */
   }
   return PfadWurdeReduziert();                                           /* klar */
}

   
            
BOOLEAN NFT_NormalformZuGleichungenAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
{
   StellenT u;

   PfadNeu(Term, objekt);
   do{                          /* Leere Terme gibt es nicht => mind. ein Versuch */
      if (NFB_GleichungAngewendet()){
         while (NFB_GleichungAngewendet());
         letzteRed = u = aktuelleStelle;        /* Stelle der Reduktion merken, hoechste red. Stelle dito */
         while (GAlleAbsteigenBisReduziertOderStop(0, letzteRed)){
            while(NFB_GleichungAngewendet());
            letzteRed = aktuelleStelle;
         }
         if (letzteRed != u){   /* Beim Aufstieg weitere Reduktionen */
            while(LiegtOberhalbAktFrame(letzteRed)){ /* Sonst alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()){ /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed, u);
            }                          /* sonst: neuen Frame anlegen */
            aktSeiGipfelAktFrame();
         }
      }
   }while(NaechsterTopDown());
   return PfadWurdeReduziert();                                           /* klar */
}

         
BOOLEAN NFT_NormalformZuGleichungenAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   StellenT u;
   BOOLEAN nochmal;

   PfadNeu(Term, objekt);
   do{                          /* Leere Terme gibt es nicht => mind. ein Versuch */
      if (NFB_GleichungAngewendet()){
         u = aktuelleStelle;    /* Stelle der Reduktion merken, hoechste red. Stelle dito */
         letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while (GAlleAbsteigenBisReduziertOderStop(0, letzteRed))
            letzteRed = aktuelleStelle;
         if (letzteRed == u)    /* Keine weitere Reduktion bei Aufstieg */
            aktuelleStelle = u; /* letzte Stelle nochmal */
         else{
            while(LiegtOberhalbAktFrame(letzteRed)){ /* Sonst alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()) /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed, u);
            aktSeiGipfelAktFrame();
         }
      }
      else
         nochmal = NaechsterTopDown();
   }while(nochmal);
   return PfadWurdeReduziert();                                           /* klar */
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Regeln ---------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zur Wurzel auf Reduzierbarkeit geprueft und ggf.      */
/* reduziert.                                                                                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN RAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (NFB_RegelAngewendet())
         return TRUE;
   }
   return FALSE;
}
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN RAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop'
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (NFB_RegelAngewendet())
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Top-Down-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste             */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFT_NormalformZuRegelnAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   StellenT u;

   PfadNeu(Term, objekt);
   while (NFB_RegelAngewendet());                           /* Es gibt keine leeren Terme, also mind. ein Red.versuch */

   while(NaechsterTopDown()){
      if (NFB_RegelAngewendet()){
         while(NFB_RegelAngewendet());
         u = letzteRed = aktuelleStelle;
         while (RAlleAufsteigenBisReduziertOderStop(letzteRed, 0)){
            while(NFB_RegelAngewendet());
            letzteRed = aktuelleStelle;
         }
         if (letzteRed == u)
            aktuelleStelle = u;
         else{
            while(LiegtOberhalbAktFrame(letzteRed)){ /* => alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()){ /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed,u); /* sonst: neuen Frame anlegen */
            }
            aktSeiGipfelAktFrame(); /* Dies ist die Stelle der hoechsten Reduktion und damit bereits zuende reduziert, */
         }                          /* braucht also nicht weiter behandelt zu werden */
      }
   }
   return PfadWurdeReduziert(); 
}

BOOLEAN NFT_NormalformZuRegelnAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg */
{
   BOOLEAN weiter;
   StellenT u;

   PfadNeu(Term, objekt);
   while (NFB_RegelAngewendet());
   weiter =  NaechsterTopDown();
   while(weiter){
      if (NFB_RegelAngewendet()){
         letzteRed = u = aktuelleStelle;        /* Stelle der Reduktion merken, hoechste red. Stelle dito */
         while(RAlleAufsteigenBisReduziertOderStop(letzteRed, 0))
            letzteRed = aktuelleStelle;
         if (letzteRed == u)
            aktuelleStelle = u;
         else{
            while(LiegtOberhalbAktFrame(letzteRed)){ /* Sonst alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()){ /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed,u);
            }                          /* sonst: neuen Frame anlegen */
            aktSeiGipfelAktFrame();
         }
      }
      else
         weiter = NaechsterTopDown(); /* Gibt es noch eine unbehandelte Stelle im Term? */
   }
   return PfadWurdeReduziert();                                           /* klar */
}

   
            
BOOLEAN NFT_NormalformZuRegelnAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
{
   StellenT u;

   PfadNeu(Term, objekt);
   do{                          /* Leere Terme gibt es nicht => mind. ein Versuch */
      if (NFB_RegelAngewendet()){
         while (NFB_RegelAngewendet());
         letzteRed = u = aktuelleStelle;        /* Stelle der Reduktion merken, hoechste red. Stelle dito */
         while (RAlleAbsteigenBisReduziertOderStop(0, letzteRed)){
            while(NFB_RegelAngewendet());
            letzteRed = aktuelleStelle;
         }
         if (letzteRed != u){   /* Beim Aufstieg weitere Reduktionen */
            while(LiegtOberhalbAktFrame(letzteRed)){ /* Sonst alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()){ /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed, u);
            }                          /* sonst: neuen Frame anlegen */
            aktSeiGipfelAktFrame();
         }
      }
   }while(NaechsterTopDown());
   return PfadWurdeReduziert();                                           /* klar */
}

         
BOOLEAN NFT_NormalformZuRegelnAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   StellenT u;
   BOOLEAN nochmal;

   PfadNeu(Term, objekt);
   do{                          /* Leere Terme gibt es nicht => mind. ein Versuch */
      if (NFB_RegelAngewendet()){
         u = aktuelleStelle;    /* Stelle der Reduktion merken, hoechste red. Stelle dito */
         letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while (RAlleAbsteigenBisReduziertOderStop(0, letzteRed))
            letzteRed = aktuelleStelle;
         if (letzteRed == u)    /* Keine weitere Reduktion bei Aufstieg */
            aktuelleStelle = u; /* letzte Stelle nochmal */
         else{
            while(LiegtOberhalbAktFrame(letzteRed)){ /* Sonst alle Frames wegwerfen, die innerhalb des hoechsten red. Teilterms liegen. */
               u = BodenAktFrame();
               aktFrameWegwerfen();
            }
            if (letzteRed !=GipfelAktFrame()) /* Hoechster red. Teilterm entspricht gerade einem ex. Frame => diesen von vorne durchlaufen */
               NFB_NeuenFrameAnlegen(letzteRed, u);
            aktSeiGipfelAktFrame();
         }
      }
      else
         nochmal = NaechsterTopDown();
   }while(nochmal);
   return PfadWurdeReduziert();                                           /* klar */
}
/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den  Term bzgl. des akt. Regel- und Gleichungssystems auf Normalform. Dabei wird aber  */
/* jeweils eine andere Variante der Top-Down-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste             */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* Es wird erst dann versucht, Gleichungen anzuwenden, wenn es keine anwendbaren Regeln mehr gibt.                         */
/* In diesem Fall wird versucht, eine anwendbare Gleichung zu finden. Gelingt dies, so wird die Gleichung angewendet, und  */
/* anschliessend wird an dieser Stelle mit Regeln fortgefahren. Dies wiederholt sich, bis es weder anwendbare Regeln noch  */
/* Gleichungen mehr gibt.                                                                                                  */
/*-------------------------------------------------------------------------------------------------------------------------*/


BOOLEAN NFT_NormalformAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   BOOLEAN reduziertG;

   NFB_Reduziert = FALSE;
   do{
      NFT_NormalformZuRegelnAufNochmal(Term, objekt);
      PfadNeu(Term, objekt);
         reduziertG = FALSE;
      do
         reduziertG = NFB_GleichungAngewendet();
      while(!reduziertG && NaechsterTopDown());
   }while(reduziertG);
   return NFB_Reduziert;
}

BOOLEAN NFT_NormalformAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg */
{
   BOOLEAN reduziertG;

   NFB_Reduziert = FALSE;
   do{
      NFT_NormalformZuRegelnAufDirekt(Term, objekt);
      PfadNeu(Term, objekt);
         reduziertG = FALSE;
      do
         reduziertG = NFB_GleichungAngewendet();
      while(!reduziertG && NaechsterTopDown());
   }while(reduziertG);
   return NFB_Reduziert;
}

BOOLEAN NFT_NormalformAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
{
   BOOLEAN reduziertG;

   NFB_Reduziert = FALSE;
   do{
      NFT_NormalformZuRegelnAbNochmal(Term, objekt);
      PfadNeu(Term, objekt);
         reduziertG = FALSE;
      do
         reduziertG = NFB_GleichungAngewendet();
      while(!reduziertG && NaechsterTopDown());
   }while(reduziertG);
   return NFB_Reduziert;
}

BOOLEAN NFT_NormalformAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN reduziertG;

   NFB_Reduziert = FALSE;
   do{
      NFT_NormalformZuRegelnAbDirekt(Term, objekt);
      PfadNeu(Term, objekt);
         reduziertG = FALSE;
      do
         reduziertG = NFB_GleichungAngewendet();
      while(!reduziertG && NaechsterTopDown());
   }while(reduziertG);
   return NFB_Reduziert;
}

