/*----------------------------------------------------------------------------------------------------
  HEAP.
  
  KP-Mengen-Verwaltung mit einem normalen Heap (ARRAY).
  Keine Sonderfunktionen wie Waisenmord etc., dafuer aber kein zusaetzlicher Verwaltungsaufwand und
  Platzbedarf ausser dem Array.

  11.03.94. Angelegt. Arnim.
  14.03.95. Heap und KPHeap zusammengelegt.

Vorgehensweise:

 Sortierreihenfolge : kleinstes Element am Anfang
 Entnehmen          : Falls Heap-Flag ok, erstes Element entnehmen, Flag auf nicht-ok setzen
                        sonst Heap reorganisieren, erstes Element entnehmen, Flag bleibt nicht-ok
 Einfuegen          : Falls Flag ok, am Ende anhaengen, reorganisieren
                        sonst am Anfang einsetzen, einsinken lassen.

 Zu jedem Heap werden zwei Verwaltungsinformationen gespeichert, die seltener eine Reorganisation
 noetig machen. Zum einen gibt es einen Eintrag 'letztesBesetztesElement', zum anderen ein Flag
 mit Namen 'HeapIstHeap'. Dieses Flag gibt an, ob der Heap in reorganisierter Form vorliegt, oder
 ob nach der letzten Entnahme eines Elementes die oberste Stelle verwaist geblieben ist.
 Dies hat den Vorteil, dass bei einer auf die Entnahme folgenden Einfuege-Operation eine Reorgani-
 sation gespart werden kann (naemlich die Reorganisation nach der Entnahme). Je nach Wert dieses
 Flags kommt dem Verweis 'letztesBesetztesElement' aber eine unterschiedliche Bedeutung zu:

 Ist das Flag TRUE, so wird tatsaechlich auf das letzte besetzte Element verwiesen.
 Ist das Flag aber FALSE, so zeigt der Verweis zwar immer noch auf das letzte besetzte Element,
     an der ersten Stelle im Heap klafft aber eine Luecke. Deshalb wird in dieser Situation nach
     dem naechsten Einfuegen der Verweis nicht erhoeht!


 Ein Sonderfall tritt in diesem Zusammenhang auf:
   Ist der Heap leer, so ist das Flag FALSE, und der Verweis zeigt auf das nullte Element.

+--------------------------------------------+
|             IstHeap       LetztesBesetztes |
+--------------------------------------------+
|leer             F                0         |
|1 Elt auf 0      T                0         |
|1 Elt auf 1      F                1         |
|n ELt ab 0       T                n         |
|n ELt ab 1       F               n+1        |
+--------------------------------------------+

 Daraus ergibt sich folgender Grob-Algorithmus:

 Einfuegen:
 - Flag = TRUE: + Verweis++
                + Heap[Verweis] = neues Element;
                + Position (Verweis div 2) rekursiv einsinken lassen

        = FALSE:+ Heap[0] = neues Element
                + Position 0 einsinken lassen
                + Flag = TRUE

 Entnehmen:
 - Flag = TRUE: + return Heap[0]
                + Flag = FALSE

        = FALSE:+ if Verweis==0 => return NULL (Heap ist leer)
                + Heap[0] = Heap[Verweis]
                + Verweis--
                + Position 0 Einsinken lassen
                + return Heap[0]
*/

/*-----------------------------------------------------------------------------*/
/*------------------------------ Includes -------------------------------------*/
/*-----------------------------------------------------------------------------*/



#include "Achtzigstel.h"
#include "antiwarnings.h"
#include "Ausgaben.h"
#include "compiler-extra.h"
#include "general.h"
#include "KPHeap.h"
#include <limits.h>  /* fuer LONG_MAX */
#include "NewClassification.h"
#include "Parameter.h"
#include "ParseInterleave.h"
#include "PQStatistik.h"
#include "SUEBasis.h"
#include "SUEM.h"
#include "SpeicherVerwaltung.h"
#include <stdlib.h>
#include "StrahlMemorium.h"
#include <string.h>
#include "TermpairRepresentation.h"

#if defined SV_BASIS_MMAP
#define KPH_USE_MMAP 1
#else
#define KPH_USE_MMAP 0
#endif

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe KPHeap.h) ----------------------------*/
/* ------------------------------------------------------------------------*/

void (* KPH_INSERTImmortal)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);

/*-----------------------------------------------------------------------------*/
/*------------------------------ Typen ----------------------------------------*/
/*-----------------------------------------------------------------------------*/

typedef struct InfoTS{          /* Info-Block */
   long int ElternNummer;
   unsigned long AnzNochNichtBehandelterKPs;
   RegelOderGleichungsT re;
   BOOLEAN LebtNoch;
#if PQ_STATISTIK
  unsigned long AnzInitialerKPs;
#endif
} InfoT;


typedef struct{                 /* Eintrag im Heap: 2 long als Gewicht */
   InfoT *Vater, *Mutter;
   TermpairT tp;
   WeightT w1, w2;
} HeapZellenT;

static HeapZellenT   *Elemente;    /* Das Array mit den eigentlichen Eintraegen.*/


static BOOLEAN IstHeap;                 /* Zeigt an, ob die Struktur z.Zt. als Heap organisiert ist
                                           oder beim naechsten Zugriff reorganisiert werden muss:
                                           T : besetzt an Positionen 0..LetztesBesetztes.
                                           F : besetzt an Positionen 1..(LetztesBesetztes+1).
                                           (LetztesBesetztes == 0 bedeutet Heap ist leer)
                                         */

static unsigned long  LetztesBesetztes;


static unsigned long DomainGran;
#if KPH_USE_MMAP
static DomaenenT Domaene;
#else
static unsigned int AllocElemente;
#endif

/* fuer Interleaving der Queues */
static unsigned int modulo;
static unsigned int threshold;

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Testing ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
#define KPH_TEST 0
/* Alternativen: 0 fuer aus, 1 fuer an */
#if KPH_TEST
#if PQ_STATISTIK == 0
#error "fuer KPH_TEST muss PQ_STATISTIK gesetzt sein"
#endif
static unsigned long int testNr = 0;
static BOOLEAN ok = TRUE;

#define HeapFehler(str, nr1, nr2) (ok = FALSE, printf("\n%ld: %s (%ld, %ld)\n", index, str, nr1, nr2))

#define ElterOk(el)                                                                                                                     \
{                                                                                                                                       \
   if (!el)                                                                                                                             \
      HeapFehler("Elternteil nicht besetzt",0L,0L);                                                                                     \
                                                                                                                                        \
   if ((el->LebtNoch) && (el->ElternNummer != 0) && !(el->re))                                                                          \
      HeapFehler("lebendem Elternteil anhaengende re nicht besetzt",0L,0L);                                                             \
                                                                                                                                        \
   if ((el->LebtNoch) && (el->re)){                                                                                                     \
      if (el->re->Nummer != el->ElternNummer)                                                                                           \
         HeapFehler("Elternteil mit falscher Nummer",el->re->Nummer , el->ElternNummer);                                                \
                                                                                                                                        \
      if (((InfoT *) TP_SUEInfo(el->re)) != el)                                                                                         \
         HeapFehler("re zeigt auf anderen Info-Block",0L,0L);                                                                           \
                                                                                                                                        \
      if (RE_IstRegel(el->re)){                                                                                                         \
         if (el->ElternNummer > RE_AnzRegelnGesamt())                                                                                   \
            HeapFehler("Regel mit falscher Nummer", el->ElternNummer, RE_AnzRegelnGesamt());                                            \
                                                                                                                                        \
         if (el->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))                                                       \
            HeapFehler("Regel mit zuvielen noch unbeh. KPs",el->AnzNochNichtBehandelterKPs,  2*(AnzKPEingefuegt-AnzKPGeloescht));       \
                                                                                                                                        \
      }                                                                                                                                 \
      else{                                                                                                                             \
         if (el->ElternNummer < -RE_AnzGleichungenGesamt())                                                                             \
            HeapFehler("Gleichung mit falscher Nummer",el->ElternNummer,  -RE_AnzGleichungenGesamt());                                  \
                                                                                                                                        \
         if (el->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))                                                       \
            HeapFehler("Gleichung mit zuvielen noch unbeh. KPs",el->AnzNochNichtBehandelterKPs, 2*(AnzKPEingefuegt-AnzKPGeloescht));    \
                                                                                                                                        \
      }                                                                                                                                 \
   }                                                                                                                                    \
}

#define ElementKleiner(elt1, elt2) ((elt1.w1<elt2.w1) || ((elt1.w1==elt2.w1) && (elt1.w2<elt2.w2)))

static void HeapTest(char *str)
{
  unsigned long int index = IstHeap?0:1;
  printf("%s HEAP1-TEST %ld %s        \r", ok?"++":"--", ++testNr, str);

  while (index <= LetztesBesetztes){
      ElterOk(Elemente [index].Vater);
      ElterOk(Elemente [index].Mutter);
      
      if ((2*index+1<= LetztesBesetztes) &&
           ElementKleiner (Elemente [2*index+1], Elemente [index]))
         HeapFehler("Sortierung falsch-1",Elemente [index].w1, Elemente [2*index+1].w1);
      if ((2*index+2<= LetztesBesetztes) &&
           ElementKleiner (Elemente [2*index+2], Elemente [index]))
         HeapFehler("Sortierung falsch-2", Elemente [index].w1, Elemente [2*index+2].w1);
      if (!Elemente [index].tp)
         HeapFehler("Termpaar fehlt!",0L,0L);
      index++;
   }
}

#else
#define HeapTest(str)
#endif /* KPH_TEST */

/*-----------------------------------------------------------------------------*/
/*------------------------------ Einsinken Lassen -----------------------------*/
/*-----------------------------------------------------------------------------*/
/* Diese Funktion laesst das Element an der Stelle 'position' in den Heap      */
/* einsinken. Dabei wird das Gewicht als Vergleichs-Mittel herangezogen.       */
/* Der Rueckgabewert zeigt an, ob noch Veraenderungen im Heap noetig waren.    */
/* Die folgenden Funktionen unterscheiden sich in der Art und Weise, wie KPs   */
/* mit gleichem Gewicht behandelt werden. Eingestellt wird die entsprechende   */
/* Funktion in InitAposteriori.                                                */
/*-----------------------------------------------------------------------------*/



/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Element-Vergleiche ----------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define C_Groesser(e1,e2) ((e1.w1 > e2.w1) || ((e1.w1 == e2.w1) && (e1.w2 > e2.w2)))

/* ----------------------------------------------------------------------------*/
/* ------- Lasse Einsinken ----------------------------------------------------*/
/* ----------------------------------------------------------------------------*/
/* Diese Funktion laesst das Element an der Stelle 'position' in den Heap      */
/* einsinken. Dabei wird das Gewicht als Vergleichs-Mittel herangezogen.       */
/* Der Rueckgabewert zeigt an, ob noch Veraenderungen im Heap noetig waren.    */
/*-----------------------------------------------------------------------------*/


static BOOLEAN LasseEinsinken (unsigned long position) 
{
   BOOLEAN        changed = FALSE;
   unsigned long  j = position*2+1;
   HeapZellenT    merker = Elemente [position];
   
   while (j<=LetztesBesetztes) {
     if ((j<LetztesBesetztes) &&                                      /* Position j+1 noch gueltig? */
	 (C_Groesser (Elemente [j], Elemente [j+1])))
       j++;                                                           /* => j hat jetzt den Wert des leichteren Nachfolgers */
     if (C_Groesser (merker, Elemente [j])) {
       Elemente [position] = Elemente [j];
       changed = TRUE;
       position = j;
       j *=2;
       j++;
     }
     else
       /* Fertig - und raus */
       break;
   }
   if (changed)
      Elemente [position] = merker;
   return changed;
}

/* Heap neu aufbauen, nachdem ein Element aus der Mitte geloescht wurde */
static void ReorganiseHole (unsigned long hole) 
{
  unsigned long pos = hole;

  /* Loch an Topposition schieben (also alles auf dem Pfad nach unten) */
  while (pos > 0) {
    Elemente [pos] = Elemente [(pos-1)/2];
    pos = (pos-1) / 2;
  }
  if (LetztesBesetztes > 0) {
    Elemente [0] = Elemente [LetztesBesetztes];
    --LetztesBesetztes;
    LasseEinsinken (0);
  } else {
    IstHeap = FALSE;
  }
}


/*-----------------------------------------------------------------------------*/
/*------------------------------ Zugriffe auf den Heap ------------------------*/
/*-----------------------------------------------------------------------------*/

#define /*unsigned long*/ HP_AnzahlEintraege(/*void*/) ( (IstHeap)? LetztesBesetztes:(LetztesBesetztes-1))
/* Gibt die Anzahl der momentan im Heap befindlichen Eintraege zurueck.*/

#define /*BOOLEAN*/ HP_HeapIstLeer(/*void*/) (!(IstHeap) && (LetztesBesetztes==0))
/* Gibt zurueck, ob der Heap leer ist oder nicht. */

#define /*BOOLEAN*/ HP_HeapIstNichtLeer(/*void*/) ((IstHeap) || (LetztesBesetztes>0))
/* Gibt zurueck, ob der Heap Elemente enthaelt oder nicht. */

/* ---------------------------------------------------------------*/
/* --------------- Typen & Prozeduren fuer normalen Heap ---------*/
/* ---------------------------------------------------------------*/

static SV_SpeicherManagerT InfoManager;

static void KPVonElternGeloescht(InfoT *Vater, InfoT *Mutter)
/* Dekrementiert den Zaehler der noch nicht behandelten KPs in Vater und Mutter und loescht
   die Info-Eintraege, falls die zug. Regel o. Gleichung als tot markiert ist und es keine
   unbehandelten KPs zu dieser Regel oder Gleichung im Heap mehr gibt.*/
{
   (Vater)->AnzNochNichtBehandelterKPs--;
   (Mutter)->AnzNochNichtBehandelterKPs--; 
   if (Vater!=Mutter){
      if (((Vater)->AnzNochNichtBehandelterKPs == 0)
          && !((Vater)->LebtNoch)){
         SV_Dealloc(InfoManager, Vater);
      }
      
      if (((Mutter)->AnzNochNichtBehandelterKPs == 0)
          && !((Mutter)->LebtNoch)){
         SV_Dealloc(InfoManager, Mutter);
      }
   }
   else if (((Vater)->AnzNochNichtBehandelterKPs == 0)
            && !((Vater)->LebtNoch)){
      SV_Dealloc(InfoManager, Vater);
   }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

                                                                                                                           
void KPH_INSERT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
{
   HeapTest("Anfang INSERT");
   IncAnzKPEingefuegt ();
   IncAnzInitialerKPs (!IsDemigod(otherParent), (((InfoT *) TP_SUEInfo(actualParent))->AnzInitialerKPs));
   if (IstHeap) {
      unsigned long position;
      
      LetztesBesetztes++;
#if !KPH_USE_MMAP

      if (LetztesBesetztes >= AllocElemente) {
	AllocElemente += DomainGran;
	Elemente = our_realloc (Elemente, sizeof (HeapZellenT) * AllocElemente, "KPHeap Elements");
      }
#endif
      Elemente [LetztesBesetztes].Vater = (InfoT *) TP_SUEInfo(actualParent);
      Elemente [LetztesBesetztes].Mutter = (InfoT *) TP_SUEInfo(otherParent);
      Elemente [LetztesBesetztes].tp = tp;
      Elemente [LetztesBesetztes].w1 = wt.w1;
      Elemente [LetztesBesetztes].w2 = wt.w2;
      
      position = LetztesBesetztes;                                    /* Da der Heap nicht leer war, sind jetzt min. 2 Elemente drin. */
      do{
         position = (position-1)/2;
      } while ((LasseEinsinken(position)) && (position > 0) ); /* Reihenfolge wichtig: Einsinkenlassen muss durchgefuehrt werden,*/
                                                                    /* auch wenn position schon null ist. */
   }
   else {                                                            /* Heap hat leeres Top-Element */
      Elemente [0].Vater = (InfoT *) TP_SUEInfo(actualParent);
      Elemente [0].Mutter = (InfoT *) TP_SUEInfo(otherParent);
      Elemente [0].tp = tp;
      Elemente [0].w1 = wt.w1;
      Elemente [0].w2 = wt.w2;

      LasseEinsinken(0);
      IstHeap = TRUE;
   }
   ((InfoT *) TP_SUEInfo(actualParent))->AnzNochNichtBehandelterKPs++;
   ((InfoT *) TP_SUEInfo(otherParent))->AnzNochNichtBehandelterKPs++;

   HeapTest("Ende INSERT");
}


BOOLEAN KPH_POSTPONE (TermpairT tp MAYBE_UNUSEDPARAM, WeightEntryT wt MAYBE_UNUSEDPARAM,
                      RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
                      RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
{
  return FALSE;
}
/* evtl. auch ohne Parameter, denn es wird ja einfach nur das zuletzt selected'e KP postponed */


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Select ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#if PUNKT_IRP_HACK
extern WeightEntryT AltesGewicht; /* Punkt-IRP-Hack */
#endif

static BOOLEAN KPH_Select_Fifo (TermpairT *tp,
                                RegelOderGleichungsT *actualParent,
                                RegelOderGleichungsT *otherParent)
{
  WeightT min = LONG_MAX;
  unsigned int PosMin = 0;
  unsigned int Pos = 0;

  if (!IstHeap && (LetztesBesetztes > 0)) {
    Elemente [0]= Elemente [LetztesBesetztes];
    LetztesBesetztes--;
    LasseEinsinken (0);
    IstHeap  = TRUE;
  }
  if (HP_HeapIstLeer())
    return FALSE;

  for (Pos = 0; Pos <= LetztesBesetztes; Pos++) {
    if (Elemente [Pos].Vater->LebtNoch && Elemente [Pos].Mutter->LebtNoch) {
      if (Elemente[Pos].w2 < min) {
        min = Elemente[Pos].w2;
        PosMin = Pos;
      }
    }
  }
  
  if (min == LONG_MAX)
    return FALSE;
  
  *tp = Elemente [PosMin].tp;
  *actualParent = Elemente [PosMin].Vater->re;
  *otherParent = Elemente [PosMin].Mutter->re;
#if PUNKT_IRP_HACK
  AltesGewicht.w1 = Elemente [PosMin].w1; /* Punkt-IRP-Hack */
  AltesGewicht.w2 = Elemente [PosMin].w2; /* Punkt-IRP-Hack */
#endif
  KPVonElternGeloescht(Elemente [PosMin].Vater, Elemente [PosMin].Mutter);

  ReorganiseHole (PosMin);

  SUE_LastSelectedWasFifo = TRUE;
  return TRUE;
}

BOOLEAN KPH_SELECT (TermpairT *tp, RegelOderGleichungsT *actualParent, RegelOderGleichungsT *otherParent)
{
  if (SUE_AnzAktivierterRE % modulo < threshold) {
    return KPH_Select_Fifo (tp, actualParent, otherParent);
  }
  HeapTest("Anfang SELECT");
  while (HP_HeapIstNichtLeer()){
      IncAnzKPEntfernt();
      if (!IstHeap){
         Elemente [0] = Elemente [LetztesBesetztes];
         LetztesBesetztes--;
         LasseEinsinken(0);
      }
      else
         IstHeap = FALSE;
      
      /* Jetzt ist Elemente [0] gueltig und der Heap restauriert. */
      /* Nach der Entnahme eines Elementes ist IstHeap auf jeden Fall FALSE! */
      
      if (Elemente [0].Vater->LebtNoch && Elemente [0].Mutter->LebtNoch){
         *tp = Elemente [0].tp;
         *actualParent = Elemente [0].Vater->re;
         *otherParent = Elemente [0].Mutter->re;
#if PUNKT_IRP_HACK
	 AltesGewicht.w1 = Elemente [0].w1; /* Punkt-IRP-Hack */
	 AltesGewicht.w2 = Elemente [0].w2; /* Punkt-IRP-Hack */
#endif
         KPVonElternGeloescht(Elemente [0].Vater, Elemente [0].Mutter);
         HeapTest("Ende SELECT" );

         SUE_LastSelectedWasFifo = FALSE;
         return TRUE;
      }
      else{
         /* Meldung, Zaehler */
         IncAnzKPSpaeterWaisenmord();
         TPR_DESTROYTERMPAIR(Elemente [0].tp, Elemente [0].Vater->re, Elemente [0].Mutter->re);
         KPVonElternGeloescht(Elemente [0].Vater, Elemente [0].Mutter);
      }
  }
  
  return FALSE;
}

      
      

BOOLEAN KPH_SELECTINITIAL (TermpairT *tp, RegelOderGleichungsT *actualParent,
                           RegelOderGleichungsT *otherParent)
{
  unsigned int Index = 0;

  HeapTest("Anfang SELECT-INITIAL");
  
  if (LetztesBesetztes == 0) /* Heap leer -> alle initialen weg */
    return FALSE;
  
  if (!IstHeap){
    Elemente [0] = Elemente [LetztesBesetztes];
    LetztesBesetztes--;
    LasseEinsinken(0);
    IstHeap = TRUE;
  }
  
  while (!IsDemigod(Elemente [Index].Vater->re) && (Elemente [Index].w1!=LONG_MIN)
         && (Index<LetztesBesetztes))
    Index++;
  
  if (Index==LetztesBesetztes) return FALSE;
  else{
    IncAnzKPEntfernt();
    *tp = Elemente [Index].tp;
    *actualParent = Elemente [Index].Vater->re;
    *otherParent = Elemente [Index].Mutter->re;
    
    KPVonElternGeloescht(Elemente [Index].Vater, Elemente [Index].Mutter);
    
    if (Index<LetztesBesetztes) Elemente [Index]=Elemente [LetztesBesetztes];
    
    if (LetztesBesetztes>0){
      int run=Index;
      LetztesBesetztes--;
      while (run>0){
        LasseEinsinken(run);
        run = (run-1)/2;
      }
      LasseEinsinken(0);
    }
    
    HeapTest("Ende SELECT-INITIAL" );
    
    return TRUE;
  }
}


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Parents ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void KPH_NEWPARENT(RegelOderGleichungsT re)
{
   InfoT *infoHP;

   SV_Alloc(infoHP,InfoManager,InfoT *);
   infoHP -> LebtNoch = TRUE;
   infoHP -> ElternNummer = TP_ElternNr(re);
   infoHP -> AnzNochNichtBehandelterKPs = 0;
   infoHP -> re = re;

#if PQ_STATISTIK   
   infoHP -> AnzInitialerKPs = 0;
#endif
   TP_SUEInfo(re) = (void *) infoHP;

   if (!TP_TermpaarIstRegel(re)){ /* dh. Gleichung */
      SV_Alloc(infoHP,InfoManager,InfoT *);
      infoHP -> LebtNoch = TRUE;
      infoHP -> ElternNummer = TP_ElternNr(re);
      infoHP -> AnzNochNichtBehandelterKPs = 0;
      infoHP -> re = TP_Gegenrichtung(re);
#if PQ_STATISTIK   
      infoHP -> AnzInitialerKPs = 0;
#endif
      TP_SUEInfo(TP_Gegenrichtung(re)) = (void *) infoHP;
   }
   HeapTest("NEWPARENT");
}


void KPH_KILLPARENT(RegelOderGleichungsT re)
{
   InfoT *info = (InfoT *)TP_SUEInfo(re);

   info-> LebtNoch = FALSE;
   if (info->AnzNochNichtBehandelterKPs == 0){
      SV_Dealloc(InfoManager, info);
   }
   else
      info ->re = NULL;
      
   TP_SUEInfo(re) = (void *) NULL;

   if (!TP_TermpaarIstRegel(re)){ /* dh. Gleichung */
      info = (InfoT *)TP_SUEInfo(TP_Gegenrichtung(re));
      info-> LebtNoch = FALSE;
      if ( info->AnzNochNichtBehandelterKPs == 0){
         SV_Dealloc(InfoManager, info);
      }
      else
         info -> re =NULL;
      TP_SUEInfo(TP_Gegenrichtung(re)) = (void *) NULL;
   }
   HeapTest("KILLPARENT");
}

void KPH_REDIRECTPARENT(void)
{
   ;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int KPH_AnzKPsToGo(void)
{
   return IstHeap?LetztesBesetztes:(LetztesBesetztes+1);
}


static INLINE void WALKTHROUGH_BODY (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
{
  if (!IstHeap && (LetztesBesetztes>0)){
    Elemente [0]= Elemente [LetztesBesetztes--];
    LasseEinsinken(0); /* Heap reorganisieren */
    IstHeap = TRUE;
  }
   
  HeapTest("Anfang WALKTHROUGH");
  if (HP_HeapIstNichtLeer()){
    unsigned long int index = 0;
    BOOLEAN resort = FALSE;
    
    while((index <= LetztesBesetztes) && IstHeap){
      IncAnzKPIRBetrachtet();

      if (!(Elemente [index].Vater->LebtNoch && Elemente [index].Mutter->LebtNoch)){
        AZ_DruckeMeldungItem();
        IncAnzKPIRWaisenmord();
        IncAnzKPEntfernt();
        TPR_DESTROYTERMPAIR(Elemente [index].tp, Elemente [index].Vater->re,
                            Elemente [index].Mutter->re);
        KPVonElternGeloescht(Elemente [index].Vater, Elemente [index].Mutter);
        if (index < LetztesBesetztes)
          Elemente [index] = Elemente [LetztesBesetztes];
        if (LetztesBesetztes) {
          LetztesBesetztes--;
        } else {
          IstHeap = FALSE;
        }
        
        resort = TRUE;
      }
      else {
        int APresult;
        PCL_ObjektNummerT pclID;
        TermT lhs = NULL, rhs = NULL;

        if (NeedsTerms) {
          TPR_TERMPAIRtoPARALLEL_INTERMED(Elemente [index].tp, &lhs, &rhs,
                                          Elemente [index].Vater->re,
                                          Elemente [index].Mutter->re, &pclID);
        }
        if ((APresult = (* AP) (lhs, rhs, Elemente [index].Vater->re,
                                Elemente [index].Mutter->re, &pclID)) == 1) {
          IncAnzKPEntfernt();
          IncAnzKPIRGeloescht();
          if (NeedsTerms) {
            TPR_DESTROYPARALLEL (Elemente [index].tp, lhs, rhs,
                                 Elemente [index].Vater->re,
                                 Elemente [index].Mutter->re);
          }
          
          KPVonElternGeloescht(Elemente [index].Vater,
                               Elemente [index].Mutter);
          if (index<LetztesBesetztes) {
            Elemente [index] = Elemente [LetztesBesetztes];
          }
          if (LetztesBesetztes) {
            LetztesBesetztes--;
          }
          else {
            IstHeap = FALSE;
          }
          resort = TRUE;
        }
        else if (APresult == 2) {
          WeightEntryT wt;
               
          IncAnzKPIRVeraendert();

          wt.w1 = Elemente [index].w1;
          wt.w2 = Elemente [index].w2;
          C_ReClassify(&wt, lhs, rhs, Elemente [index].Vater->re,
                       Elemente [index].Mutter->re);
          if (NeedsTerms) {
            Elemente [index].tp = TPR_PARALLELtoTERMPAIR (Elemente [index].tp, lhs, rhs,
                                                            Elemente [index].Vater->re,
                                                            Elemente [index].Mutter->re, TRUE, pclID);
          }
          Elemente [index].w1 = wt.w1;
          Elemente [index].w2 = wt.w2;
          resort = TRUE;
          index++;
        } else {               /* noop */
          if (NeedsTerms) {
            Elemente [index].tp = TPR_PARALLELtoTERMPAIR (Elemente [index].tp, lhs, rhs,
                                                            Elemente [index].Vater->re,
                                                            Elemente [index].Mutter->re, TRUE, pclID);
          }
          index++;
        }
      }
    }
    
    if ((LetztesBesetztes>0) && resort) {
      index = (LetztesBesetztes-1) / 2;
      while (index>0) {
        index--;
        LasseEinsinken(index);
      }
    }
  }
  HeapTest("Ende WALKTHROUGH");
}


void KPH_WALKTHROUGH(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
{
  /* der gcc sollte folgenden Code inlinen und optimieren */
  if (NeedsTerms)
    WALKTHROUGH_BODY(AP, TRUE);
  else
    WALKTHROUGH_BODY(AP, FALSE);
}


void KPH_WATCHINGWALKTHROUGH(PQ_WatchingWalkthroughProcedureT AP)
{
   unsigned long int index;

   if (!IstHeap && (LetztesBesetztes==0)) return; /* Heap ist leer */
   index = IstHeap?0:1;
   while(index <= LetztesBesetztes){
     
      if ((Elemente [index].Vater->LebtNoch && Elemente [index].Mutter->LebtNoch)) {
        WeightEntryT wt;

        wt.w1 = Elemente [index].w1;
        wt.w2 = Elemente [index].w2;
        
        (* AP)(wt, Elemente [index].tp, Elemente [index].Vater->re, Elemente [index].Mutter->re);
      }
      index++;
   }
}

void KPH_TermVerfolgung (void (* OperationAufTermen) (TermT, TermT))
{
   unsigned long int index;

   if (!IstHeap && (LetztesBesetztes==0))
     return; /* Heap ist leer */

   index = IstHeap?0:1;
 
   while (index <= LetztesBesetztes){
     TPR_OperationWaehrendTermverfolgung(OperationAufTermen, Elemente [index].tp,
                                         Elemente [index].Vater->re,
                                         Elemente [index].Mutter->re);
     index++;
   }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static unsigned long int Lebende;

static void LebendsKPsZaehlen(WeightEntryT wt MAYBE_UNUSEDPARAM, TermpairT tp MAYBE_UNUSEDPARAM, RegelOderGleichungsT actualP MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherP MAYBE_UNUSEDPARAM)
{
   Lebende++;
}


void KPH_STATISTICS(InfoStringT is)
{
   if (!(IS_IsIn("all", is) || (IS_IsIn("PQ", is))))
      return;                   /* Modul nicht angesprochen */
   if (IS_IsIn("standard", is)){
      IO_DruckeKurzUeberschrift("Priority-Queue: Regular heap", '+');
      IO_DruckeStatistikZeile("actual number of elements in heap", HP_HeapIstLeer()?0:(LetztesBesetztes - IstHeap?0:1));
#if PQ_STATISTIK
      Lebende = 0;
      KPH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
      PQStatistik(is, Lebende);
#endif
   }
   else if(IS_IsIn("latex-head", is)){
      fprintf(IO_Strom, "PQ: total &");
#if PQ_STATISTIK
      PQStatistik(is, 0);
#endif
   }
   else if(IS_IsIn("intern", is))
      IO_DruckeStatistikZeile("KPH: domain granularity", DomainGran);
   else if(IS_IsIn("config", is)){
      fprintf(IO_Strom, "KPH: Statistik %s, ", PQ_STATISTIK ?" On":"Off");
#if KPH_TEST
      fprintf(IO_Strom, "Heap-Test-Modus  On\n");
#else
      fprintf(IO_Strom, "Heap-Test-Modus Off\n");
#endif
   }
}


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Clean up --------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void KPH_SoftAbraeumen(void)
/* Nur Nicht-SV-Manager-abhaengiges leeren, aber Wiederaufsatz-kompatibel */
{
#if KPH_USE_MMAP
  SM_StrahldomaeneLoeschen(Domaene);
#else
  our_dealloc (Elemente);
  Elemente = NULL;
#endif
}

void KPH_Abraeumen(void)
/* Alles loeschen*/
{
  unsigned long int i;

  /* Folgendes wird benoetigt, damit Speicherstatistik stimmt */
  for (i=IstHeap?0:1; i<=LetztesBesetztes; i++) {
    TPR_DESTROYTERMPAIR(Elemente[i].tp, Elemente[i].Vater->re, Elemente[i].Mutter->re);
    KPVonElternGeloescht(Elemente [i].Vater, Elemente [i].Mutter);
  }

#if KPH_USE_MMAP
  SM_NichtGefrorenenStrahlLoeschen(Elemente);
  SM_StrahldomaeneLoeschen(Domaene);
#else
  our_dealloc (Elemente);
  Elemente = NULL;
#endif
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN KPH_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  int fifo, heuristic;

  if (IS_ValueOfOption(DEF, USR, "domgran")) {
      DomainGran = atoi(IS_ValueOfOption(DEF, USR, "domgran"));
   } else {
#if KPH_USE_MMAP   
     DomainGran = 10; /* 10 Seiten */
#else
     DomainGran = 10000; /* 10000 KPs */
#endif
   }
#if KPH_USE_MMAP   
   Domaene = SM_NeueStrahldomaene(DM_KPH, DomainGran, "Domain for regular heap");
   Elemente = SM_NeuerStrahl(Domaene);
#else
   AllocElemente = DomainGran;
   Elemente = our_alloc (sizeof (HeapZellenT) * AllocElemente);
#endif
   
#if !COMP_POWER
   fprintf(IO_Strom, "Regular Heap. statistics = %s. "
           "domgran = %ld. ", 
           PQ_STATISTIK ? "on":"off", DomainGran);
#endif
   if (PI_ParseInterleave (USR, &fifo, &heuristic)
       || PI_ParseInterleave (DEF, &fifo, &heuristic)) {
     modulo = fifo + heuristic;
     threshold = fifo;
#if !COMP_POWER
     fprintf (IO_Strom, "  Interleaving of CP Queues\n");
     fprintf (IO_Strom, "    Fifo: %d, Heuristic: %d\n", fifo, heuristic);
#endif
   } else {
     modulo = 1;
     threshold = 0;
   }
   
   IstHeap = FALSE;        /* Dh. die Wurzel ist unbesetzt ! */

   LetztesBesetztes = 0;

   SV_ManagerInitialisieren(&InfoManager, InfoT, PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (Heap)2");

   KPH_INSERTImmortal = KPH_INSERT;
   return TRUE;
}

void KPH_ActSize(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  Lebende = 0;
  KPH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
  
  *AktGueltigeKPZahl = Lebende;
#if PQ_STATISTIK
  *AktKPZahl = AnzKPEingefuegt - AnzKPGeloescht;
#else
  *AktKPZahl = Lebende;
#endif
}
