/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Packen.c                                                                                                   */
/* - ********                                                                                                   */
/* -                                                                                                            */
/* - Dieses Modul stellt die Funktionalitaet zum Ein- und Auspacken von Termen in Byte-Ketten zur Verfuegung.   */
/* - Diese Byteketten koennen dann entweder zur platzsparenden Ablage der KP-Menge verwendet werden oder        */
/* - aber zum Versenden von Termen im Rahmen von DISCOUNT,                                                      */
/* -                                                                                                            */
/* - 27.03.95 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

/* Dokumentation:

Funktionalitaet etc.:

1. Terme sollen in Byte-Ketten abgelegt bzw. aus solchen restauriert werden koennen.

2. Terme als Byteketten sollen als Grundlage der KP-Menge herhalten koennen. Dazu ist eine
   entsprechende Speicherverwaltung zu etablieren. Diese soll:
   - fuer alle Byteketten-Laengen in Vierer-Schritten Freispeicherlisten fuehren,
     (-> Array von Speichermanagern)
   - fuer Byteketten oberhalb einer bestimmten Laenge einzeln den Platz allokieren

3. Da Terme in der KP-Menge nur paarweise auftreten, werden Termpaare hier gemeinsam eingepackt.
   Dadurch entsteht nur ein- statt zweimal Platzverschwendung durch Alignment.

4. Jedes gepackte Termpaar steht in einem Byte-String im Speicher. Dieser String enthaelt als erstes Byte
   die Nummer des entsprechenden Speichermanagers. Ist der Platz einzeln allokiert worden, so findet sich dort
   die Zahl 0xff == 255. Dadurch ergibt sich folgende Restriktion:
   es duerfen max. 254 Speichermanager angelegt werden!

*/
#include "Ausgaben.h"
#include "compiler-extra.h"
#include "general.h"
#include "Packen.h"
#include "pcl.h"
#include "SUEBasis.h"
#include "SpeicherVerwaltung.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"

#include "Stringterms.h"

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------- Typdefinitionen fuer eingepackte Terme --------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#if COMP_PCL

typedef struct PCLTS{
   PCL_ObjektNummerT pclID;
   byte *tp;
} *PCLTermpaarT;

#endif

typedef TermzellenZeigerT TermpaarT;                     /* nicht gepacktes, aber aneinandergehaengtes Termpaar */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------- Speichermanagement fuer eingepackte Terme/ Globale Variablen ----------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

const int FixAlignment  = sizeof(struct{char c; int i;}) - sizeof(int);
static int Alignment;
static unsigned int AnzInitManager;

#define PseudoManagerNummer 255 /* Manager-Nummer, die bei einzeln allokierten PTermpaaren eingetragen wird. */

#define PK_AnzahlInitialerManager (AnzInitManager)
/* Anzahl der Manager, die zur Verwaltung der gepackten Termpaare verwendet werden sollen. Dabei wird der verwaltete
   Speicherplatz in Alignment-Byte-Schritten erhoeht. Die Manager verwalten also Speicherpakete der Groessen
   4, 8, 16, ... 100 Byte. Darueberhinausgehende Anforderungen werden solo, d.h. ueber alloc und dealloc behandelt.
*/

static unsigned int EinzelnAnlegenSchwelle; /* Als Variable, damit der Wert nicht jedesmal neu berechnet werden muss. */

static SV_SpeicherManagerT *SpeicherManagerArray; /* direkt in maximaler Groesse anlegen */

static unsigned long int AnzahlGepackterTermpaareOhneManager;

#if COMP_PCL
static SV_SpeicherManagerT PCLManager;
#endif
/* -------------------------------------------------------------------------------------------------------------*/
/* ------------------- Anlegen und Freigeben von gepackten Termpaaren ------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static PTermpaarT HoleTermpaar(unsigned long int groesse)
/* Liefert den Speicherbereich fuer ein Termpaar der Laenge groesse. */
{
   PTermpaarT pp;


   if (groesse  >= EinzelnAnlegenSchwelle){
      AnzahlGepackterTermpaareOhneManager++;
      pp = (PTermpaarT) our_alloc((groesse+1)*sizeof(byte)); /* +1 wegen Speichermanager-Index */
      *pp = (unsigned char) PseudoManagerNummer;
   }
   else{
      groesse /= Alignment; /* (*) */
      SV_Alloc(pp,SpeicherManagerArray[groesse],PTermpaarT);
      *pp = (unsigned char) groesse;
   }
   return pp;
}

#define TermpaarFreigeben(/*PTermpaarT*/ pp)                    \
/* Termpaar der angegebnen Groesse freigeben.*/                 \
{                                                               \
   if ((unsigned char) *pp == PseudoManagerNummer)              \
      our_dealloc(pp);                                          \
   else{                                                        \
     unsigned char ManagerNummer = (unsigned char) *pp;         \
     SV_Dealloc(SpeicherManagerArray[ManagerNummer], pp);       \
  }                                                             \
}

/* 

(*) Berechnung des zu einer Groesse gehoerigen Speichermanagers.:

Gesucht wird zu groesse das naechsthoehere Vielfache von Alignment, dh.:

    das kleinste x, so dass x*Alignment >= groesse.
<=> das kleinste x, so dass x*Alignment > groesse -1.
<=> das kleinste x, so dass x > (groesse-1)/Alignment.

Es gilt: (groesse-1)/Alignment >= (groesse-1) div Alignment
=> das gesuchte x ist > (groesse-1) div Alignment,
und da das kleinste solche (ganzzahlige) x gesucht wird, ist die gesuchte Zahl

   x = (groesse-1) div Alignment + 1.

Da aber immer ein Byte mehr fuer die Managernummer benoetigt wird:

   x = (groesse-1+1) div Alignment + 1
     = groesse div Alignment + 1.

Da ferner die Manager nicht mit der Nummer 1 sondern der Nummer 0 beginnen, muss noch eins abgezogen werden:

   x = groesse div Alignment + 1 - 1
     = groesse div Alignment
     = groesse / Alignment.
*/




static PTermpaarT TermpaarEinpacken(TermT lhs, TermT rhs)
/* Termpaar lhs=rhs einpacken und als gepacktes Termpaar zurueckgeben.
   Dieses steht dann bereits als eigenstaendiges Objekt zur Verfuegung. */
{
   PTermpaarT Termpaar;
   
   Termpaar = HoleTermpaar(STT_TermpaarPlatzbedarf(lhs,rhs));
   /* hier wird in das erste Byte des gepackte Termpaares die Nummer des */
   /* verwendeten Managers eingetragen. */

   STT_TermpaarEinpackenFragil(lhs, rhs, Termpaar+1);


   return (TermpairT) Termpaar;
   
}

#define GepacktesTermpaarLoeschen(pp)           \
/* Termpaar ohne Auspacken wegwerfen.*/         \
{                                               \
   TermpaarFreigeben(pp);                       \
}

static void TermpaarAuspacken(PTermpaarT tp, TermT *lhsPtr, TermT *rhsPtr)
{
   STT_TermpaarAuspacken(&tp,lhsPtr,rhsPtr);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize-------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static BOOLEAN PK_SchonInitialisiert = FALSE;

BOOLEAN PK_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* Etablieren der Speichermanager und diverse Vorbereitungen. */
{
   unsigned int i;
   char *stepStr, *manStr;
   long int STEP;

   STT_InitAposteriori();

   if (!PK_SchonInitialisiert){
     if ((stepStr = IS_ValueOfOption(DEF, USR, "step"))){
       STEP = atoi(stepStr);
       if (STEP <= 0) STEP = 1;
     }
     else 
       STEP = 1;
     
     if ((manStr = IS_ValueOfOption(DEF, USR, "managers"))){
       signed int Konversion = atoi(manStr);
       AnzInitManager = max(Konversion,1);
     }
     else 
       AnzInitManager = 25;
     
     SpeicherManagerArray = array_alloc(AnzInitManager,SV_SpeicherManagerT);
     
     Alignment = FixAlignment * STEP;
#if !COMP_POWER
     fprintf(IO_Strom, "STEP = %ld bytes, ", STEP);
#endif
   }
#if !COMP_POWER
   fprintf(IO_Strom, "Alignment = %d bytes, ", Alignment);
   fprintf(IO_Strom, "AnzInitManager  = %d\n", AnzInitManager);
#endif

   EinzelnAnlegenSchwelle = Alignment * PK_AnzahlInitialerManager;

   if (!PK_SchonInitialisiert)
     for (i=0; i<PK_AnzahlInitialerManager; i++) {
       SpeicherManagerArray[i].block_size = 0;                                        /* Initialisierung erzwingen */
       SV_BeliebigenManagerInitialisieren(&(SpeicherManagerArray[i]), (i+1)*Alignment, 
                                          PK_BlocksizeInBytes/(Alignment*(i+1)), FALSE, "Gepackte Termpaare");
     }
   AnzahlGepackterTermpaareOhneManager=0;
#if COMP_PCL
   SV_ManagerInitialisieren(&PCLManager, struct PCLTS,2000, "PCL-Manager (gepackt)");
#endif
   PK_SchonInitialisiert = TRUE;
   return TRUE;
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- New Flats generated ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void PK_NewFlatsGenerated(void)
{
   PCL_StartChain();
}

#if COMP_PCL

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

TermpairT PK_Flat2TP (TermT lhs, TermT rhs, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
                      UTermT actualPosition MAYBE_UNUSEDPARAM, PCL_ObjektNummerT pclID)
{
   PCLTermpaarT tp;
   SV_Alloc(tp,PCLManager,PCLTermpaarT);
   tp->tp = TermpaarEinpacken(lhs, rhs);
   TO_TermeLoeschen(lhs, rhs);
   tp->pclID = pclID;
   PCL_SaveChain();
   return tp;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PK_TP2Par (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, 
                RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclIDPtr)
/* Terme auspacken, Termpaar behalten. */
{
   *pclIDPtr = ((PCLTermpaarT) tp)->pclID;
   TermpaarAuspacken(((PCLTermpaarT) tp)->tp+1, lhsPtr, rhsPtr);
   PCL_StartChain();
}



/* -------------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PK_Par2Flat (TermpairT tp, TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, 
                  RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
/* Destroys the termpair-representation irrecoverable. Gepacktes Termpaar freigeben. */
{
   GepacktesTermpaarLoeschen(((PCLTermpaarT) tp)->tp);
   SV_Dealloc(PCLManager, (PCLTermpaarT) tp);
   PCL_SaveChain();
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

TermpairT PK_Par2TP (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, 
                     RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM, BOOLEAN changed, PCL_ObjektNummerT pclID)
{
   if (changed) {
      GepacktesTermpaarLoeschen(((PCLTermpaarT) tp)->tp);
      ((PCLTermpaarT) tp)->tp = (TermpairT) TermpaarEinpacken(lhs, rhs);
      ((PCLTermpaarT) tp)->pclID = pclID;
   }
   TO_TermeLoeschen(lhs, rhs);
   PCL_SaveChain();
   return tp;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void PK_DestroyTP  (TermpairT tp, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
/* Termpaar zerstoeren */
{
   GepacktesTermpaarLoeschen(((PCLTermpaarT) tp)->tp);
   SV_Dealloc(PCLManager, (PCLTermpaarT) tp);
}

void  PK_DestroyPar (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
/* Termpaar und Flachterme zerstoeren */
{
   TO_TermeLoeschen(lhs, rhs);
   GepacktesTermpaarLoeschen(((PCLTermpaarT) tp)->tp);
   SV_Dealloc(PCLManager, (PCLTermpaarT) tp);
   PCL_DeleteChain();
}

/* ------------------------------------------------------------------------------------------------------*/
/* ------- Unambiguous Comparison -----------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------*/

BOOLEAN PK_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                  TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2)
/* Vom Entstehungszeitpunkt unabhaengiger Vergleich von Termpaaren.
   Hier: Eltern-Nummern, dann Symb-Lex.
*/
{
   if (TP_ElternNr(father1) > TP_ElternNr(father2)) return TRUE;
   else if (TP_ElternNr(father1) < TP_ElternNr(father2)) return FALSE;
   else                         /* father1 == father2 */
      if (TP_ElternNr(mother1) > TP_ElternNr(mother2)) return TRUE;
      else if (TP_ElternNr(mother1) < TP_ElternNr(mother2)) return FALSE;
      else{                     /* beide Elternpaare identisch */
         return STT_TPTestSymbLex(((PCLTermpaarT) tp1)->tp+1, ((PCLTermpaarT) tp2)->tp+1);
      }
}   

#else
/* ROUTINEN OHNE PCL */
  
/* -------------------------------------------------------------------------------------------------------------*/
/* ------- FLATTERMS to TERMPAIR -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

TermpairT PK_Flat2TP (TermT lhs, TermT rhs, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
                      UTermT actualPosition MAYBE_UNUSEDPARAM, PCL_ObjektNummerT pclID MAYBE_UNUSEDPARAM)
{
   PTermpaarT tp = TermpaarEinpacken(lhs, rhs);
   TO_TermeLoeschen(lhs, rhs);
   return tp;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- TERMPAIR to PARALLEL --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PK_TP2Par (TermpairT tp, TermT *lhsPtr, TermT *rhsPtr, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, 
                RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM, PCL_ObjektNummerT *pclIDPtr MAYBE_UNUSEDPARAM)
/* Terme auspacken, Termpaar behalten. */
{
   TermpaarAuspacken(((PTermpaarT) tp)+1, lhsPtr, rhsPtr);
}



/* -------------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to FLATTERMS -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PK_Par2Flat (TermpairT tp, TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, 
                  RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
/* Destroys the termpair-representation irrecoverable. Gepacktes Termpaar freigeben. */
{
   GepacktesTermpaarLoeschen((PTermpaarT) tp);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- PARALLEL to TERMPAIR --------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

TermpairT PK_Par2TP (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, 
                     RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM, BOOLEAN changed, PCL_ObjektNummerT pclID MAYBE_UNUSEDPARAM)
{
   if (changed) {
      GepacktesTermpaarLoeschen((PTermpaarT) tp);
      tp = (TermpairT) TermpaarEinpacken(lhs, rhs);
   }
   TO_TermeLoeschen(lhs, rhs);
   return tp;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- DESTROY ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void PK_DestroyTP  (TermpairT tp, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
/* Termpaar zerstoeren */
{
   GepacktesTermpaarLoeschen((PTermpaarT) tp);
}

void  PK_DestroyPar (TermpairT tp, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM)
/* Termpaar und Flachterme zerstoeren */
{
   TO_TermeLoeschen(lhs, rhs);
   GepacktesTermpaarLoeschen((PTermpaarT) tp);
}

BOOLEAN PK_UnambiguousGreaterThan(TermpairT tp1, RegelOderGleichungsT father1, RegelOderGleichungsT mother1,
                                  TermpairT tp2, RegelOderGleichungsT father2, RegelOderGleichungsT mother2)
/* Vom Entstehungszeitpunkt unabhaengiger Vergleich von Termpaaren.
   Hier: Eltern-Nummern, dann Symb-Lex.
*/
{
   if (TP_ElternNr(father1) > TP_ElternNr(father2)) return TRUE;
   else if (TP_ElternNr(father1) < TP_ElternNr(father2)) return FALSE;
   else                         /* father1 == father2 */
      if (TP_ElternNr(mother1) > TP_ElternNr(mother2)) return TRUE;
      else if (TP_ElternNr(mother1) < TP_ElternNr(mother2)) return FALSE;
      else{                     /* beide Elternpaare identisch */
         return STT_TPTestSymbLex((PTermpaarT) tp1, (PTermpaarT) tp2);
      }
}   

#endif


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PK_Statistics(InfoStringT IS)
{
   if (IS_IsIn("standard", IS)){
      IO_DruckeKurzUeberschrift("Termpair Representation: packed", '+');
      IO_DruckeStatistikZeile("Anzahl gepackter Termpaare, die ohne Manager angelegt wurden", AnzahlGepackterTermpaareOhneManager);
   }
   else if (IS_IsIn("latex-head", IS)){
      fprintf(IO_Strom, "Singly packed pairs & ");
   }
   
}


void PK_PrintTermpair (TermpairT tp, RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM, RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM, const char*trenner)
{
#if COMP_PCL
   STT_DruckePTermpaar((((PCLTermpaarT) tp)->tp)+1, trenner);
#else
   STT_DruckePTermpaar(((PTermpaarT) tp)+1, trenner);
#endif
}


/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PK_CleanUp(void)
{
   ;
}

