/*=================================================================================================================
  =                                                                                                               =
  = Subgleichungen.c                                                                                              =
  =                                                                                                               =
  = Realisiereung einer Menge von Subgleichungen                                                                  =
  =                                                                                                               =
  = 28.07.94 Angelegt. Arnim.                                                                                     =
  =                                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =================================================================================================================*/


/* -------------------------------------------------------------------------------------------------------------*/
/* - Anforderungen: + Sukzessives Einfuegen von Subgleichungen einer Gleichung                                  */
/* -                + Loeschen aller Subgleichungen zu einer Gleichung (wenn eine Gleichung geloescht wird)     */
/* -                + Alle Subgleichungen aller Gleichungen durchgehen und Subsumptionstest mit neuer Gleichung */
/* -                  durchfuehren.                                                                             */
/* -                                                                                                            */
/* - Da die dritte Anforderung wohl am haeufigsten durchgefuehrt weren wird, ist hier auf moeglichst schnelle   */
/* - Implementierung zu achten.                                                                                 */
/* -                                                                                                            */
/* - Moegliche Struktur: einfach verkettete Liste.                                                              */
/* - Begruendung: + Die dritte Anforderung ist sehr schnell ausfuehrbar (O(n) fuer alle Subgleichungen).        */
/* -              + Das Einfuegen von Subgleichungen ist in O(1) moeglich.                                      */
/* -              + Das Loeschen aller Subgleichungen zu einer Gleichung ist zwar nur in O(n) moeglich, dies    */
/* -                kann aber durch das Abspeichern einer Referenz bei jeder Gleichung auf O(1) gedrueckt wer-  */
/* -                den, wenn die Liste zweiseitig verkettet wird.                                              */
/* -                                                                                                            */
/* - Alternativ-Struktur: Subsumptions-Baum                                                                     */
/* - Dieser Baum sollte dann die Subgleichungen als aneinanderhaengende Terme enthalten. Der Subsumptionstest   */
/* - mit einer neuen Gleichung wuerde sich dann auf den Matchversuch in einen DS-Baum reduzieren. Ferner sollten*/
/* - dann auch alle Gleichungen aus der GM als zusammenhaengende Terme in einen eigenen Baum gehaengt werden, um*/
/* - auch den aktiven Subsumptionstest zu unterstuetzen (wird die neue Gleichung von der Gleichungsmenge subs.).*/
/* - Vorteile : + schnellerer Subsumptionstest beim aktiven Test                                                */
/* -            + evtl. schnellerer Subsumptionstest beim passiven Test                                         */
/* -                                                                                                            */
/* - Nachteile: + Loeschen und Einfuegen ist erheblich langsamer                                                */
/* -            + hoher Programmieraufwand                                                                      */
/* -                                                                                                            */
/* -                                                                                                            */
/* - Tatsaechlich gewaehlte Struktur/Vorgehensweise:                                                            */
/* -                                                                                                            */
/* - + aktive Subsumption: neue Gleichung l=r. Mit Gleichungsbaum Matches suchen und auf Blattebene in der/den  */
/* -   rechten Seiten fortsetzen. Dies geschieht fuer alle Subgleichungen von l=r, und zwar nur einmal je       */
/* -   Subgleichung, da die Gleichungen bereits in beiden Richtungen im E-Baum gespeichert sind.                */
/* - + passive Subsumption: die Gleichungen werden zunaechst auf Top-Ebene getestet, und zwar geschieht dies    */
/* -   wiederum mit Hilfe des E-Baums: verlaengertes Interreduktionsproblem: Fortsetzung auf Blattebene in der/ */
/* -   den rechten Seiten der betroffenen Gleichung(en). Anschliessend werden in einer gesonderten Struktur     */
/* -   alle echten Subgleichungen der Gleichungen aus GM ueberprueft. Hierfuer wurde folgende Struktur gewaehlt:*/
/* -   - eine doppelt verkettete Liste von Kopf-Zellen, die jeweils fuer solche Gleichungen eingerichtet werden,*/
/* -     die echte Subgleichungen enthalten. Die Kopf-Zellen enthalten dann einen Verweis auf die Gleichungen   */
/* -     selbst sowie einen Zeiger auf die erste Subgleichung.                                                  */
/* -   - die Gleichungen erhalten einen Verweis auf ihre Kopf-Zelle in der Subgleichungs-Liste, um beim         */
/* -     Loeschen der Gleichung die Subgleichungen mitloeschen zu koennen, ohne lange suchen zu muessen.        */
/* -   - die an den Kopf-Zellen angehaengten Subgleichungen sind als einseitig verkettete Liste realisiert,     */
/* -     da sie nur als ganzes geloescht werden muessen.                                                        */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* - Ausblick : Da beim passiven Subsumptionstest (wird eine alte Gleichung von einer neuen subs.) nur Matches  */
/* -            in einen Baum berechnet werden muessen, bietet sich hier vielleicht eine ganz andere Struktur   */
/* -            an, z.B. Path-Indexing o.ae.                                                                    */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "Subgleichungen.h"
#include "TermOperationen.h"
#include "Puffer.h"
#include "Ausgaben.h"
#include "SpeicherVerwaltung.h"
#include "RUndEVerwaltung.h"	/* nur fuer Druck-Funktion */

 /* ------------- Die Subgleichungsmenge (SGM) ------------------------------------------------------------------*/

SV_SpeicherManagerT KopfZellenManager, SubgleichungsManager;


#define SGM_AnzahlKopfZellen 30
#define SGM_AnzahlSubgleichungen 100

static SGM_ReferenzT ErsteKopfZelle;

/* ----------------- Initialisieren ------------------------*/

void SGM_InitAposteriori(void)
/* Nur aufrufen, wenn der externe Modus proof oder completion ist.*/
{
  SV_ManagerInitialisieren(&KopfZellenManager, struct SGM_ReferenzTS, SGM_AnzahlKopfZellen, "Kopfzellen Subgleichungsmenge");
  SV_ManagerInitialisieren(&SubgleichungsManager, SubgleichungsT, SGM_AnzahlSubgleichungen, "Subgleichungen");
  ErsteKopfZelle = NULL;
}

SGM_ReferenzT SGM_SubgleichungenZuObjektAufnehmen(SGM_BetroffenesObjektT objekt)
/* Fuegt die echten Subgleichungen (aus dem SGPuffer) zu gleichung in die SGM ein. Zurueckgegeben wird der Verweis auf die 
   Kopfzelle, falls echte Subgleichungen existierten, andernfalls NULL. Mit der zurueckgegebenen Referenz wird das Loeschen
   der Subgleichungen zu gleichung gestartet.*/
{
   SGM_ReferenzT neu;
   SubgleichungsT *sg, *vorg;

   if (PU_SGPufferNichtLeer()){
      SV_Alloc(neu,KopfZellenManager,SGM_ReferenzT );
      neu->naechste = ErsteKopfZelle;
      neu->vorige = NULL;
      neu->objekt = objekt;
      if (ErsteKopfZelle != NULL){
	 ErsteKopfZelle->vorige = neu;
      }
      ErsteKopfZelle = neu;
      SV_Alloc(neu->erste,SubgleichungsManager,SubgleichungsT *);
      vorg = neu->erste;
      PU_SGAuslesen(&(vorg->links), &(vorg->rechts));
      vorg->naechste = NULL;
      while(PU_SGPufferNichtLeer()){
	 SV_Alloc(sg,SubgleichungsManager,SubgleichungsT *);
	 sg -> naechste = NULL;
	 PU_SGAuslesen(&(sg->links), &(sg->rechts));
	 vorg -> naechste = sg;
	 vorg = sg;
      }
      return neu;
   }
   else{
      return NULL;		/* keine (echten) Subgleichungen im Puffer */
   }
}

static BOOLEAN nochSGs;
static SubgleichungsT *aktSG;
static SGM_ReferenzT aktKopfZelle;

void SGM_LoescheSubgleichungen(SGM_ReferenzT Referenz)
/* Loescht die zu einer Gleichung eingetragenen Subgleichungen. Diese Prozeddur muss immer dann aufgerufen werden,
   wenn eine Gleichung geloescht wird, und zwar unabhaengig davon, ob die Referenz==NULL ist oder nicht. */
{
   SubgleichungsT *temp, *naechste;

   if ((aktKopfZelle != NULL) && (Referenz == aktKopfZelle)) {               /* SGs an aktueller Durchlaufposition */
     if (aktKopfZelle->naechste != NULL){
       aktKopfZelle = aktKopfZelle->naechste;
       aktSG = aktKopfZelle->erste;
     }
     else
       nochSGs = FALSE;
   }
   if (Referenz !=NULL){
      if (ErsteKopfZelle == Referenz)
	 ErsteKopfZelle = Referenz->naechste;
      else			/* Referenz ist nicht die erste Kopfzelle in der Liste, aber ggf. die letzte */
	 Referenz->vorige->naechste = Referenz->naechste;
      if (Referenz->naechste != NULL)
	 Referenz->naechste->vorige = Referenz->vorige; /* ist NULL, wenn Referenz die erste KZ in der Liste ist */
      temp = Referenz->erste;
      do{			/* Die Subgleichungsliste zu der Gleichung enthaelt mindestens ein Element. */
	 naechste = temp->naechste;
	 SV_Dealloc(SubgleichungsManager, temp);
	 temp = naechste;
      }while(temp != NULL);
      SV_Dealloc(KopfZellenManager, Referenz);
   }
				/* kein else, da fuer Referenz==NULL keine Subgleichungen existieren */
}

void SGM_DruckeSGM(void)
{
   SGM_ReferenzT runny;
   SubgleichungsT *temp;

   fprintf(IO_Strom, "-----\n");
   runny = ErsteKopfZelle;
   while (runny != NULL){
      temp = runny->erste;
      do{			/* Die Subgleichungsliste zu der Gleichung enthaelt mindestens ein Element. */
	 fprintf(IO_Strom, "%ld.",TP_ElternNr((GleichungsT) (runny->objekt)));
	 IO_DruckeFlex("%t = %t\n", temp->links, temp->rechts);
	 temp = temp->naechste;
      }while(temp != NULL);
      runny = runny->naechste;
   }
}

/* ----------------- Durchlauf der SGM ---------------------*/

void SGM_StarteDurchlauf(void)
/* Durchlauf initialisieren */
{
   if (ErsteKopfZelle != NULL){
      nochSGs = TRUE;
      aktKopfZelle = ErsteKopfZelle;
      aktSG = aktKopfZelle->erste;
   }
   else
      nochSGs = FALSE;
}


void SGM_NaechsteSubgleichung(UTermT *links, UTermT *rechts, SGM_ReferenzT *objekt)
/* Naechste zu behandelnde Subgleichung holen; links und rechts enthalten nach dem Aufruf die linke und
   rechte Seite dieser Subgleichung.
   Vorbed.: Es gibt noch zu behandelnde SGs.
*/
{
   *links = aktSG->links;
   *rechts = aktSG->rechts;
   *objekt = aktKopfZelle->objekt;
   if (aktSG->naechste != NULL)
      aktSG = aktSG->naechste;
   else if (aktKopfZelle->naechste != NULL){
      aktKopfZelle = aktKopfZelle->naechste;
      aktSG = aktKopfZelle->erste;
   }
   else
      nochSGs = FALSE;
}

void SGM_NaechsteGleichung(void)
/* Schaltet beim Durchlauf auf die Subgleichungen der naechsten Gleichung, falls sie existieren.
   Wird verwendet, wenn die aktuelle Subgleichung subsummiert werden konnte und deshalb die uebrigen Subgleichungen
   der aktuellen Gleichung nicht mehr betrachtet werden muessen.*/
{
  if (aktKopfZelle->naechste != NULL){
    aktKopfZelle = aktKopfZelle->naechste;
    aktSG = aktKopfZelle->erste;
  }
  else
    nochSGs = FALSE;
}

BOOLEAN SGM_NochSubgleichungen(void)
/* Liefert TRUE, wenn es in diesem Durchlauf noch unbehandelte SGs gibt.*/
{
   return nochSGs;
}
